/***********************************************************************
 * This file is part of module_imageviewer.
 *
 * module_imageviewer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imageviewer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imageviewer.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QTimer>
#include <QDebug>
#include <QMediaPlayer>
#include <QApplication>
#include <QDateTime>

#ifdef _WIN32
#include <QVideoWidget>
#endif

#include "imagepreloader.h"
#include "misctools.h"

ImagePreloader::ImagePreloader()
    : _currentPosition(0)
    , _rawDisplayed(true)
    //, _player(new QMediaPlayer())
    //, _videoOut(0)
{
#ifdef _WIN32
        // windows needs something to render to
        // our "test" player will render on a video widget
        //_videoOut = new QVideoWidget;
        //_videoOut->setVisible(false);
        //_player->setVideoOutput(_videoOut);
#endif

    connect(&_loadingThread, SIGNAL(doneLoading(MiscTools::pQImage,QString)),
            SLOT(onImageLoaded(MiscTools::pQImage,QString)),
            Qt::QueuedConnection);
    //connect(&_mediaTimer, SIGNAL(timeout()), SLOT(checkForVideo()));
}

ImagePreloader::~ImagePreloader()
{    
    disconnect();
}

void ImagePreloader::setRawFormats(QStringList rawFormats)
{
    _myRawFormats = rawFormats;
    for (int i=0; i<_myRawFormats.size(); i++) {
        _myRawFormats[i] = _myRawFormats[i].toLower();
    }
}

ImageDataPtr ImagePreloader::getFileInfoPtr(const int position)
{
    QMutexLocker l(&_updateList);
    if (position < 0 || position > _mapToOrig.size() || _mapToOrig[position] > _imagesOrig.size()) {
        return 0;
    }
    return _imagesOrig[_mapToOrig[position]].file;
}

ImageDataPtr ImagePreloader::getFileInfoPtrOverallList(const int position)
{
    QMutexLocker l(&_updateList);
    if (position < 0 || position > _imagesOrig.size()) {
        return 0;
    }
    return _imagesOrig[position].file;
}

int ImagePreloader::setImages(const QList<ImageDataPtr>& newImages,
                              const int positionOrig,
                              int& positionNew)
{
    QMutexLocker l(&_updateList);

    QList<LoadImageInfo> newListTmp;
    for (int i=0; i<(int) newImages.size(); i++) {
        LoadImageInfo tmp = LoadImageInfo(newImages[i]);
        newListTmp.push_back( tmp );
    }

    _imagesOrig.clear();
    _imagesOrig = newListTmp;
    _loadingThread.clear();

    positionNew = _createImageMap(positionOrig >= 0 && positionOrig < newImages.size() ? positionOrig : 0);
    _currentPosition = positionNew;

    _requestNewLoadings();
    _loadingThread.addTarget(_imagesOrig[_mapToOrig[_currentPosition]].file->getFileInfo().absoluteFilePath(),
            true);

    return _mapToOrig.size();
}

int ImagePreloader::setRawDisplayed(const bool rawDisp,
                                    int& newNumImages)
{
    QMutexLocker l(&_updateList);
    if (_rawDisplayed == rawDisp) {
        // nothing has changed
        newNumImages = _mapToOrig.size();
        return -1;
    }

    _rawDisplayed = rawDisp;

    // empty anyways ?
    if (_currentPosition < 0
            || _currentPosition >= _imagesOrig.size()
            || _mapToOrig.size() == 0) {
        return -1;
    }
    // find position in raw image
    const int positionRaw = _mapToOrig[_currentPosition];
    // create new map with that position
    _createImageMap(positionRaw);

    _loadingThread.clear();
    _requestNewLoadings();

    newNumImages = _mapToOrig.size();
    return _currentPosition;
}

int ImagePreloader::_createImageMap(const int desiredPositionOrig)
{
    int desiredPosition = desiredPositionOrig;
    _mapToOrig.clear();
    _currentPosition = 0;

    if (_rawDisplayed) {
        // all raw images are displayed, so
        // we create an identity map
        _currentPosition = desiredPosition;
        _mapToOrig.resize(_imagesOrig.size());
        for (int i=0; i<_mapToOrig.size(); i++) {
            _mapToOrig[i] = i;
        }
    } else {
        // do not show raw images, if there's a corresponding
        // (same basename) file which is not of raw format
        const int s = (int)_imagesOrig.size();
        for (int i=0; i<s; i++) {
            const QString basename = _imagesOrig[i].file->getFileInfo().completeBaseName();
            const QString ext = _imagesOrig[i].file->getFileInfo().suffix().toLower();
            const QString path = _imagesOrig[i].file->getFileInfo().canonicalPath();

            if (_myRawFormats.contains(ext)) {
                // if it is indeed a raw file, then try to find if there exists a
                // corresponding jpg file. if not DO add raw file
                bool foundOther = false;
                for (int j=0; j<s; j++) {
                    if (j == i) {
                        continue;
                    }
                    if (_imagesOrig[j].file->getFileInfo().completeBaseName() == basename
                            && _imagesOrig[j].file->getFileInfo().canonicalPath() == path) {
                        const QString ext2 = _imagesOrig[j].file->getFileInfo().suffix().toLower();

                        if (!_myRawFormats.contains(ext2)) {
                            // found another file
                            foundOther = true;

                            // if the current raw file happened to be wished to be displayed
                            // change desired position to the other, not-raw file
                            if (i == desiredPosition) {
                                desiredPosition = j;
                            }
                            break;
                        }
                    }
                }
                if (!foundOther) {
                    _mapToOrig.append(i);
                }
            } else {
                _mapToOrig.append(i);
            }
        }

        // find corrected current position,
        // this does apply for the position shift
        // and the change of file if desiredposition was
        // a raw file
        for (int i=0; i<_mapToOrig.size(); i++) {
            if (_mapToOrig[i] == desiredPosition) {
                _currentPosition = i;
            }
        }
    }

    return _currentPosition;

}

void ImagePreloader::checkForVideo()
{
	/*
    bool newData = false;
    if (_player->mediaStatus() != QMediaPlayer::NoMedia) {
        if (_player->mediaStatus() == QMediaPlayer::LoadingMedia) {
            _mediaTimer.stop();
            _mediaTimer.setSingleShot(true);
            _mediaTimer.setInterval(100);
            _mediaTimer.start();
        } else {
            // find video again
            const QFileInfo test(_player->media().canonicalUrl().toLocalFile());
            for (int j=0; j<(int) _mapToOrig.size(); j++) {
                const int posOrig = _mapToOrig[j];
                if (_imagesOrig[posOrig].file->getFileInfo() ==  test) {
                    if (_player->mediaStatus() == QMediaPlayer::InvalidMedia) {
                        _imagesOrig[posOrig].status = LoadingStatusFailedLoading;
                        emit errorLoading(test.canonicalFilePath());
                    } else {
                        _imagesOrig[posOrig].status = LoadingStatusIsVideo;
                        emit doneLoading(test.canonicalFilePath());
                    }
                    break;
                }
            }
            _player->stop();
            newData = true;
        }
    }

    if (!_playlist.isEmpty() &&
            (_player->mediaStatus() == QMediaPlayer::NoMedia || newData) ) {
        _player->setMedia(QUrl::fromLocalFile(_playlist.first()));
        _playlist.pop_front();
        _mediaTimer.stop();
        _mediaTimer.setSingleShot(true);
        _mediaTimer.setInterval(100);
        _mediaTimer.start();
    }*/
}

void ImagePreloader::onImageLoaded(const MiscTools::pQImage image,
                                   const QString& canonicalFilePath)
{
    QMutexLocker l(&_updateList);
    MiscTools::pQPixmap pixmap (QPixmap::fromImage(*image));
    // update respective file
    for (int j=0; j<(int) _mapToOrig.size(); j++) {
        const int posOrig = _mapToOrig[j];
        if (posOrig < 0 || posOrig >= _imagesOrig.size()) {
            continue;
        }
        if (_imagesOrig[posOrig].file->getFileInfo().canonicalFilePath() ==  canonicalFilePath) {
            if (pixmap->width() > 0 && pixmap->height() > 0) {
                _imagesOrig[posOrig].status = LoadingStatusIsLoaded;
            } else if (_imagesOrig[posOrig].status != LoadingStatusIsTestedForVideo) {
				if (_videoFormats.contains(_imagesOrig[posOrig].file->getFileInfo().suffix().toLower())) {
					_imagesOrig[posOrig].status = LoadingStatusIsVideo;
					emit doneLoading(canonicalFilePath);
				}
				else {
					_imagesOrig[posOrig].status = LoadingStatusFailedLoading;
					emit errorLoading(canonicalFilePath);
				}
				/*
                _imagesOrig[posOrig].status = LoadingStatusIsTestedForVideo;
                // test if it is a video
                _playlist.append(canonicalFilePath);
                _mediaTimer.stop();
                _mediaTimer.setSingleShot(true);
                _mediaTimer.setInterval(100);
                _mediaTimer.start();
				*/
                break;
            }

            // adapt to possible inherent orientation of the image
            // const QString suff = _imagesOrig[posOrig].file->getFileInfo().suffix().toLower();
            const int diffRot = _imagesOrig[posOrig].addDesiredRotation;
            // make sure the original transform already cares
            // for user desired rotation (i.e. if still loading and the
            // user has already pressed "rotate"):
            QTransform transform = QTransform().rotate(diffRot);
                *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform);

            emit doneLoading(canonicalFilePath);
            break;
        }
    }
}

void ImagePreloader::_requestNewLoadings()
{
    // otherwise check for next/previous/first/last elements to be loaded
    for (int i=0; i<NUM_IMAGES_TO_EACH_SIDE; i++) {
        //end
        int p = i + (_mapToOrig.size() - NUM_IMAGES_TO_EACH_SIDE) ;
        if ( p >=0 ) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _loadingThread.addTarget(_imagesOrig[_mapToOrig[p]].file->getFileInfo().absoluteFilePath(),
                        false,
                        true);
            }
        }
        // start
        p = (NUM_IMAGES_TO_EACH_SIDE-1) - i;
        if (p < _mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _loadingThread.addTarget(_imagesOrig[_mapToOrig[p]].file->getFileInfo().absoluteFilePath(),
                        false,
                        true);
            }
        }
        // previous
        p = _currentPosition-(NUM_IMAGES_TO_EACH_SIDE-i);
        if ( p >=0 && p <_mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _loadingThread.addTarget(_imagesOrig[_mapToOrig[p]].file->getFileInfo().absoluteFilePath(),
                        false,
                        true);
            }
        }
        // following
        p = _currentPosition+(NUM_IMAGES_TO_EACH_SIDE-i);
        if ( p >=0 && p <_mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _loadingThread.addTarget(_imagesOrig[_mapToOrig[p]].file->getFileInfo().absoluteFilePath(),
                        false,
                        true);
            }
        }
    }

    // will start after possibly high priority has been added
    QTimer::singleShot(0, &_loadingThread, SLOT(start()));

    // check for out of range, to delete pixmaps (don't keep too much memory)
    const int s = _mapToOrig.size();
    const int buff = 5;
    for (int i=0; i<s; i++) {
        if ( abs(i -s) > buff * NUM_IMAGES_TO_EACH_SIDE
             && i > buff * NUM_IMAGES_TO_EACH_SIDE
             && abs(i-_currentPosition) > buff * NUM_IMAGES_TO_EACH_SIDE) {
            if (_imagesOrig[_mapToOrig[i]].status == LoadingStatusIsLoaded) {
                _imagesOrig[_mapToOrig[i]].status = LoadingStatusNotLoaded;
                *_imagesOrig[_mapToOrig[i]].pixmap = QPixmap();
            }
        }
    }
}

bool ImagePreloader::isVideoAtPosition(const int position) const
{
    QMutexLocker l(&_updateList);
    return (!(position < 0 || position >= _mapToOrig.size()
               || _mapToOrig[position] < 0 || _mapToOrig[position]  >= _imagesOrig.size())
            && (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsVideo) );
}

MiscTools::pQPixmap ImagePreloader::getImageAtPosition(const int position,
                                           bool& found,
                                           QString& filepath,
                                           bool& isloading,
                                           bool& isVideo)
{
    QMutexLocker l(&_updateList);
    found = true;

    if (position < 0
            || position >= _mapToOrig.size()
            || _mapToOrig[position] < 0
            || _mapToOrig[position]  >= _imagesOrig.size()) {
        found = false;
        filepath = "";
        return MiscTools::pQPixmap();
    }

    // adapt position
    filepath = _imagesOrig[_mapToOrig[position]].file->getFileInfo().canonicalFilePath();
    _currentPosition = position;

    isloading = _imagesOrig[_mapToOrig[position]].status == LoadingStatusIsLoading
            || _imagesOrig[_mapToOrig[position]].status == LoadingStatusNotLoaded;

    filepath = _imagesOrig[_mapToOrig[position]].file->getFileInfo().canonicalFilePath();
    if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsLoaded) {
        found = true;
        if (_imagesOrig[_mapToOrig[position]].pixmap != 0) {
            return *_imagesOrig[_mapToOrig[position]].pixmap;
        } else {
            return MiscTools::pQPixmap();
        }
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsVideo) {
        found = false;
        return MiscTools::pQPixmap();
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusFailedLoading) {
        found = false;
        return MiscTools::pQPixmap();
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsTestedForVideo) {
        isVideo = true;
        found = false;
        return MiscTools::pQPixmap();
    } else {
        // the desired pixmap has not been loaded yet
        // first of all check, if it is already being loaded by a thread

        found = false;
        _requestNewLoadings();

        // place here to make sure it has high priority
        _loadingThread.addTarget(filepath, true);

        return MiscTools::pQPixmap();
    }
}


void ImagePreloader::clearImages()
{
    QMutexLocker l(&_updateList);
    _loadingThread.clear();
    _imagesOrig.clear();
    _currentPosition=0;
    _mapToOrig.clear();
}

void ImagePreloader::setDesiredOrientation(const int position,
                                           const int addRotation)
{
    QMutexLocker l(&_updateList);
    if (position < 0 || position >= _mapToOrig.size()) {
        qCritical() << "invalid index for rotation";
        return;
    }
    const int posOrig = _mapToOrig[position];
    _imagesOrig[posOrig].addDesiredRotation = (_imagesOrig[posOrig].addDesiredRotation + addRotation) % 360;
    if (_imagesOrig[posOrig].status == LoadingStatusIsLoaded) {
        QTransform transform;
        *_imagesOrig[posOrig].pixmap = _imagesOrig[posOrig].pixmap->transformed(transform.rotate(addRotation));
    }
}
