/***********************************************************************
 * This file is part of module_imageviewer.
 *
 * module_imageviewer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imageviewer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imageviewer.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include "imageviewerloader.h"

#include <QDebug>
#include <QSvgRenderer>
#include <QDesktopWidget>
#include <QPainter>

#include "misctools.h"


ImageViewerLoader::ImageViewerLoader(QObject *parent)
    : QThread(parent)
    , _imSize (MiscTools::MiscTools::retrieveDesktopResolution().size())
    , _stopped(false)
    , _abortCurrentStartNext(false)
{
    //connect(this, SIGNAL(finished()), SLOT(deleteLater()));
}

ImageViewerLoader::~ImageViewerLoader()
{
    disconnect();
    _fileListMutex.lock();
    _files.clear();
    _stopped = true;
    _condition.wakeOne();
    _fileListMutex.unlock();
    wait();
}

void ImageViewerLoader::deleteInstance()
{
    // is it necessary to put this at the end of the even
    this->deleteLater();
}

void ImageViewerLoader::clear()
{
    _abortAndLoadNext = true;
    QMutexLocker l(&_fileListMutex);
    _files.clear();
}

void ImageViewerLoader::addTarget(const QString& target,
                                  const bool highPriority,
                                  const bool noStart)
{
    QMutexLocker l(&_fileListMutex);

    if (_current == target)
    {
        // already loading -> do nothing
        return;
    }

    if (highPriority) {
        _abortCurrentStartNext = true;
    }

    _files.push_front(target);

    // find duplicates and delete (work from back)
    for (int i=_files.size()-1; i>=0; i--) {
        for (int j=i-1; j>=0; j--) {
            if (_files[i] == _files[j] || _files[i] == _current) {
                _files.erase(_files.begin() + i);
            }
        }
    }

    if (noStart) {
        return;
    }

    startComputation();
}

void ImageViewerLoader::startComputation()
{
    _stopped = false;
    if (!isRunning()) {
        start(QThread::HighPriority);
    } else {
        _condition.wakeOne();
    }
}


#define TEST_ABORT(s) \
    if (_stopped || _abortAndLoadNext) { \
        _current = ""; \
        continue; \
    }

void ImageViewerLoader::run()
{
    while (!_stopped)
    {
        _abortAndLoadNext = false;
        {
            QMutexLocker l(&_fileListMutex);
            if (_files.size() == 0) {
                _condition.wait(&_fileListMutex);
                // continue: in the meantime it might have been stopped
                // or set empty
                continue;
            }
            _current = _files.first();
            _files.pop_front();
        }
        MiscTools::pQImage image;
        image = MiscTools::MiscTools::loadImageOrRaw(_current, true, &_abortAndLoadNext);
        TEST_ABORT(s);

        if (image->isNull() || image->width() == 0 || image->height() == 0) {
            // now try to load as SVG
            QSvgRenderer renderer(_current);

            if (renderer.isValid()) {
                //const QRect deskSize = MiscTools::MiscTools::retrieveDesktopResolution();
                const double deskRatio = _imSize.width() / double(_imSize.height());
                const QSize suggestedSize = renderer.defaultSize();
                const double suggRatio = suggestedSize.width() / double(suggestedSize .height());
                QSize targetSize;
                if (deskRatio <= suggRatio) {
                    targetSize.setWidth(_imSize.width());
                    targetSize.setHeight(_imSize.width()/suggRatio);
                } else {
                    targetSize.setHeight(_imSize.height());
                    targetSize.setWidth(_imSize.height()*suggRatio);
                }
                //qDebug() << _file << deskRatio << suggRatio <<targetSize;
                *image = QImage(targetSize, QImage::Format_ARGB32);
                image->fill(Qt::white);
                TEST_ABORT(s)

                        QPainter painter(image.get());
                //painter.setBackground(Qt::green);
                //painter.setBrush(Qt::red);
                //painter.setBackgroundMode(Qt::OpaqueMode);
                //painter.setPen(Qt::white);*/
                renderer.render(&painter);
                painter.end();
                TEST_ABORT(s);
                *image = image->convertToFormat(QImage::Format_RGB32,
                                                Qt::ImageConversionFlag::DiffuseDither
                                                | Qt::ImageConversionFlag::ThresholdAlphaDither);

            }
        } else {
            *image = image->scaled(_imSize, Qt::KeepAspectRatio, Qt::SmoothTransformation);
        }

        TEST_ABORT(s);
        emit doneLoading(image, _current);
        _current = "";
    }
}
