/***********************************************************************
 * This file is part of module_imageviewer.
 *
 * module_imageviewer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imageviewer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imageviewer.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include "viewimagedialog.h"
#include "ui_viewimagedialog.h"
#include <QApplication>
#include <QSettings>
#include <QTimer>
#include <QDebug>
#include <QDir>
#include <QTime>
#include <QContextMenuEvent>
#include <QMenu>
#include <QKeySequence>
#include <QKeyEvent>
#include <QGraphicsDropShadowEffect>
#include <QStyle>
#include <QWhatsThis>
#include <QTransform>
#include <QFrame>
#include <QDesktopWidget>
#include <QVBoxLayout>
#include <QFont>

#include <cstdlib>

#include "imageviewerloader.h"
#include "misctools.h"

#define IMAGE_ORGANIZATION_ORG "Dominik Rueß"
#define IMAGE_ORGANIZATION_APP "ImageViewer"

static QRectF removeMargin11945(const QRectF& sceneRect, const QSize& viewerSize) {
    // http://stackoverflow.com/questions/18387276/qt-4-8-5-qgraphicsviewfitinview-fails-to-fit-exactly
    // https://bugreports.qt-project.org/browse/QTBUG-11945
    const int bugMargin = 2;
    const double mx = sceneRect.width()/viewerSize.width()*bugMargin;
    const double my = sceneRect.height()/viewerSize.height()*bugMargin;
    return sceneRect.adjusted(mx, my, -mx, -my);
}

// singleton static pointer
ViewImageDialog* ViewImageDialog::_instance = NULL;

ViewImageDialog::ViewImageDialog(QWidget *parent)
    : QDialog(parent, Qt::Window)
    , ui(new Ui::ViewImageDialog)
    , _currentPosition(-1)
    , _basenameLength(0)
    , _reset(true)
    , _currNumImages(0)
    , _fullScreen(false)
    , _fileOverlayActive(true)
    , _filenameOverlay(NULL)
	, _videoControlsGroup(NULL)
    , _timeToDisappear(3000)
    , _videoPlayer(new QMediaPlayer(0, QMediaPlayer::VideoSurface))
    , _scene(new QGraphicsScene(this))
    , _textItem(new QGraphicsTextItem)
    , _pixmapItem(new QGraphicsPixmapItem)
    , _videoItem(new QGraphicsVideoItem)
    , _currentDisplayMode(DisplayMode_Text)
{
	setFocusPolicy(Qt::StrongFocus);
	ui->setupUi(this);
	QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));

    ui->graphicsView->setBackgroundRole(QPalette::Base);
    ui->graphicsView->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    ui->graphicsView->setAlignment(Qt::AlignCenter);
    ui->graphicsView->setBackgroundRole(QPalette::Dark);
    ui->graphicsView->viewport()->setMouseTracking(true);
    ui->graphicsView->viewport()->installEventFilter(this);
	
    ui->graphicsView->setScene(_scene);
    _videoItem->setVisible(true);
	_videoItem->setSize(MiscTools::MiscTools::retrieveDesktopResolution().size());
    
	_videoItem->setZValue(1);
    _pixmapItem->setVisible(false);
    _pixmapItem->setZValue(2);
    _textItem->setZValue(10);
    _textItem->setPos(0,0);
    QGraphicsDropShadowEffect *shadow1 = new QGraphicsDropShadowEffect();
    shadow1->setBlurRadius(5);
    shadow1->setOffset(3,3);
    _textItem->setGraphicsEffect(shadow1);
    _textItem->setDefaultTextColor(QColor(0,255,148,255));
    QFont font;
    font.setPointSize(1.5*font.pointSizeF());
    font.setFamily("Verdana");
    _textItem->setFont(font);
    _pixmapItem->setPos(0,0);
    _videoItem->setPos(0,0);
	_scene->addItem(_videoItem);
	_scene->addItem(_textItem);
	_scene->addItem(_pixmapItem);

    _videoControlsGroup = new QFrame();
    _videoControlsGroup->setLayout(new QHBoxLayout());
    _videoControlsGroup->setGeometry(0,0,MiscTools::MiscTools::retrieveDesktopResolution().width(),75);
    connect(&_controlsDisappearTimer, SIGNAL(timeout()), SLOT(disappearControls()));

    _playButton = new QToolButton(_videoControlsGroup);
    _playButton->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    _videoControlsGroup->layout()->addWidget(_playButton);
    _videoPositionSlider = new QSlider( Qt::Horizontal, _videoControlsGroup);
    _videoControlsGroup->layout()->addWidget(_videoPositionSlider);
    _muteButton = new QToolButton(_videoControlsGroup);
    _muteButton->setCheckable(true);
    _muteButton->setChecked(settings.value("viewimage_mute", false).toBool());
    muteToggled();
    _videoPlayer->setMuted(_muteButton->isChecked());
    _videoControlsGroup->layout()->addWidget(_muteButton);
    _videoVolumeSlider = new QSlider(Qt::Horizontal, _videoControlsGroup);
    _videoVolumeSlider->setMaximumWidth(100);
    _videoVolumeSlider->setRange(0,100);
    _videoVolumeSlider->setValue(settings.value("viewimage_volume", 100).toInt());
    _videoPlayer->setVolume(_videoVolumeSlider->value());
    _videoControlsGroup->layout()->addWidget(_videoVolumeSlider);
    static_cast<QVBoxLayout*>(this->layout())->insertWidget(1, _videoControlsGroup);
    //_groupProxy = _scene->addWidget(_videoControlsGroup);
    //_groupProxy->setZValue(100);

    connect(_videoPositionSlider, SIGNAL(valueChanged(int)), SLOT(sliderPositionChange(int)));
    connect(_playButton, SIGNAL(clicked()), SLOT(playPauseToggled()));
    connect(_videoVolumeSlider, SIGNAL(valueChanged(int)), _videoPlayer, SLOT(setVolume(int)));
    connect(_muteButton, SIGNAL(toggled(bool)), _videoPlayer, SLOT(setMuted(bool)));
    connect(_muteButton, SIGNAL(toggled(bool)), SLOT(muteToggled()));
    setVideoControlsVisible(false);


    _displayText(ImageDataPtr(0), tr("No file list provided... (yet)"));

    setWindowTitle(tr("Image Viewer"));
    resize(500, 400);

    zoomInAct = new QAction(tr("Zoom &In (25%)"), this);
    zoomInAct->setShortcut(QKeySequence(tr("Ctrl++")));
    connect(zoomInAct, SIGNAL(triggered()), this, SLOT(zoomIn()));
    zoomInAct->setShortcutContext(Qt::WindowShortcut);
    addAction(zoomInAct);

    zoomOutAct = new QAction(tr("Zoom &Out (25%)"), this);
    zoomOutAct->setShortcut(QKeySequence(tr("Ctrl+-")));
    connect(zoomOutAct, SIGNAL(triggered()), this, SLOT(zoomOut()));
    zoomOutAct->setShortcutContext(Qt::WindowShortcut);
    addAction(zoomOutAct);

    normalSizeAct = new QAction(tr("&Normal Size"), this);
    normalSizeAct->setShortcut(QKeySequence(tr("Ctrl+S")));
    connect(normalSizeAct, SIGNAL(triggered()), this, SLOT(normalSize()));
    normalSizeAct->setShortcutContext(Qt::WindowShortcut);
    addAction(normalSizeAct);

    fitToWindowAct = new QAction(tr("&Fit to Window"), this);
    fitToWindowAct->setCheckable(true);
    fitToWindowAct->setShortcut(QKeySequence(tr("Ctrl+F")));
    connect(fitToWindowAct, SIGNAL(triggered()), this, SLOT(fitToWindowToggle()));
    fitToWindowAct->setShortcutContext(Qt::WindowShortcut);
    addAction(fitToWindowAct);

    fullScreenAct = new QAction(tr("&Full screen mode"), this);
    fullScreenAct->setCheckable(true);
    fullScreenAct->setShortcut(QKeySequence(tr("F11")));
    fullScreenAct->setCheckable(true);
    connect(fullScreenAct, SIGNAL(triggered()), this, SLOT(_toggleFullscreen()));
    fullScreenAct->setShortcutContext(Qt::WindowShortcut);
    fullScreenAct->setToolTip(tr("Enter or leave fullscreen mode"));
    fullScreenAct->setStatusTip(fullScreenAct->toolTip());
    fullScreenAct->setWhatsThis(fullScreenAct->toolTip());
    addAction(fullScreenAct);

    nextImageAct = new QAction(tr("Next Image"), this);
    nextImageAct->setShortcut(QKeySequence(tr("PgDown", "prev image")));
    connect(nextImageAct, SIGNAL(triggered()), this, SLOT(on_toolButton_next_clicked()));
    nextImageAct->setShortcutContext(Qt::WindowShortcut);
    addAction(nextImageAct);

    lastImageAct = new QAction(tr("Last Image"), this);
    lastImageAct->setShortcut(QKeySequence(tr("End", "last image")));
    connect(lastImageAct, SIGNAL(triggered()), this, SLOT(on_toolButton_toEnd_clicked()));
    lastImageAct->setShortcutContext(Qt::WindowShortcut);
    addAction(lastImageAct);

    firstImageAct = new QAction(tr("Previous Image"), this);
    firstImageAct->setShortcut(QKeySequence(tr("Home", "first image")));
    connect(firstImageAct, SIGNAL(triggered()), this, SLOT(on_toolButton_toStart_clicked()));
    firstImageAct->setShortcutContext(Qt::WindowShortcut);
    addAction(firstImageAct);

    prevImageAct = new QAction(tr("Previous Image"), this);
    prevImageAct->setShortcut(QKeySequence(tr("PgUp", "next image")));
    connect(prevImageAct, SIGNAL(triggered()), this, SLOT(on_toolButton_previous_clicked()));
    prevImageAct->setShortcutContext(Qt::WindowShortcut);
    addAction(prevImageAct);

    rotLeftAct = new QAction(tr("Rotate Left"), this);
    rotLeftAct->setShortcut(QKeySequence(tr("l", "rotate image to the left")));
    connect(rotLeftAct, SIGNAL(triggered()), this, SLOT(on_toolButton_rotLeft_clicked()));
    rotLeftAct->setShortcutContext(Qt::WindowShortcut);
    addAction(rotLeftAct);

    rotRightAct = new QAction(tr("Rotate Right"), this);
    rotRightAct->setShortcut(QKeySequence(tr("r", "rotate image to the right")));
    connect(rotRightAct, SIGNAL(triggered()), this, SLOT(on_toolButton_rotRight_clicked()));
    rotRightAct->setShortcutContext(Qt::WindowShortcut);
    addAction(rotRightAct);

    QTimer::singleShot(0, this, SLOT(loadSettings()));

    _loader.setRawDisplayed(!ui->checkBox_hideRaw->isChecked(), _currNumImages);
	
    connect(&_loader, SIGNAL(doneLoading(QString)),
            this, SLOT(_doneLoading(QString)), Qt::QueuedConnection);
    connect(&_loader, SIGNAL(errorLoading(QString)),
            this, SLOT(_doneLoading(QString)), Qt::QueuedConnection);

    connect(ui->checkBox_hideRaw, SIGNAL(toggled(bool)),
            this, SLOT(on_checkBox_hideRaw_toggled(bool)));

	
    this->setContextMenuPolicy(Qt::CustomContextMenu);
    connect (this, SIGNAL(customContextMenuRequested(QPoint)),
             this, SLOT(_contextMenuEvent(const QPoint&)));

    _filenameOverlay = new QLabel(this);//ui->scrollArea);
    QGraphicsDropShadowEffect* effect = new QGraphicsDropShadowEffect();
    effect->setBlurRadius(3);
    effect->setOffset(1.0);
    _filenameOverlay->setGraphicsEffect(effect);
    _filenameOverlay->setStyleSheet("QLabel { color : #66FF66; }");

    _filenameOverlay->setVisible(isFullScreen());

    QStyle * style = this->style();
    ui->toolButton_whatsthis->setIcon(style->standardPixmap(QStyle::SP_MessageBoxQuestion));
	
	
	connect(_videoPlayer, SIGNAL(stateChanged(QMediaPlayer::State)), this, SLOT(playerStateChanged(QMediaPlayer::State)), Qt::QueuedConnection);
	connect(_videoPlayer, SIGNAL(mediaStatusChanged(QMediaPlayer::MediaStatus)), this, SLOT(mediaStatusChanged(QMediaPlayer::MediaStatus)), Qt::QueuedConnection);
	connect(_videoPlayer, SIGNAL(positionChanged(qint64)), this, SLOT(positionChanged(qint64)), Qt::QueuedConnection);
	connect(_videoPlayer, SIGNAL(durationChanged(qint64)), this, SLOT(durationChanged(qint64)), Qt::QueuedConnection);
	connect(_videoPlayer, SIGNAL(error(QMediaPlayer::Error)), this, SLOT(handleError(QMediaPlayer::Error)), Qt::QueuedConnection);
	
    this->installEventFilter(this);
	
	_videoPlayer->setVideoOutput(_videoItem);
}


bool ViewImageDialog::eventFilter(QObject *obj, QEvent *event) {
    if (event->type() == QEvent::MouseMove && _currentDisplayMode == DisplayMode_Video) {
        setVideoControlsVisible(true);
        _controlsDisappearTimer.stop();
        _controlsDisappearTimer.setSingleShot(true);
        _controlsDisappearTimer.setInterval(_timeToDisappear);
        _controlsDisappearTimer.start();
    } else if ( event->type() == QEvent::MouseButtonDblClick ) {
        if (obj->parent() == static_cast<QObject*>(ui->graphicsView)
                || obj == static_cast<QObject*>(_scene)
                || obj == static_cast<QObject*>(ui->graphicsView)) {
            event->accept();
            _toggleFullscreen();
        }
    } else if ( event->type() == QEvent::KeyPress) {
        const QKeyEvent *ke = static_cast<QKeyEvent *>(event);
        if(ke->key()==Qt::Key_Space) {
            // start/stop playing
            if (_currentDisplayMode == DisplayMode_Video) {
                _playButton->click();
            }
			event->accept();
        }
		else if (ke->key() == Qt::Key_Escape && _fullScreen) {
			_toggleFullscreen();
			event->accept();
		}
    }
    return QWidget::eventFilter(obj, event);
}

void ViewImageDialog::resetSettings()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.clear();
}

void ViewImageDialog::loadSettings()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    //this->restoreGeometry(settings.value("album_viewimage_geometry").toByteArray());
    if (isFullScreen()) {
        _fullScreen = true;
        _toggleFullscreen();
    }
    ui->checkBox_hideRaw->setChecked(settings.value("album_viewimage_rawchecked", true).toBool());
}

ViewImageDialog::~ViewImageDialog()
{
    disconnect();
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    if (_fullScreen) {
        _toggleFullscreen();
    }
    settings.setValue("album_viewimage_geometry", this->saveGeometry());
    settings.setValue("album_viewimage_rawchecked", ui->checkBox_hideRaw->isChecked());
    settings.setValue("viewimage_volume", _videoPlayer->volume());
    settings.setValue("viewimage_mute", _muteButton->isChecked());
    delete ui;
}

ViewImageDialog* ViewImageDialog::instance(QWidget* parent)
{
  Q_UNUSED(parent);
  if(_instance == NULL){
    _instance = new ViewImageDialog;
  }
  return _instance;
}

void ViewImageDialog::closeInstance()
 {
     if(_instance == NULL) {
         return;
     } else {
         _instance->close();
     }
 }

void ViewImageDialog::setRawFormats(const QStringList &rawList)
{
    _loader.setRawFormats(rawList);
}


void ViewImageDialog::closeEvent ( QCloseEvent * )
{
    _videoPlayer->stop();
    _videoPlayer->setMedia(QMediaContent());
    _loader.clearImages();
}

ImageDataPtr ViewImageDialog::getFileInfoPtr(const int position)
{
    return _loader.getFileInfoPtr(position);
}

int ViewImageDialog::getNumFilesOverall() const
{
    return _loader.getNumFilesOverall();
}

ImageDataPtr ViewImageDialog::getFileInfoPtrOverallList(const int position)
{
    return _loader.getFileInfoPtrOverallList(position);
}

void ViewImageDialog::setImages(const QList<ImageDataPtr>& newImages,
                                const int positionOrig,
                                const int basenameLength)
{    
    int newPosition = positionOrig;
    _currNumImages = _loader.setImages(newImages, positionOrig, newPosition);
    ui->spinBox_number->setMinimum(1);
    ui->spinBox_number->setMaximum(_currNumImages);
    _basenameLength = basenameLength;
    loadPosition(newPosition);
}


void ViewImageDialog::on_checkBox_hideRaw_toggled(bool checked)
{
    const int newPos = _loader.setRawDisplayed(!checked,
                                               _currNumImages);
    ui->spinBox_number->setMaximum(_currNumImages);
    if (newPos >= 0) {
        loadPosition(newPos);
    }
}

void ViewImageDialog::loadPosition(const int positionOrig)
{
    int position = positionOrig;
    if (positionOrig == -1) {
        position = _currentPosition;
    } else if (position >= _currNumImages) {
        position = _currNumImages -1 ;
    }
    QString filepath;
    bool wasFound = false, videoloading = false, isLoading = false;
    MiscTools::pQPixmap pixmap;

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    pixmap = _loader.getImageAtPosition(position, wasFound, filepath, isLoading, videoloading);

    QApplication::restoreOverrideCursor();

    _currentPosition = position;
    _originalPixmap = pixmap;
    _showPixmap(filepath);
	bool resetVideo = false;
    if (wasFound) {
        if (_currentDisplayMode != DisplayMode_Image) {
            // viewport recovery doesn't quite work ...
            // TODO
            ui->graphicsView->setTransform(_lastTransformation);
            ui->graphicsView->setSceneRect(_lastSceneRect);
        }
        _pixmapItem->setPixmap(*_originalPixmap);
        _pixmapItem->setPos(0,0);
        _displayImage();
		resetVideo = true;
    } else if (_loader.isVideoAtPosition(position) 
			   && _videoPlayer->state() == QMediaPlayer::PlayingState) {
		_videoPlayer->stop();
        _videoPlayer->setMedia(QMediaContent());
        QTimer::singleShot(0, this, SLOT(loadPosition()));
        return;
    } else if (_loader.isVideoAtPosition(position)) {
		_videoItem->setVisible(true);
		_nextVideo = filepath;
		QTimer::singleShot(0, this, SLOT(loadVideo()));
        //_videoPlayTimer.setInterval(100);
        //_videoPlayTimer.setSingleShot(true);
        //_videoPlayTimer.start();
    } else if (!isLoading && !videoloading){
        _displayText(_loader.getFileInfoPtr(position),
                         tr("Could not load file \n'%1'").
                     arg(QFileInfo(filepath).fileName()));
		resetVideo = true;
    } else {
		if (videoloading) {
			_videoItem->setVisible(true);
		}
		else {
			resetVideo = true;
		}
        _displayText(_loader.getFileInfoPtr(position),
                     tr("Please wait, loading file ..."));
    }
	/*
	if (resetVideo == true) {
		_videoPlayer->stop();
		_videoPlayer->setMedia(0);
	}*/

    if (position == 0) {
        ui->toolButton_toStart->setEnabled(false);
        ui->toolButton_previous->setEnabled(false);
    }
    if (position == _currNumImages -1) {
        ui->toolButton_toEnd->setEnabled(false);
        ui->toolButton_next->setEnabled(false);
    }
    if (position > 0) {
        ui->toolButton_toStart->setEnabled(true);
        ui->toolButton_previous->setEnabled(true);
    }
    if (position < _currNumImages -1) {
        ui->toolButton_toEnd->setEnabled(true);
        ui->toolButton_next->setEnabled(true);
    }
}

void ViewImageDialog::loadVideo()
{
	_currentDisplayMode = DisplayMode_Video;
	_videoPlayer->setMedia(QUrl::fromLocalFile(_nextVideo));
}


void ViewImageDialog::_doneLoading(const QString& filepath)
{
    const ImageDataPtr tmp = getFileInfoPtr(_currentPosition) ;
    if (tmp.get() != 0) {
        const QString& tmp2 = tmp->getFileInfo().canonicalFilePath();
        if (filepath == tmp2) {
            QTimer::singleShot(0, this, SLOT(loadPosition()));
        }
    }
}


void ViewImageDialog::_showPixmap(const QString& fileName)
{
    QString subdirFilename = QDir::toNativeSeparators(fileName.right(fileName.length()-_basenameLength ));
    QString str = tr("out of %1. Filename: %2").arg(_currNumImages).arg(subdirFilename);
    ui->spinBox_number->setValue(_currentPosition+1);
    QFontMetrics fm( ui->label_file->font() );
    const int maxWidth = qMin(600, ui->label_file->width());
    if( maxWidth < fm.width( str ) ){
        ui->label_file->setToolTip( str );
        ui->label_file->setText( ui->label_file->fontMetrics().elidedText( str, Qt::ElideMiddle, maxWidth ));
    } else {
        ui->label_file->setText( str );
    }
    _updateFileOverlay(false, tr("file %1 ").arg(ui->spinBox_number->value())+ str);
    emit(newImageLabel());
}

void ViewImageDialog::resizeEvent(QResizeEvent * /* event */)
{
    switch(_currentDisplayMode) {
    case DisplayMode_Image:
        if (ui->toolButton_zoomWindow->isChecked() && _pixmapItem != 0) {
            fitToWindow(_pixmapItem);
        }
        break;
    case DisplayMode_Video:
        fitToWindow(_videoItem);

        break;
    case DisplayMode_Text:
    default:
        if (ui->toolButton_zoomWindow->isChecked()
                && _pixmapItem != 0
                && _pixmapItem->isVisible()) {
            fitToWindow(_pixmapItem);
        }
        ui->graphicsView->setTransform(QTransform());
        ui->graphicsView->centerOn(_textItem);
        break;
    }
}

//icon loading wrong

void ViewImageDialog::zoomIn()
{
    ui->toolButton_zoomWindow->setChecked(false);
    scaleImage(1.25);
}

void ViewImageDialog::zoomOut()
{
    ui->toolButton_zoomWindow->setChecked(false);
    scaleImage(0.8);
}

void ViewImageDialog::normalSize()
{

    ui->toolButton_zoomWindow->setChecked(false);

    ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    ui->graphicsView->setTransform(QTransform());
    if (_pixmapItem != 0) ui->graphicsView->centerOn(_pixmapItem);
    //_updateImageLabel();
}

void ViewImageDialog::fitToWindowToggle()
{
    if (_currentDisplayMode == DisplayMode_Image) {
        fitToWindow(_pixmapItem);
    } else if (_currentDisplayMode == DisplayMode_Video) {
        fitToWindow(_videoItem);
    }
}


void ViewImageDialog::fitToWindow(QGraphicsItem* item)
{
    if (item == 0) {
        return;
    }
    ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);    

    ui->graphicsView->setTransform(QTransform());
    ui->graphicsView->setSceneRect(item->sceneBoundingRect());
    ui->graphicsView->centerOn(item);
    const double scale = std::max(item->sceneBoundingRect().width()/ui->graphicsView->size().width(),
                                 item->sceneBoundingRect().height()/ui->graphicsView->size().height());

    if (_videoControlsGroup != 0) {
        _videoControlsGroup->setGeometry(0,
                                     ui->graphicsView->size().height() - 75,
                                     ui->graphicsView->size().width(),
                                     75);
    }

    QTransform transf = ui->graphicsView->transform();
    ui->graphicsView->setTransform(transf.scale(1.0/scale,1.0/scale));

    return;
    //ui->graphicsView->centerOn(0, 0);
    /*
    ui->graphicsView->setSceneRect(0,0,item->boundingRect().width(),item->boundingRect().height());

    ui->graphicsView->setTransform(QTransform());
    ui->graphicsView->centerOn(item);
    ui->graphicsView->fitInView(item, Qt::KeepAspectRatio);
    ui->graphicsView->centerOn(item);
    ui->graphicsView->fitInView(item, Qt::KeepAspectRatio);
    QRectF tmp = removeMargin11945(ui->graphicsView->sceneRect(),
                                   ui->graphicsView->size());
    qDebug() << "GERD" << item->boundingRect() << ui->graphicsView->sceneRect() << tmp << item->isVisible();
    ui->graphicsView->setSceneRect(tmp);
    */

    const QRectF itemRect = item->sceneBoundingRect();
    const double aspect = itemRect.size().width() / double(itemRect.size().height());
    const double aspectScene = ui->graphicsView->size().width() / double(ui->graphicsView->size().height());
    QRectF sceneRect;
    if (aspect > aspectScene) {
        sceneRect.setX(0);
        sceneRect.setWidth(itemRect.width());
        const double d = (1.0-aspectScene/aspect) * double(ui->graphicsView->size().height());
        sceneRect.setY(-0.5*d);
        sceneRect.setHeight(itemRect.height() + 0.5*d);
    } else {
        sceneRect.setY(0);
        sceneRect.setHeight(itemRect.height());
        const double d = (1.0 - aspectScene/aspect ) * double(ui->graphicsView->size().width());
        sceneRect.setX(-0.5*d);
        sceneRect.setWidth(itemRect.height() + 0.5*d);
    }
    QRectF tmp = removeMargin11945(sceneRect,
                                   ui->graphicsView->size());
    ui->graphicsView->setSceneRect(tmp);
    //ui->graphicsView->setSceneRect(item->sceneBoundingRect());
}

void ViewImageDialog::keyPressEvent(QKeyEvent *e) {

    if (e->key() == Qt::Key_Space) {
        if (_currentDisplayMode == DisplayMode_Video) {
            _playButton->click();
        }
    } else if(e->key() == Qt::Key_Escape) {
        // prevent closing/minimizing
    } else {
            QWidget::keyPressEvent(e);
    }
}

void ViewImageDialog::scaleImage(double factor)
{
    ui->graphicsView->scale(factor, factor);
}

void ViewImageDialog::on_toolButton_zoomIn_clicked()
{
    zoomInAct->trigger();
}

void ViewImageDialog::on_toolButton_zoomOut_clicked()
{
    zoomOutAct->trigger();
}

void ViewImageDialog::on_toolButton_zoomReal_clicked()
{
    normalSizeAct->trigger();
}

void ViewImageDialog::on_toolButton_zoomWindow_clicked()
{
    fitToWindowAct->trigger();
}

void ViewImageDialog::on_toolButton_toStart_clicked()
{
    if (_currentDisplayMode == DisplayMode_Image) {
        _lastTransformation = ui->graphicsView->transform();
        _lastSceneRect = ui->graphicsView->sceneRect();
    }
    loadPosition(0);
}

void ViewImageDialog::on_toolButton_previous_clicked()
{
    if (_currentDisplayMode == DisplayMode_Image) {
        _lastTransformation = ui->graphicsView->transform();
        _lastSceneRect = ui->graphicsView->sceneRect();
    }
    loadPosition(_currentPosition-1);
}

void ViewImageDialog::on_toolButton_next_clicked()
{
    if (_currentDisplayMode == DisplayMode_Image) {
        _lastTransformation = ui->graphicsView->transform();
        _lastSceneRect = ui->graphicsView->sceneRect();
    }
    loadPosition(_currentPosition+1);
}

void ViewImageDialog::on_toolButton_toEnd_clicked()
{
    if (_currentDisplayMode == DisplayMode_Image) {
        _lastTransformation = ui->graphicsView->transform();
        _lastSceneRect = ui->graphicsView->sceneRect();
    }
    loadPosition(_currNumImages - 1);
}

void ViewImageDialog::on_pushButton_leave_clicked()
{
    _videoPlayer->stop();
    _videoPlayer->setMedia(QMediaContent());
    _loader.clearImages();
    _displayText(ImageDataPtr(0), tr("Empty"));

    this->close();
}

void ViewImageDialog::_contextMenuEvent(const QPoint &pos)
{
    QMenu menu(this);
    menu.addAction(fullScreenAct);
    QAction* overlay = NULL;
    if (_fullScreen) {
        menu.addSeparator();
        overlay = new QAction(tr("Filename &overlay"), this);
        overlay->setCheckable(true);
        overlay->setChecked(_fileOverlayActive);
        overlay->setToolTip(tr("Disable or enable file path display for fullscreen mode"));
        overlay->setStatusTip(fullScreenAct->toolTip());
        overlay->setWhatsThis(fullScreenAct->toolTip());
        menu.addAction(overlay);
    }

    QAction *a = menu.exec(pos + this->pos()); //ui->graphicsView->mapFromScene(pos));
    menu.close();
    if (a == overlay && a != 0)  {
        _fileOverlayActive = overlay->isChecked();
        _filenameOverlay->setVisible(_fileOverlayActive);
        emit fileOverlayChanged();
    } else if (a == fullScreenAct && a != 0){
        // _toggleFullscreen will already be called by qaction itself
    }
    //delete overlay;
}

void ViewImageDialog::_updateFileOverlay(const bool onlyPos,
                                         const QString text)
{
    if (!onlyPos && _filenameOverlay != 0) {
        QFontMetrics fm( _filenameOverlay->font() );
        const int maxWidth = ui->graphicsView->width()-10;
        _filenameOverlay->setMinimumWidth(maxWidth);
        if( maxWidth < fm.width( text ) ){
            _filenameOverlay->setText(text);
        } else {
            _filenameOverlay->setText(_filenameOverlay->fontMetrics().elidedText( text, Qt::ElideMiddle, maxWidth ));
        }
    }
}

void ViewImageDialog::_toggleFullscreen()
{
    _fullScreen = !_fullScreen;

	if (_fullScreen) {
        ui->graphicsView->setStyleSheet("background-color: black;");
        ui->graphicsView->setFrameShape(QFrame::NoFrame);
        ui->graphicsView->setFrameShadow(QFrame::Plain);
		ui->frame_controls->setVisible(false);
		ui->frame_impath->setVisible(false);
		ui->frame_raw->setVisible(false);
		ui->verticalLayout->setMargin(0);
#ifndef _WIN32
		const QString session = QString(getenv("DESKTOP_SESSION")).toLower();

        if (session == "ubuntu" || session == "gnome") {
            // http://stackoverflow.com/a/24078775/1150303
            this->setWindowFlags(Qt::Window);
            this->showFullScreen();
        } else {
			showFullScreen();
        }
#else
		showFullScreen();
#endif
		// resize preview thumbnail to fit full screen
		if (_currentDisplayMode == DisplayMode_Text
			&& _pixmapItem != 0
			&& _pixmapItem->isVisible())
		{
			fitToWindow(_pixmapItem);
		}

        if (_fileOverlayActive) {
            if (_filenameOverlay != 0) _filenameOverlay->setVisible(true);
        }
    } else {
        ui->graphicsView->setFrameShape(QFrame::StyledPanel);
        ui->graphicsView->setFrameShadow(QFrame::Sunken);
        ui->verticalLayout->setMargin(9);
        ui->graphicsView->setStyleSheet("");
        ui->frame_controls->setVisible(true);
        ui->frame_impath->setVisible(true);
        ui->frame_raw->setVisible(true);

        //QTimer::singleShot(0, this, SLOT(_updateImageLabel()));
        if (_filenameOverlay != 0) _filenameOverlay->setVisible(false);

		showNormal();

		// resize preview thumbnail to fit full screen
		if (_currentDisplayMode == DisplayMode_Text
			&& _pixmapItem != 0
			&& _pixmapItem->isVisible())
		{
			fitToWindow(_pixmapItem);
		}
    }

    fullScreenAct->setChecked(_fullScreen);


    if (_currentDisplayMode == DisplayMode_Video) {
        setVideoControlsVisible(false);
    }
}

void ViewImageDialog::on_toolButton_fullscreen_clicked()
{
    _toggleFullscreen();
}

void ViewImageDialog::on_toolButton_rotLeft_clicked()
{
    _loader.setDesiredOrientation(_currentPosition, -90);
    QTimer::singleShot(0, this, SLOT(loadPosition()));
}

void ViewImageDialog::on_toolButton_rotRight_clicked()
{
    _loader.setDesiredOrientation(_currentPosition, 90);
    QTimer::singleShot(0, this, SLOT(loadPosition()));
}


void ViewImageDialog::on_toolButton_whatsthis_clicked()
{
    QWhatsThis::enterWhatsThisMode();
}

void ViewImageDialog::on_spinBox_number_valueChanged(int arg1)
{
    const int val = arg1 - 1;
    if (_currNumImages > 0 && val != _currentPosition) {
        if (_currentDisplayMode == DisplayMode_Image) {
            _lastTransformation = ui->graphicsView->transform();
            _lastSceneRect = ui->graphicsView->sceneRect();
        }
        loadPosition(val);
    }
}

void ViewImageDialog::_displayText(const ImageDataPtr image,
                                   const QString& text,
                                   const bool updateStatus,
                                   const bool hideVideo)
{
    if (_videoPlayer != 0) _videoPlayer->stop();
    if (_pixmapItem != 0) _pixmapItem->setVisible(false);
    if (hideVideo) {
        if (_videoItem != 0) _videoItem->setVisible(false);
        if (_videoControlsGroup != 0) _videoControlsGroup->setVisible(false);
    }
    if (_textItem != 0) {
        ui->graphicsView->setTransform(QTransform());
        bool found = false;
        if (image.get() != 0
                && _pixmapItem != 0
                && image->getThumbnail().get() != 0)
        {
            QMutexLocker l(&_protectPixmapItem);
            if(!(image->getThumbnail()->isNull())
                    && image->getThumbnail()->width() > 0
                    && image->getThumbnail()->height() > 0)
            {
                found  = true;
                _textItem->setVisible(false);
                /*if (image->getThumbnail()->width() > _textItem->boundingRect().width()) {
                _pixmapItem->setPixmap(QPixmap::fromImage(
                    image->getThumbnail()->scaled(QSize(_textItem->boundingRect().width(),
                                                         image->getThumbnail()->height()),
                                                  Qt::KeepAspectRatio)));
            } else {
                _pixmapItem->setPixmap(QPixmap::fromImage(*(image->getThumbnail())));
            }*/
                //QImage* test = image->getThumbnail().get();
               _pixmapItem->setPixmap(QPixmap::fromImage(*(image->getThumbnail())));
                _pixmapItem->setVisible(true);
                /*
            QPointF pos = _textItem->pos();
            pos.setX(pos.x() -
                     0.5* (_pixmapItem->boundingRect().width() -
                     _textItem->boundingRect().width()) );
            pos.setY(pos.y() + _textItem->boundingRect().height() + 10);
            _pixmapItem->setPos(pos);

            pos.setX(_textItem->pos().x() + _textItem->boundingRect().width() * 0.5);
            pos.setY(_textItem->pos().y()
                     + _textItem->boundingRect().height() * 0.5
                     + _pixmapItem->boundingRect().height() * 0.5);
            */

                if (ui->toolButton_zoomWindow->isChecked() ) {
                    fitToWindow(_pixmapItem);
                } else {
                    ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
                    ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
                }
            }
        }
        if (!found) {
            _textItem->setVisible(true);
            _textItem->setPlainText(text);
            ui->graphicsView->centerOn(_textItem);
            ui->graphicsView->setSceneRect(_textItem->sceneBoundingRect());
        }
        if (updateStatus) {
            _currentDisplayMode = DisplayMode_Text;
        }
    }
    ui->toolButton_zoomIn->setEnabled(false);
    ui->toolButton_zoomOut->setEnabled(false);
    ui->toolButton_zoomWindow->setEnabled(false);
    ui->toolButton_zoomReal->setEnabled(false);
    ui->toolButton_rotRight->setEnabled(false);
    ui->toolButton_rotLeft->setEnabled(false);
    ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
}

void ViewImageDialog::_displayImage()
{
    if (_videoPlayer != 0) _videoPlayer->stop();
    if (_videoItem != 0) _videoItem->setVisible(false);
    if (_textItem != 0) _textItem->setVisible(false);
    if (_videoControlsGroup != 0) _videoControlsGroup->setVisible(false);
    ui->toolButton_zoomIn->setEnabled(true);
    ui->toolButton_zoomOut->setEnabled(true);
    ui->toolButton_zoomWindow->setEnabled(true);
    ui->toolButton_zoomReal->setEnabled(true);
    ui->toolButton_rotRight->setEnabled(true);
    ui->toolButton_rotLeft->setEnabled(true);
    if (_pixmapItem != 0) {
        QMutexLocker l(&_protectPixmapItem);
        _pixmapItem->setVisible(true);
        if (ui->toolButton_zoomWindow->isChecked() ) {
            fitToWindow(_pixmapItem);
        } else {
            ui->graphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
            ui->graphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
        }
    }
    _currentDisplayMode = DisplayMode_Image;
}

void ViewImageDialog::_displayVideo()
{
	if (_videoPlayer != 0) {
		if (_videoPlayer->mediaStatus() == QMediaPlayer::LoadingMedia) {
			return;
		}
		if (_textItem != 0 && _pixmapItem != 0)
		{
			if (_pixmapItem->isVisible()
				&& (_videoPlayer->mediaStatus() == QMediaPlayer::LoadingMedia
					|| _videoPlayer->state() == QMediaPlayer::StoppedState
					) )
			{
				// do nothing, keep the preview until completely loaded
				// and playing
			}
			else {
				_pixmapItem->setVisible(false);
				_textItem->setVisible(false);
			}
		}
		ui->toolButton_zoomIn->setEnabled(false);
		ui->toolButton_zoomOut->setEnabled(false);
		ui->toolButton_zoomWindow->setEnabled(false);
		ui->toolButton_zoomReal->setEnabled(false);
		ui->toolButton_rotRight->setEnabled(false);
		ui->toolButton_rotLeft->setEnabled(false);
		if (_videoItem != 0) {
			_videoItem->setVisible(true);

			ui->graphicsView->centerOn(_videoItem);
			fitToWindow(_videoItem);
			_currentDisplayMode = DisplayMode_Video;
			if (_videoPlayer->state() != QMediaPlayer::PlayingState
				&& _videoPlayer->state() != QMediaPlayer::PausedState
				&& _videoPlayer->mediaStatus() == QMediaPlayer::LoadedMedia) {
				QTimer::singleShot(0, _videoPlayer, SLOT(play()));
			}
		}
	}
}

void ViewImageDialog::playerStateChanged(const QMediaPlayer::State& state)
{
    _videoPositionSlider->setEnabled(state == QMediaPlayer::PlayingState);
    _playButton->setEnabled(_videoPlayer->mediaStatus() != QMediaPlayer::InvalidMedia);

    switch(state) {
    case QMediaPlayer::PlayingState:
        _playButton->setIcon(style()->standardIcon(QStyle::SP_MediaPause));
		if (_pixmapItem != 0) {
			_pixmapItem->setVisible(false);
		}
		if (_textItem != 0) {
			_textItem->setVisible(false);
		}
        break;
    default:
        _playButton->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
        break;
    }
}

void ViewImageDialog::mediaStatusChanged(const QMediaPlayer::MediaStatus& status)
{
    if (status == QMediaPlayer::QMediaPlayer::EndOfMedia) {
        _displayText(ImageDataPtr(0), tr("video has finished playing, press play to start over"), false, true);
    }
	else if (status == QMediaPlayer::InvalidMedia
		|| status == QMediaPlayer::StalledMedia)
	{
		_displayText(ImageDataPtr(0), tr("video error"), false, true);
	}
	else if (status == QMediaPlayer::LoadedMedia
			 && _currentDisplayMode == DisplayMode_Video) {
		_displayVideo();
	}
}

void ViewImageDialog::positionChanged(const qint64 pos)
{
    const bool tmp = _videoPositionSlider->blockSignals(true);
	if (_currentDisplayMode == DisplayMode_Video) {
		_videoPositionSlider->setValue(pos);
		_videoPositionSlider->blockSignals(tmp);
	}
}

void ViewImageDialog::durationChanged(const qint64 duration)
{
    _videoPositionSlider->setRange(0,duration);
    _videoPositionSlider->setTickInterval(0);
    _videoPositionSlider->setPageStep(duration/10);
    _videoPositionSlider->setSingleStep(duration/50);
}

void ViewImageDialog::playPauseToggled()
{
    if (_videoPlayer->mediaStatus() == QMediaPlayer::EndOfMedia) {
        loadPosition(_currentPosition);
    } else {
        switch(_videoPlayer->state()) {
        case QMediaPlayer::PlayingState:
            _videoPlayer->pause();
            break;
        default:
            _videoPlayer->play();
            break;
        }
    }
}


void ViewImageDialog::muteToggled()
{
    _muteButton->setIcon(_muteButton->isChecked() ?
                             style()->standardIcon(QStyle::SP_MediaVolumeMuted) :
                             style()->standardIcon(QStyle::SP_MediaVolume));
}

void ViewImageDialog::handleError(const QMediaPlayer::Error err)
{
    qWarning() << "Mediaplayer error" << _videoPlayer->media().canonicalUrl() << err;
}

void ViewImageDialog::sliderPositionChange(const int pos)
{
    const bool tmp = _videoPlayer->blockSignals(true);
    _videoPlayer->setPosition(pos);
    _videoPlayer->blockSignals(tmp);
}

void ViewImageDialog::setVideoControlsVisible(const bool visible)
{
	if (_videoControlsGroup != 0) {
		_videoControlsGroup->setVisible(visible);
		_playButton->setVisible(visible);
		_muteButton->setVisible(visible);
		_videoPositionSlider->setVisible(visible);
		_videoVolumeSlider->setVisible(visible);
		_videoControlsGroup->repaint();	
	}
}

void ViewImageDialog::disappearControls()
{
    setVideoControlsVisible(false);
    _controlsDisappearTimer.stop();
    if (_currentDisplayMode == DisplayMode_Video) {
		QTimer::singleShot(0, this, SLOT(_displayVideo()));
    }
}

