/***********************************************************************
 * This file is part of module_misc.
 *
 * module_misc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_misc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_misc.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDir>
#include <QDebug>
#include <QCryptographicHash>
#include <QDateTime>
#include <QString>
#include <QVector>
#include <QLocale>
#include <QSettings>
#include <QDesktopWidget>

#include "misctools.h"
#include "imagedata.h"
#include <cassert>
#include <exception>

#include <stdio.h>
#include <libexif/exif-loader.h>
#include <libexif/exif-data.h>
#include <libraw/libraw.h>

extern "C" {
#include "jpeg-data.h"
}

#ifdef _WIN32
#include <qstring.h>
#endif

namespace MiscTools
{

#ifdef _WIN32
QString MiscTools::getDevicePath(const QString& deviceID,
								 const QString& number)
{
    QSettings settings(QString("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Class\\"
                               "%1\\%2\\DeviceData").arg(deviceID).arg(number),
						QSettings::NativeFormat);
    QString out = settings.value("WpdDevicePnPID", "").toString();
	return out;
}
#endif

bool MiscTools::_dialogLock = false;

bool MiscTools::lockDialog()
{
    if (_dialogLock) {
        return false;
    }
    _dialogLock = true;
    return true;
}

void MiscTools::unlockDialog()
{
    _dialogLock = false;
}

QRect MiscTools::retrieveDesktopResolution()
{
    QDesktopWidget d;

    // multi-screen support
    QRect s = d.screenGeometry(d.screenNumber());
    return s;
}

QString MiscTools::appDataPath()
{
#ifdef _WIN32
    QString out = QDir::homePath();
    QDir search(out);

    if (search.exists(QDir::toNativeSeparators("AppData/Roaming"))) {
        out += "/Appdata/Roaming";
    } else if (search.exists(QDir::toNativeSeparators("AppData/Local"))) {
        out += "/Appdata/Local";
    } else if (search.exists(QDir::toNativeSeparators("Application Data"))) {
        out += "/Application Data";
    }
    //return QDir::toNativeSeparators(out + QDir::toNativeSeparators("/Application Data/"));
    return out;
#else
    return QDir::toNativeSeparators(QDir::homePath());
#endif
}

pQImage MiscTools::LibRawImageToQImage(const uchar *data,
                                       const int width,
                                       const int height,
                                       const int nCols,
                                       const int colorBits)
{
    int colorSize = (colorBits % 8) == 0 ? colorBits / 8 : ceil(colorBits / 8.0);
    pQImage outPt;
    int numPixels = width * height;
    int pixelSize =  nCols * colorSize;
    uchar* pixels = new uchar[numPixels * 3];
    for (int i = 0; i < numPixels; i++, data += pixelSize) {
        if (nCols == 3) {
            // this ordering produces good results - why?!
            pixels[i * 3] = data[3*colorSize];
            pixels[i * 3 + 1] = data[colorSize];
            pixels[i * 3 + 2] = data[2*colorSize];
        } else {
            pixels[i * 3] = data[0];
            pixels[i * 3 + 1] = data[0];
            pixels[i * 3 + 2] = data[0];
        }
    }
    // immediately create a copy since otherwise we'd have to
    // 'delete[] pixels' somewhere else, ourselves
    // see http://doc.qt.io/qt-5.5/qimage.html#QImage-6
    *outPt = QImage(pixels, width, height, width * 3,
                           QImage::Format_RGB888).copy();
    delete[] pixels;
    return outPt;
}

void MiscTools::LibRawImagePerformFlip(const int flip,
                                       QImage& image)
{
    if (flip != 0) {
        QTransform rotation;
        int angle = 0;
        if (flip == 3) angle = 180;
        else if (flip == 5) angle = -90;
        else if (flip == 6) angle = 90;
        if (angle != 0) {
            rotation.rotate(angle);
            image = image.transformed(rotation);
        }
    }
}

pQImage MiscTools::loadImageOrRaw(const QString& filename,
                                  const bool fast,
                                  const bool* abort)
{
    pQImage out;

    *out = QImage(filename);

    if (!out->isNull() && out->width() >0 && out->height() > 0) {

        return out;
    }

    // try reading as RAW file
    LibRaw RawProcessor;
    RawProcessor.imgdata.params.gamm[0] = 1.0;
    RawProcessor.imgdata.params.gamm[1] = 0.0;
    if (fast)
    {
        RawProcessor.imgdata.params.user_qual = 0; // fastest interpolation (linear)
        RawProcessor.imgdata.params.bright = 1.5;
        RawProcessor.imgdata.params.use_camera_wb = 1;
        RawProcessor.imgdata.params.use_auto_wb = 0;
        RawProcessor.imgdata.params.med_passes = 0;
        RawProcessor.imgdata.params.dcb_enhance_fl=0;
        RawProcessor.imgdata.params.fbdd_noiserd = 0;
        RawProcessor.imgdata.params.es_med_passes = 0;
        RawProcessor.imgdata.params.use_rawspeed = 1;
    }
    if(  LIBRAW_SUCCESS == RawProcessor.open_file(filename.toLocal8Bit().data()))
    {
        if (abort != 0 && *abort) {
            RawProcessor.recycle();
            return out;
        }
        if( LIBRAW_SUCCESS == RawProcessor.unpack() )
        {
            if (abort != 0 && *abort) {
                RawProcessor.recycle();
                return out;
            }
            if (LIBRAW_SUCCESS == RawProcessor.dcraw_process())
            {
                if (abort != 0 && *abort) {
                    RawProcessor.recycle();
                    return out;
                }
                libraw_processed_image_t* output = RawProcessor.dcraw_make_mem_image();
                if (abort != 0 && *abort) {
                    LibRaw::dcraw_clear_mem(output);
                    RawProcessor.recycle();
                    return out;
                }

                if (LIBRAW_IMAGE_JPEG == output->type )
                {
                    out->loadFromData((uchar*)output->data,
                                           output->data_size,
                                           "JPEG");
                    MiscTools::MiscTools::LibRawImagePerformFlip(
                                RawProcessor.imgdata.sizes.flip, *out);
                } else if (LIBRAW_IMAGE_BITMAP == output->type) {
                    *out = *MiscTools::MiscTools::LibRawImageToQImage(
                                (uchar*)output->data,
                                output->width,
                                output->height,
                                output->colors,
                                output->bits);
                }
                LibRaw::dcraw_clear_mem(output);
            }
            RawProcessor.recycle();
        }
    }

    return out;
}

QByteArray MiscTools::getMD5Sum(const QString & path)
{
    QFile file(path);

    if (file.open(QIODevice::ReadOnly | QIODevice::Unbuffered)) {
        QCryptographicHash md5Hash(QCryptographicHash::Md5);
        md5Hash.addData(file.readAll());
        file.close();
        return md5Hash.result();
    } else {
        return QByteArray();
    }
}


bool MiscTools::changeTimeStamp(const QString& filenameIn,
                                const QString& filenameOut,
                                const QDateTime& newDate,
                                const bool saveAsTextIfNotWOrks)
{
    // only JPEG
    QString ext = QFileInfo(filenameIn).completeSuffix().toLower();

    if (ext == "jpg" || ext == "jpeg") {
        ExifLoader *l;
        /* Create an ExifLoader object to manage the EXIF loading process */
        l = exif_loader_new();

        bool found = false;
        if (l) {
            ExifData *ed;

            /* Load the EXIF data from the image file */
            exif_loader_write_file(l, filenameIn.toLocal8Bit().data());

            /* Get a pointer to the EXIF data */
            ed = exif_loader_get_data(l);

            /* The loader is no longer needed--free it */
            exif_loader_unref(l);
            l = NULL;

            // the source file didn't have exif tags
            if (!ed) {
                ed = exif_data_new();
            }

            // next modify target time stamps exif data

            QVector<QPair<ExifTag, ExifIfd> > tags(QVector<QPair<ExifTag, ExifIfd> >()
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_ORIGINAL, EXIF_IFD_EXIF)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_DIGITIZED, EXIF_IFD_EXIF)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME, EXIF_IFD_0));
            for (int i=0; i<(int)tags.size(); i++) {
                /* See if this tag exists */
                ExifEntry *entry =
                    exif_data_get_entry ( ed,
                                          tags[i].first);
                if (!entry) {
                    entry = exif_entry_new ();
                    exif_content_add_entry (ed->ifd[tags[i].second], entry);
                    exif_entry_initialize (entry, tags[i].first);
                    if (!entry) {
                        qCritical() << "not enough memory";
                        throw "Not enough memory.";
                    }
                }

                // if (entry->data) free (entry->data); // free old data
                entry->components = 20;
                entry->size = sizeof (char) * entry->components;
                entry->data = (unsigned char*)malloc (entry->size + 1);
                if (entry->data == 0) {
                    qCritical() << "not enough memory";
                    throw "Not enough memory.";
                }

#ifdef _MSC_VER
                const QByteArray dateStr = newDate.toString(TOOLS_TIME_FORMAT).toLocal8Bit();
                strcpy_s ((char *) entry->data, dateStr.size()+1, dateStr.data());
#else
                strcpy ((char *) entry->data, newDate.toString(TOOLS_TIME_FORMAT).toLocal8Bit().data());
#endif
                entry->data[19] = 0;
                found = true;
            }

            /* Parse the JPEG file from source and save it in target */
            JPEGData* jdata;
            jdata = jpeg_data_new ();
            jpeg_data_load_file (jdata, filenameIn.toLocal8Bit().data());

            jpeg_data_set_exif_data (jdata, ed);

            /* Save the modified image. */
            jpeg_data_save_file (jdata, filenameOut.toLocal8Bit().data());
            jpeg_data_unref (jdata);

            exif_data_unref (ed);
        }

        if (found) {
            return true;
        }
    }

    // saving has not worked, so write text file, if requested
    if (saveAsTextIfNotWOrks) {
        if (!_rawFormats.contains( QFileInfo(filenameOut).suffix().toUpper()))
        {
            const QString textFilenameOut = filenameOut + ".txt";
            QFile file(textFilenameOut);
            file.open(QIODevice::WriteOnly | QIODevice::Text);
            QTextStream out(&file);
            out << "acquisition date: " << newDate.toString(TOOLS_TIME_FORMAT) << "\n";
            file.close();
            return true;
        }
    }
    return false;
}

bool MiscTools::getTimeStampOrientationEXIF(const QString& filenameIn,
                                            QDateTime& dateTime,
                                            OrientationJPG& orientation)
{
    // only JPEG

    if (true) { //ext == "jpg" || ext == "jpeg") {
        ExifLoader *l;
        /* Create an ExifLoader object to manage the EXIF loading process */
        l = exif_loader_new();

        if (l) {
            ExifData *ed;

            /* Load the EXIF data from the image file */
            exif_loader_write_file(l, filenameIn.toLocal8Bit().data());

            /* Get a pointer to the EXIF data */
            ed = exif_loader_get_data(l);

            /* The loader is no longer needed--free it */
            exif_loader_unref(l);
            l = NULL;

            // the source file didn't have exif tags
            if (!ed) {
                exif_loader_unref(l);
                return false;
            }

            // next modify target time stamps exif data


            ExifEntry *entryOrientation =
                exif_data_get_entry ( ed, EXIF_TAG_ORIENTATION);
            if (entryOrientation != 0 && entryOrientation->data)
            {
               // qDebug() <<"orientation" << entryOrientation->data;
                orientation = (OrientationJPG)entryOrientation->data[0];
            }

            QVector<QPair<ExifTag, ExifIfd> > tags(QVector<QPair<ExifTag, ExifIfd> >()
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_ORIGINAL, EXIF_IFD_EXIF)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME, EXIF_IFD_0)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_DIGITIZED, EXIF_IFD_EXIF));
            for (int i=0; i<(int)tags.size(); i++) {
                /* See if this tag exists */
                ExifEntry *entry =
                    exif_data_get_entry ( ed,
                                          tags[i].first);

                if (entry != 0 && entry->data)
                {
                    int j=0;
                    while (entry->data[j] != 0 && j < 22) j++;
                    QString dateStr = QString::fromLocal8Bit((char*)entry->data, j);
                    QString format = "yyyy:MM:dd HH:mm:ss";
                    if (dateStr.length() == format.length())
                    {
                        dateTime = QDateTime::fromString(dateStr, format);
                        exif_data_unref (ed);
                        return true;
                    }
                }
            }

            exif_data_unref (ed);
        }
    }
    return false;
}

bool MiscTools::getTimeStampAndOrientation(const QString& filename,
                                           QDateTime& timeOut,
                                           OrientationJPG& orientationOut)
{
    orientationOut = OrientationJPG_normal;
    timeOut = QDateTime();

    getTimeStampOrientationEXIF(filename, timeOut, orientationOut);

    if (_rawFormats.contains(QFileInfo(filename).suffix().toUpper()))
    {
        LibRaw RawProcessor;
        if(  LIBRAW_SUCCESS == RawProcessor.open_file(filename.toLocal8Bit().data()))
        {
            timeOut = QDateTime::fromTime_t(RawProcessor.imgdata.other.timestamp);
        }
        RawProcessor.recycle();
    }
    return timeOut.isValid();
}




}
// end namespace misctools
