/***********************************************************************
 * This file is part of module_randomimageframe.
 *
 * module_randomimageframe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_randomimageframe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_randomimageframe.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QTimer>
#include <QHBoxLayout>
#include <qmath.h>
#include <QDebug>

#include "randomimagelist.h"


RandomImageList::RandomImageList(const qint32 thumbsize,
                                 const qint32 timeToChange,
                                 const QList<ImageChange>& modes,
                                 QWidget *parent)
    : QFrame(parent)
    , _thumbsize(thumbsize)
    , _timeToChange(timeToChange)
    , _changeModes(modes)
    , _lastImageIndex(-1)
    , _lastViewIndex(-1)
    , _started(false)
{
    QHBoxLayout* layout = new QHBoxLayout();
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    this->setLayout(layout);

    if (_changeModes.size() == 0) {
        _changeModes.push_back(ImageChange(QEasingCurve::InOutElastic, 1000, 2.0));
        _changeModes.push_back(ImageChange(QEasingCurve::InOutBounce, 1000, 2.0));
        _changeModes.push_back(ImageChange(QEasingCurve::InOutBack, 750, 3.0));
        _changeModes.push_back(ImageChange(QEasingCurve::InOutQuad, 500, 2.0));
    }
}

RandomImageList::~RandomImageList()
{
    _timer.stop();
    clearImages();
}

QSize RandomImageList::minimumSizeHint () const
{
    const int realSize = _thumbsize + ADD_BORDER_TO_IMAGE;
    QSize out(realSize, realSize);
    return out;
}

QSize RandomImageList::sizeHint () const
{
    return minimumSizeHint();
}

void RandomImageList::setTimeToChange(const qint32 timeToChange)
{
    _timeToChange = timeToChange;
}

void RandomImageList::setThumbnailsize(const qint32 thumbsize)
{
    _thumbsize = thumbsize;
}

void RandomImageList::setChangeModes(const QList<ImageChange>& modes)
{
    _changeModes = modes;
}

void RandomImageList::clearImages()
{
    _timer.stop();
    disconnect(&_timer, SIGNAL(timeout()), this, SLOT(runAnimation()));
    _images.clear();
    _currInd.clear();
    for (int i=0; i<_imageWidgets.size(); i++) {
        MiscTools::ImageFlip* tmp = _imageWidgets[i];
        layout()->removeWidget(tmp);
        delete tmp;
        tmp = 0;
    }
    _imageWidgets.clear();
}

void RandomImageList::addImage(const ImageDataPtr &img)
{
    _images.push_back(img);
    addWidget();
}

void RandomImageList::startAnimation()
{
    _started = true;
    _timer.stop();
    _timer.setInterval(0);
    _timer.setSingleShot(true);
    connect(&_timer, SIGNAL(timeout()), SLOT(runAnimation()));
    _timer.start();

}

void RandomImageList::stopAnimation()
{
    _timer.stop();
    disconnect(&_timer, SIGNAL(timeout()), this, SLOT(runAnimation()));
    _started = false;
}

void RandomImageList::runAnimation()
{
    if (!_started) {
        return;
    }

    _timer.stop();
    _timer.setInterval(250+qrand() % _timeToChange);
    _timer.setSingleShot(true);
    _timer.start();
    if (_imageWidgets.size() == 0
            || _images.size() == 0) {
        return;
    }

    // find next image widget
    const int n = _imageWidgets.size();
    int nextInd = -1;
    if (n > 0) {
        for (int i=0; i<10; i++) {
            nextInd = qrand() % n;
            if (nextInd != _lastViewIndex) {
                break;
            }
        }
        if (nextInd == _lastViewIndex && n > 1) {
            nextInd = -1;
        }
    }
    if (nextInd < 0) {
        return;
    }

    // don't use too many preview images
    if (nextInd > _images.size()*2) {
        return;
    }

    _lastViewIndex = nextInd;

    // find next image
    const int nI = _images.size();
    nextInd = -1;
    if (nI > n) {
        for (int i=0; i<10; i++) {
            nextInd = qrand() % nI;
            if (nextInd != _lastImageIndex) {
                break;
            }
        }
        if (nextInd == _lastImageIndex) {
            nextInd = -1;
        }
    }
    if (nextInd < 0) {
        return;
    }
    _lastImageIndex = nextInd;

    // find animation
    const int nAn = qrand() % _changeModes.size();

    // update widget
    _imageWidgets[_lastViewIndex]->setAnimationMode(_changeModes[nAn].type,
                                                    _changeModes[nAn].amplitude);
    _imageWidgets[_lastViewIndex]->setTime(_changeModes[nAn].time);
    _imageWidgets[_lastViewIndex]->setNewImage(_images[_lastImageIndex]->getThumbnail());
    _currInd[_lastViewIndex] = _lastImageIndex;
}

void RandomImageList::addWidget()
{
    const int nPossible = getNumPossible();
    const int nGiven = _imageWidgets.size();
    if (nGiven >= nPossible) {
        return;
    }
    // first not found
    for (int i=0; i<nPossible - nGiven; i++) {
        int fnd = -1;
        const int nIm = _images.size();
        for (int j=0; j<nIm; j += 2) {
            if (!_currInd.contains(j)) {
                fnd = j;
                break;
            }
        }
        if (fnd >= 0 ) {
            const int nAn = qrand() % _changeModes.size();
            MiscTools::ImageFlip * imflip = new MiscTools::ImageFlip();
            layout()->addWidget(imflip);
            _imageWidgets.push_back(imflip);
            _currInd.push_back(fnd);
            const int newInd = _imageWidgets.size() - 1;
            _imageWidgets[newInd]->setAnimationMode(_changeModes[nAn].type,
                                                    _changeModes[nAn].amplitude);
            _imageWidgets[newInd]->setTime(_changeModes[nAn].time);
            _imageWidgets[newInd]->setNewImage(_images[fnd]->getThumbnail());
        }
    }
}

void RandomImageList::delWidget()
{
    if (_imageWidgets.size() > 0) {
        MiscTools::ImageFlip* first = _imageWidgets[0];
        layout()->removeWidget(first);
        _imageWidgets.removeFirst();
        _currInd.remove(0);
        delete first;
        first = 0;
    }
}

qint32 RandomImageList::getNumPossible() const
{
    QRect rect = geometry();
    return qFloor(float(rect.width()) / float(_thumbsize + ADD_BORDER_TO_IMAGE));
}

void RandomImageList::resizeEvent(QResizeEvent *)
{
    const int nGiven = _imageWidgets.size();
    const int nPossible = getNumPossible();
    const int nIm = _images.size();

    if (nGiven < nPossible) {
        // add new widgets
        if (nIm >= nPossible) {
            for (int i=0; i < (nPossible - nGiven); i++) {
                addWidget();
            }
        }
    } else if (nGiven > nPossible) {
        // remove some widgets
        for (int i=0; i < (nGiven - nPossible); i++) {
            delWidget();
        }
    }
}
