/***********************************************************************
 * This file is part of module_thumbnailcache.
 *
 * module_thumbnailcache is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_thumbnailcache is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_thumbnailcache.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDebug>
#include <QDir>
#include <QSqlQuery>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDateTime>
#include <QBuffer>
#include <QSqlError>
#include <QSvgRenderer>
#include <QPainter>
#include <QApplication>
#include <QThread>
#include <QSqlDriver>
#include <QSqlField>
#include <stdio.h>

#include "fastimagethumbnail.h"
#include "qexifimageheader-1.h"
#include "misctools.h"
#include "imagedata.h"

#include "videothumbnails.h"

// initialization of member variables
QMutex FastImageThumbnail::databaseMutex;
QMutex FastImageThumbnail::_ioMutex;

// #define // DEBUG if (false)
FastImageThumbnail::FastImageThumbnail(const QString& location, QObject* parent )
    : QObject(parent)
    , _db(NULL)
    , _movieOverlayGraphic(":/thumbnail/moviewframe.png")
    , _videoThumbnailLoader(VideoThumbnailTools::instance())
{
    _connectToLocalDB(location);

    //*_movieOverlayGraphic = _movieOverlayGraphic->scaled(QSize(256,256), Qt::KeepAspectRatio, Qt::SmoothTransformation).convertToFormat(QImage::Format_RGB32, Qt::ImageConversionFlag::DiffuseDither | Qt::ImageConversionFlag::ThresholdAlphaDither);

    connect(_videoThumbnailLoader, SIGNAL(newThumbnail(QString,QImage)),
            this, SLOT(newVideoThumbnail(QString,QImage)));
    connect(_videoThumbnailLoader, SIGNAL(errorThumbnail(QString)),
            this, SLOT(errorVideoThumbnail(QString)));
}

void FastImageThumbnail::newVideoThumbnail(const QString& videoFilename,
                                           const QImage& image)
{
    const QFileInfo info(videoFilename);

    if (!image.isNull() && image.width() > 0 && image.height() > 0)
    {
        static int Gerd = 0;
        Gerd++;
        MiscTools::pQImage pImage(image.copy());
        resizeToThumbSize(*pImage);

        if (!_movieOverlayGraphic.isNull()) {
            QPainter painter(pImage.get());
            painter.drawImage(0,0, _movieOverlayGraphic.scaled(
                              pImage->size(),
                              Qt::IgnoreAspectRatio,
                              Qt::SmoothTransformation));
            painter.end();
        }

        _updateCache(info, pImage);

        emit newThumbnail(info, pImage);
    } else {
        emit errorThumb(info);
    }
}

void FastImageThumbnail::errorVideoThumbnail(const QString& videoFilename)
{
    emit errorThumb(QFileInfo(videoFilename));
}

FastImageThumbnail::~FastImageThumbnail()
{
    _disconnect();
    delete _db;
    _db = 0;
    RawProcessor.recycle();
    //delete _player;
    //delete _player2;
    //delete _videoWidget;
}

FastImageThumbnail::Thumbnail_Type FastImageThumbnail::getThumbnail(const QFileInfo& fileinfo,
                                                                    MiscTools::pQImage &thumbnailOut,
                                                                    const bool onlyFast)
{
    thumbnailOut = QImage();

    const QString completePath = fileinfo.canonicalFilePath();
    QImage thumbnail;

    // 0. try to obtain the thumbnail from the metadata (
    // try exif data
    QExifImageHeader exif(completePath);
    thumbnail = exif.thumbnail().copy();
    if (!thumbnail.isNull()
            && thumbnail.width() > 0
            && thumbnail.height() > 0) {
        orientThumbnail(completePath, thumbnail);
        _removeBlackBorder(thumbnail);
        thumbnailOut = thumbnail;
        return THUMB_TYPE_IMAGE;
    }

    // 1. try to access local pivot thumbnail database storage
    if (_checkLocalCache(fileinfo, thumbnailOut)) {
        // DEBUG qDebug() << "loading thumbnail from local image cache";
        return THUMB_TYPE_IMAGE;
    }

    // 2. try to obtain the thumbnail from the raw metadata
    thumbnail = _tryLoadThumbnail(completePath);
    if (!thumbnail.isNull() && thumbnail.width() > 0 && thumbnail.height() > 0) {
        if (resizeToThumbSize(thumbnail)) {
            // save if the size had to be reduced -> faster loading next time
            _removeBlackBorder(thumbnail);
            *thumbnailOut = thumbnail;
            _updateCache(fileinfo, thumbnailOut);
        } else {
            *thumbnailOut = thumbnail;
        }
        return THUMB_TYPE_IMAGE;
    }

    if (onlyFast) {
        return THUMB_TYPE_NO_FAST_THUMB;
    }

    // otherwise (re-)compute
    thumbnail = QImage(completePath);
	
    if (!thumbnail.isNull()
            && thumbnail.width() > 0
            && thumbnail.height() > 0)
    {
        orientThumbnail(completePath, thumbnail);
    }
    else
    {
        // try reading as RAW file
        RawProcessor.imgdata.params.gamm[0] = 1.0;
        RawProcessor.imgdata.params.gamm[1] = 0.0;
        RawProcessor.imgdata.params.user_qual = 0; // fastest interpolation (linear)
        RawProcessor.imgdata.params.use_camera_wb = 1;
        if(  LIBRAW_SUCCESS == RawProcessor.open_file(completePath.toLocal8Bit().data()))
        {
            if( LIBRAW_SUCCESS == RawProcessor.unpack() )
            {
                if (LIBRAW_SUCCESS == RawProcessor.dcraw_process())
                {
                    libraw_processed_image_t* output = RawProcessor.dcraw_make_mem_image();

                    if (LIBRAW_IMAGE_JPEG == output->type )
                    {
                        thumbnail.loadFromData((uchar*)output->data,
                                               output->data_size,
                                               "JPEG");
                        MiscTools::MiscTools::LibRawImagePerformFlip(
                                    RawProcessor.imgdata.sizes.flip, thumbnail);
                    } else if (LIBRAW_IMAGE_BITMAP == output->type) {
                        thumbnail = *MiscTools::MiscTools::LibRawImageToQImage(
                                    (uchar*)output->data,
                                    output->width,
                                    output->height,
                                    output->colors,
                                    output->bits);
                    }
                    LibRaw::dcraw_clear_mem(output);
                }
                RawProcessor.recycle();
            }
        }
    }

    if ((thumbnail.isNull()
            || thumbnail.width() == 0
            || thumbnail.height() == 0)
            && _videoFormats.contains(fileinfo.suffix().toLower()) )
    {
        _videoThumbnailLoader->extractFrame(completePath);
        return THUMB_TYPE_TEST_VIDEO;
    } else {
        resizeToThumbSize(thumbnail);
        orientThumbnail(completePath, thumbnail);
        // update cache
        *thumbnailOut = thumbnail;
        if (!thumbnail.isNull() && thumbnail.width() > 0 && thumbnail.height() > 0) {
            _updateCache(fileinfo, thumbnailOut);
            return THUMB_TYPE_IMAGE;
        } else {
            return THUMB_TYPE_FAILED_LOADING;
        }
    }
}

bool FastImageThumbnail::_connectToLocalDB(const QString& filename)
{
    QMutexLocker l(&databaseMutex);
    QString locationPath = QDir::toNativeSeparators(filename);
    // DEBUG qDebug() << locationPath;

    QSqlDatabase::removeDatabase("qt_sql_default_connection");
    delete _db;
    _db = new QSqlDatabase;
    *_db = QSqlDatabase::addDatabase("QSQLITE");
    _db->setDatabaseName(locationPath);
    if (!_db->open()) {
        // maybe the cache is corrupted
        QDir tmp(locationPath.left(locationPath.lastIndexOf(QDir::separator())));
        if (tmp.exists(locationPath.right(-1+locationPath.length() - locationPath.lastIndexOf(QDir::separator())))) {
            // DEBUG qDebug() << "could not connect to db - now deleting file and retrying";
            QFile::remove(locationPath);
        }
        if (!_db->open()) {
            return false;
        }
    }

    {
        QSqlQuery query(*_db);
        QString queryStr = QString("CREATE TABLE IF NOT EXISTS %1"
                " ( %2 varchar(32768), %3 INT, %4 BLOB);").
                arg(TABLE_IMAGE).arg(FIELD_FILE_NAME).arg(FIELD_FILE_TIME).
                arg(FIELD_FILE_IMAGE);
        if (!query.exec(queryStr))
        {
            // DEBUG qDebug() << "could not insert table";
        } else {
            // DEBUG qDebug() << "successful insertion of table";
        }
    }

    // DEBUG qDebug() << _db->lastError().text();

    // todo: hidden/system file attribute of thumbnails database for windows
    return true;
}


void FastImageThumbnail::_disconnect()
{
    //db.removeDatabase();
    QMutexLocker l(&databaseMutex);
    _db->close();
    delete _db;
    _db = NULL;
    QSqlDatabase::removeDatabase("QSQLITE");
    QSqlDatabase::removeDatabase("qt_sql_default_connection");
}

bool FastImageThumbnail::_checkLocalCache(const QFileInfo& fileinfo,
                                          MiscTools::pQImage& out)
{
    QMutexLocker l(&databaseMutex);

    bool status = false;

    QSqlTableModel model(0, *_db);
    model.setTable(TABLE_IMAGE);
    QString filter =  QString("%1='%2'").arg(FIELD_FILE_NAME).arg(fileinfo.canonicalFilePath());
    model.setFilter(filter);
    model.select();
    if (model.rowCount() > 0) { //model() > 0) {
        QSqlRecord record = model.record(0);
        const QVariant v = record.value(1);
        const uint dbtimeI = v.toUInt();
        //QDateTime dbTime(QDateTime::fromTime_t(dbtimeI) );
        QDateTime other = fileinfo.lastModified();
        // last recorded modification time and real modification coincide ?
        if (dbtimeI == other.toTime_t()) {
            QByteArray imgData = record.value(2).toByteArray();
            if (out->loadFromData(imgData)) {
                status = true;
            }
        } else {
            // DEBUG qDebug() << "file outdated, need to recompute thumbnail";
            // otherwise delete entries
            model.setFilter(QString("%1='%2.'").arg(FIELD_FILE_NAME).arg(fileinfo.canonicalFilePath()));
            model.select();
            if (model.rowCount() >= 1) {
                model.removeRows(0, model.rowCount());
                model.submitAll();
            }
        }
    } else {
        // DEBUG qDebug() << fileinfo.canonicalFilePath()<< " not found in DB";
    }

    return status;
}

bool FastImageThumbnail::_updateCache(const QFileInfo& fileinfo,
                                      const MiscTools::pQImage& thumbnail)
{
    QByteArray ba;
    QBuffer buffer(&ba);
    buffer.open(QIODevice::WriteOnly);
    thumbnail->save(&buffer, "png", 85);

    QMutexLocker l(&databaseMutex);

    // DEBUG qDebug() << "inserting";
    QSqlTableModel model(0, *_db);
    model.setTable(TABLE_IMAGE);
    int row=0;
    model.insertRows(row, 1);
    // DEBUG qDebug() << fileinfo.canonicalFilePath();
    model.setData(model.index(row, 0), fileinfo.canonicalFilePath());
    model.setData(model.index(row, 1), fileinfo.lastModified().toTime_t());
    model.setData(model.index(row, 2), QVariant(ba));
    if (!model.submitAll()) {
        // DEBUG qDebug() << "error submit";
        return false;
    }
    return true;
}

void FastImageThumbnail::orientThumbnail(const QString& fn,
                                         QImage& thumbnail)
{
    QDateTime dt;
    MiscTools::OrientationJPG orientation;
    MiscTools::MiscTools::getTimeStampAndOrientation(fn, dt, orientation);

    switch (orientation) {
    case MiscTools::OrientationJPG_Rot180:
        thumbnail = thumbnail.transformed(QTransform().rotate(180));
        break;
    case MiscTools::OrientationJPG_rotLeft:
        thumbnail = thumbnail.transformed(QTransform().rotate(90));
        break;
    case MiscTools::OrientationJPG_rotRight:
        thumbnail = thumbnail.transformed(QTransform().rotate(-90));
        break;
    case MiscTools::OrientationJPG_flippedH:
    case MiscTools::OrientationJPG_flippedV:
    case MiscTools::OrientationJPG_flippedHRotLeft:
    case MiscTools::OrientationJPG_flippedHRotRight:
        // flipping of qpixmap seems not as trivial
        // to perform efficiently
        // (e.g. conversion to QImage etc)
        qWarning() << "orientation tags with flipped images"
                 << " not supported yet";
    case MiscTools::OrientationJPG_normal:
    default:
        break;
    }
}

void FastImageThumbnail::_removeBlackBorder(QImage &image)
{
    static int b = 0;
    image.save(QString("%1.png").arg(b++));
    const QRgb black = qRgba(0,0,0,255);
	const int blackMax = 100;
    // UPPER
    int maxU = 0;
    for (; maxU < image.height(); maxU++) {
        bool foundNonBlack = false;
        for (int i=0; i< image.width(); i++) {
			QRgb pixel = image.pixel(i, maxU);
			const int m = qMax(qRed(pixel), qMax(qGreen(pixel), qBlue(pixel)));
            if (m > blackMax) {
                foundNonBlack = true;
                break;
            }
        }
        if (foundNonBlack) {
            break;
        }
    }
    if (maxU == image.height()) {
        // all black image -> return
        return;
    }

    // LOWER
    int minL = image.height()-1;
    for (; minL>=maxU; minL--) {
        bool foundNonBlack = false;
        for (int i=0; i< image.width(); i++) {
			QRgb pixel = image.pixel(i, minL);
			const int m = qMax(qRed(pixel), qMax(qGreen(pixel), qBlue(pixel)));
			if (m > blackMax) {
                foundNonBlack = true;
                break;
            }
        }
        if (foundNonBlack) {
            break;
        }
    }
    // crop, but only if both sides are roughly the same number of lines
	if (abs(maxU - image.height() + minL) < 3 // same bar width
		&& abs(maxU - minL) > 10 // not both close to center
		)
	{
		image = image.copy(0, 0, image.width(), minL + 1);
		image = image.copy(0, maxU, image.width(), image.height() - maxU);
	}

    // LEFT
    int maxL = 0;
    for (; maxL<image.width(); maxL++) {
        bool foundNonBlack = false;
        for (int i=0; i< image.height(); i++) {
			QRgb pixel = image.pixel(maxL, i);
			const int m = qMax(qRed(pixel), qMax(qGreen(pixel), qBlue(pixel)));
			if (m > blackMax) {
                foundNonBlack = true;
                break;
            }
        }
        if (foundNonBlack) {
            break;
        }
    }
    // crop

    // RIGHT
    int minR = image.width()-1;
    for (; minR>=maxL; minR--) {
        bool foundNonBlack = false;
        for (int i=0; i< image.height(); i++) {
			QRgb pixel = image.pixel(minR, i);
			const int m = qMax(qRed(pixel), qMax(qGreen(pixel), qBlue(pixel)));
			if (m > blackMax) {
                foundNonBlack = true;
                break;
            }
        }
        if (foundNonBlack) {
            break;
        }
    }

	// crop, but only if both sides are roughly the same number of lines
	if (abs(maxL - image.width() + minR) < 3 // same bar width
		&& abs(maxL - minR) > 10 // not both close to center
		&& maxL != 0)
	{
		image = image.copy(0, 0, minR+1, image.height());
		image = image.copy(maxL, 0, image.width() - maxL, image.height());
	}
}

QImage  FastImageThumbnail::_tryLoadThumbnail(const QString& lpszPathName )
{
    QMutexLocker l(&_ioMutex);
    QImage thumbnail;

    // try reading as RAW file (with possibly embedded thumbnail)
    if(  LIBRAW_SUCCESS == RawProcessor.open_file(lpszPathName.toLocal8Bit().data()))
    {
        if( LIBRAW_SUCCESS == RawProcessor.unpack_thumb() )
        {
            if (LIBRAW_THUMBNAIL_JPEG == RawProcessor.imgdata.thumbnail.tformat )
            {
                thumbnail.loadFromData((uchar*)RawProcessor.imgdata.thumbnail.thumb,
                                 RawProcessor.imgdata.thumbnail.tlength,
                                 "JPEG");
                MiscTools::MiscTools::LibRawImagePerformFlip(
                            RawProcessor.imgdata.sizes.flip, thumbnail);
            } else if (LIBRAW_THUMBNAIL_BITMAP == RawProcessor.imgdata.thumbnail.tformat) {
                thumbnail = *MiscTools::MiscTools::LibRawImageToQImage(
                            (uchar*)RawProcessor.imgdata.thumbnail.thumb,
                            RawProcessor.imgdata.thumbnail.twidth,
                            RawProcessor.imgdata.thumbnail.theight,
                            RawProcessor.imgdata.thumbnail.tcolors);
            }
        }
        RawProcessor.recycle();
    }

    return thumbnail;
}

bool FastImageThumbnail::resizeToThumbSize(QImage& thumbnail)
{
    if (thumbnail.width() > THUMBNAIL_MAX_SIZE
            || thumbnail.height() > THUMBNAIL_MAX_SIZE)
    {
        thumbnail = thumbnail.scaled(QSize(THUMBNAIL_MAX_SIZE, THUMBNAIL_MAX_SIZE),
                                     Qt::KeepAspectRatio,
                                     Qt::FastTransformation );
        return true;
    }
    else {
        return false;
    }
}
