/***********************************************************************
 * This file is part of module_thumbnailcache.
 *
 * module_thumbnailcache is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_thumbnailcache is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_thumbnailcache.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef FASTIMAGETHUMBNAIL_H
#define FASTIMAGETHUMBNAIL_H

#include <QObject>
#include <QFileInfo>
#include <QMutex>
#include <QSqlDatabase>
#include <QTimer>
#include <libraw/libraw.h>

#define THUMBNAIL_MAX_SIZE 256
#define THUMBNAIL_DATAFILE ".mythumbs.sqlite.db"

#define FIELD_FILE_TIME "filetime"
#define FIELD_FILE_IMAGE "image"
#define FIELD_FILE_NAME "filename"
#define TABLE_IMAGE "imagecache"

#include "smartpointers.h"

class VideoThumbnailTools;

/**
 * A threadsafe class for fast obtaining of thumbnails
 *
 * The class checks for thumbnails in metadata. If not found it
 * tries to search in a local cache (SQLite database). If still not
 * found or outdated cache, it would try to (re-)compute the
 * thumbnail. All methods are threadsafe. But this has not been
 * tested veeery extensely.
 *
 * todo: windows, set hidden attribute
 *
 * @author Dominik Rueß dominik-ruess@gmx.de
 * @date 2011-09-06
 */
class FastImageThumbnail : public QObject
{
    Q_OBJECT

public:

    enum Thumbnail_Type{
        THUMB_TYPE_NO_FAST_THUMB,
        THUMB_TYPE_FAILED_LOADING,
        THUMB_TYPE_IMAGE,
        THUMB_TYPE_VIDEO,
        THUMB_TYPE_TEST_VIDEO
    };

    /**
      * init the database connection
      * @param QString location name of the sqlite database file
      */
    FastImageThumbnail(const QString& location, QObject* parent = 0);

    /**
      * destroy and disconnect
      */
    ~FastImageThumbnail();

    /**
     * get the thumbnail for a given file
     *
     * the function tries to obtain the thumbnail in the following order:
     * 1. does a thumbnail in the metadata exist ?
     * 2. does a thumbnail in the local database file exist ?
     * 2a. if yes, is it up to date ?
     * 3. (re-)compute thumbnail and save to local database file
     */
     Thumbnail_Type getThumbnail(const QFileInfo& fileinfo,
                                 MiscTools::pQImage& thumbnailOut,
                                 const bool onlyFast = false);
     /**
      * make database operations atomic
      */
     static QMutex databaseMutex;

signals:
     void newThumbnail(const QFileInfo &file, const MiscTools::pQImage);
     void errorThumb(const QFileInfo&);

private slots:
     void newVideoThumbnail(const QString &videoFilename, const QImage &image);

     void errorVideoThumbnail(const QString &videoFilename);
private:

    /**
     * convert a raw image to a QImage (including flipping of y-dimension)
     */
     //bool _convertToQImage(FIBITMAP * input, MiscTools::pQImage& out);

     bool _connectToLocalDB(const QString& filename);

     void _disconnect();

     bool _checkLocalCache(const QFileInfo& fileinfo,
                           MiscTools::pQImage& out);

     bool _updateCache(const QFileInfo& fileinfo,
                       const MiscTools::pQImage& thumbnail);

     void orientThumbnail(const QString& fn, QImage& thumbnail);

     /**
      * @brief _removeBlackBorder some embedded thumbnails have black borders
      * around the image, this is removed here
      */
     void _removeBlackBorder(QImage&);

    /**
     * try to load a thumbnail from an image path (try QImage and libraw)
     */
     QImage _tryLoadThumbnail(const QString& lpszPathName);

    /**
     * load a complete image
     */
     //FIBITMAP* _loadImage(const QString& lpszPathName);

    // Variables ---------------------------------------------------------------

    /**
     * make I/O operations atomic
     * However, this might not be a problem at all. But it does
     * cascade the heavy io image operations, possibly
     * making the single ones faster.
     */
    static QMutex _ioMutex;

    /**
      * holds the permanent connection to the database
      */
    QSqlDatabase* _db;


    /**
     * make sure it is small enough (e.g. 256, see THUMBNAIL_MAX_SIZE)
     */
    bool resizeToThumbSize(QImage& thumbnail);

    QImage _movieOverlayGraphic;

    LibRaw RawProcessor;

    VideoThumbnailTools * _videoThumbnailLoader;
};

#endif // FASTIMAGETHUMBNAIL_H
