#include "videothumbnails.h"

#include <QThread>
#include <QMediaResource>
#include <QImage>
#include <QVideoFrame>
#include <QApplication>
#include <QtMultimediaWidgets/QVideoWidget>
#include <QDebug>
#include <QFileInfo>
#include <QTimer>
#include <QMediaPlaylist>

#include "qandroidmultimediautils.h"

VideoThumbnailTools* VideoThumbnailTools::_instance = 0;

VideoThumbnailTools::~VideoThumbnailTools()
{
    disconnect();    
    _thumbnailPlayer->deleteLater();
    _videoProbe->deleteLater();
    _videoWidget->deleteLater();
}

void VideoThumbnailTools::extractFrame(const QString &filename,
                              const float position)
{
   QMutexLocker l(&_protectWaitingQueue);
    _waitingQueue.push_back(QPair<QString, float>(QFileInfo(filename).canonicalFilePath(),
                                                  position));
	if (_waitingQueue.size() == 1) {
		_waitingQueue.push_back(_waitingQueue.first()); // first element will be deleted upon call of "_nextinput"
		QTimer::singleShot(0, this, SLOT(_nextInput()));
	}
}

VideoThumbnailTools::VideoThumbnailTools()
	: _thumbnailPlayer(new QMediaPlayer)
	, _videoProbe(new QVideoProbe())
	, _videoWidget(new QVideoWidget())
	, _numFramesReceived(0)
{
	setupPlayer();
}

void VideoThumbnailTools::setupPlayer()
{
    _thumbnailPlayer->setVideoOutput(_videoWidget);
    _videoWidget->setVisible(false);
    _thumbnailPlayer->setMuted(true);
    connect(_videoProbe, SIGNAL(videoFrameProbed(QVideoFrame)),
            this, SLOT(newThumbFrame(QVideoFrame)), Qt::QueuedConnection);

    _videoProbe->setSource(_thumbnailPlayer);

    connect(_thumbnailPlayer, SIGNAL(stateChanged(QMediaPlayer::State)),
            this, SLOT(mediaStateChanged(QMediaPlayer::State)), Qt::QueuedConnection);
    connect(_thumbnailPlayer, SIGNAL(mediaStatusChanged(QMediaPlayer::MediaStatus)),
            this, SLOT(mediaStatusChanged(QMediaPlayer::MediaStatus)), Qt::QueuedConnection);

    connect(_thumbnailPlayer, SIGNAL(error(QMediaPlayer::Error)),
            this, SLOT(mediaError(QMediaPlayer::Error)), Qt::QueuedConnection);
}

void VideoThumbnailTools::resetPlayer()
{
    _thumbnailPlayer->stop();
	_thumbnailPlayer->setMedia(QMediaContent());
	_videoProbe->setSource((QMediaObject*)0);

    _isWaiting = QDateTime();
}

void VideoThumbnailTools::mediaStatusChanged(QMediaPlayer::MediaStatus status)
{
    switch(status) {
    case QMediaPlayer::LoadedMedia:
	{
		// start playing and seek to position if wanted and possible
		if (_positionWanted != 0.0f && _thumbnailPlayer->isSeekable()) {
			const qint64 newPosition = _thumbnailPlayer->duration() * _positionWanted;
			_thumbnailPlayer->setPosition(newPosition);
		}
		_thumbnailPlayer->play();
        break;
    }
	case QMediaPlayer::BufferedMedia:
		if (_thumbnailPlayer->state() != QMediaPlayer::PlayingState)
			_thumbnailPlayer->play();
		break;
    case QMediaPlayer::InvalidMedia:
    case QMediaPlayer::StalledMedia:
    case QMediaPlayer::EndOfMedia:
    {
		QMutexLocker l(&_protectWaitingQueue);
		if (_waitingQueue.size() > 0) {
			emit errorThumbnail(_waitingQueue.first().first);
		}
		QMediaContent media = _thumbnailPlayer->media();
		if (!media.isNull()) {
			resetPlayer();
		}
		QTimer::singleShot(0, this, SLOT(_nextInput()));
        break;
    }
	case QMediaPlayer::NoMedia:
	{
		QMutexLocker l(&_protectWaitingQueue);
		if (_waitingQueue.size() > 0) {
			QTimer::singleShot(0, this, SLOT(_nextInput()));
		}
	}
    case QMediaPlayer::LoadingMedia:
    default:
        break;
    }
}

void VideoThumbnailTools::mediaError(QMediaPlayer::Error)
{
    QMutexLocker l(&_protectWaitingQueue);
	if (_waitingQueue.size() > 0) {
		emit errorThumbnail(_waitingQueue.first().first);
		QTimer::singleShot(0, this, SLOT(_nextInput()));
	}
    resetPlayer();
}

void VideoThumbnailTools::mediaStateChanged(QMediaPlayer::State state)
{
    switch(state) {
    case QMediaPlayer::PlayingState:
    default:
        break;
    }
}

VideoThumbnailTools *VideoThumbnailTools::instance()
{
    static CGuard g;   // Speicherbereinigung
    if (!_instance)
        _instance = new VideoThumbnailTools ();
    return _instance;
}

void VideoThumbnailTools::_nextInput()
{
    QMutexLocker l(&_protectWaitingQueue);
    if (_waitingQueue.size() > 0 && !_isWaiting.isValid()) {
		_waitingQueue.pop_front();
		if (_waitingQueue.size() == 0) {
			// we're done
			return;
		}
        _isWaiting = QDateTime::currentDateTime();
        const QString& filename = _waitingQueue.first().first;
        _positionWanted = _waitingQueue.first().second;
        QMediaPlayer * player = _thumbnailPlayer;
		_numFramesReceived = 0;
		_videoProbe->setSource(_thumbnailPlayer);
        player->setMedia(QUrl::fromLocalFile(filename));
    } else if (_isWaiting.isValid()) {
        if (_isWaiting.msecsTo(QDateTime::currentDateTime()) > VIDEO_TOOLS_THUMBNAIL_MAX_WAIT_TIME) {
            // don't wait for too long, just abort loading
			if (_waitingQueue.size() > 0) {
				emit errorThumbnail(_waitingQueue.first().first);
			}
            resetPlayer();
			QTimer::singleShot(0, this, SLOT(_nextInput()));
        } else {
            QTimer::singleShot(5, this, SLOT(_nextInput()));
        }
    }
}

void VideoThumbnailTools::newThumbFrame(QVideoFrame frameOrig)
{
	if ( (!frameOrig.isValid() && _numFramesReceived++ < MAX_WAIT_FRAME) 
		|| _thumbnailPlayer->state() != QMediaPlayer::PlayingState)

	{
		return;
	}
	QImage result = qt_imageFromVideoFrame(frameOrig);
	QMutexLocker l(&_protectWaitingQueue);
	if (_waitingQueue.size() > 0) {
		const QString filename = _waitingQueue.first().first;
		if (!result.isNull() && result.width() > 0 && result.height() > 0) {
			emit newThumbnail(filename, result);
		}
		else {
			emit errorThumbnail(filename);
		}
    }
	resetPlayer();
#ifndef _WIN32
    QTimer::singleShot(0, this, SLOT(_nextInput()));
#endif
}


VideoThumbnailTools::CGuard::~CGuard()
{
    if( NULL != VideoThumbnailTools::_instance )
    {
        VideoThumbnailTools::_instance->deleteLater();
        VideoThumbnailTools::_instance = NULL;
    }
}
