#ifndef VIDEOTOOLS_H
#define VIDEOTOOLS_H

#include <QObject>
#include <QMediaPlayer>
#include <QThread>
#include <QImage>
#include <QVideoFrame>
#include <QVideoProbe>
#include <QAbstractVideoSurface>
#include <QWaitCondition>
#include <QMutex>
#include <QList>
#include <QDateTime>

#define VIDEO_TOOLS_THUMBNAIL_MAX_WAIT_TIME 1500
#define MAX_WAIT_FRAME 5

/**
 * singleton class, such that only 2 mediaplayer instances are needed
 * https://de.wikibooks.org/wiki/C%2B%2B-Programmierung:_Entwurfsmuster:_Singleton
 *
 * A note for Ubuntu users: The gstreamer-0.10-ffmpeg package is needed for some MP4
 * videos (H264). However this is not distributed with modern Ubuntu versions.
 * try to install with (http://askubuntu.com/a/575888):
 */
class VideoThumbnailTools : public QObject
{
    Q_OBJECT
public:
    /**
     * @brief instance return an instance of the video tools class (singleton pattern)
     * @return VideoTools pointer instance
     */
    static VideoThumbnailTools* instance ();

    /**
     * @brief extractFrame extract a frame from a video
     * It will try to seek to the specified position and
     * emit a signal with the frame or an error signal
     * @param filename the filename of the vdieo
     * @param position position between 0.0f and 1.0f
     */
    void extractFrame(const QString& filename,
                      const float position = 0.5f);

signals:
    void newThumbnail(QString filename, QImage image);
    void errorThumbnail(QString filename);

private slots:

    void newThumbFrame(QVideoFrame frameOrig);
    void mediaStateChanged(QMediaPlayer::State state);
    void mediaStatusChanged(QMediaPlayer::MediaStatus);

private:
    // ------------ singleton stuff -------------
    static VideoThumbnailTools* _instance;
    VideoThumbnailTools ();
    VideoThumbnailTools ( const VideoThumbnailTools& );
    ~VideoThumbnailTools ();

    class CGuard
    {
    public:
        ~CGuard();
    };
    // ------------ end singleton stuff -------------

private slots:
	void setupPlayer();

    void _nextInput();

    void resetPlayer();

    void mediaError(QMediaPlayer::Error);
private:

    QMediaPlayer* _thumbnailPlayer;
    QVideoProbe* _videoProbe;
    QVideoWidget* _videoWidget;

    QMutex _protectWaitingQueue;
    QList<QPair<QString, float> > _waitingQueue;

	int _numFramesReceived;

    float _positionWanted;

    QDateTime _isWaiting;
};

#endif // VIDEOTOOLS_H
