/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifdef WIN32
#include <WinSock2.h>
#endif

#include <QtWidgets>
#include <QMessageBox>
#include <QSettings>
#include <QMessageBox>
#include <QDebug>
#include <QMenu>
#include <QWhatsThis>
#include <QStyle>
#include <QTimer>
#include <QHeaderView>

#include "adapttimesdialog.h"
#include "ui_adapttimesdialog.h"
#include "settings.h"
#include "imagedataextended.h"
#include "tools.h"

#include <viewimagedialog.h>

#include <imageviewmodel.h>


AdaptTimesDialog::AdaptTimesDialog(QWidget *parent)
    : QDialog(parent)
    ,ui(new Ui::AdaptTimesDialog)
    ,_init(true)
    , _anchorPos(-1)
    , _targetPos(-1)
    , _infoHasBeenShown(false)
{
    ui->setupUi(this);

    loadSettings();

    _viewReference = new QAction(tr("View &reference images"), this);
    _viewTarget = new QAction(tr("View &target images"), this);
    _viewAll = new QAction(tr("View &merged album of both directories"), this);
    ui->tableWidget_comparison->setContextMenuPolicy(Qt::CustomContextMenu);
    QObject::connect (ui->tableWidget_comparison, SIGNAL(customContextMenuRequested(const QPoint &)),
                      this, SLOT(_contextMenuEvent(const QPoint&)));
    ui->tableWidget_comparison->setContextMenuPolicy(Qt::CustomContextMenu);
    QObject::connect (ui->tableWidget_comparison, SIGNAL(customContextMenuRequested(const QPoint &)),
                      this, SLOT(_contextMenuEvent(const QPoint&)));

    connect(ui->comboBox_referenceDir, SIGNAL(currentIndexChanged(int)),
            this, SLOT(display()));
    connect(ui->comboBox_targetDir, SIGNAL(currentIndexChanged(int)),
            this, SLOT(display()));
    ui->tableWidget_comparison->setItemDelegate(&_adaptTimesDelegate);

    connect(ui->spinBox_hours, SIGNAL(editingFinished()),
            this, SLOT(_recomputeInputDiff()));
    connect(ui->spinBox_min, SIGNAL(editingFinished()),
            this, SLOT(_recomputeInputDiff()));
    connect(ui->spinBox_sec, SIGNAL(editingFinished()),
            this, SLOT(_recomputeInputDiff()));

    _init = false;    
    //ui->tableWidget_comparison->resizeColumnsToContents();
    QHeaderView* header = ui->tableWidget_comparison->horizontalHeader();
    header->setSectionResizeMode(QHeaderView::Stretch);
}

void AdaptTimesDialog::loadSettings()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    this->restoreGeometry(settings.value("adaptTimesAssistant_dialog_geometry").toByteArray());

    const int thumbSize = settings.value("adaptTimesAssistant_thumbnailsize", 64).toInt();
    _adaptTimesDelegate.setThumbnailSize(thumbSize);
    ui->spinBox_thumb->setValue(thumbSize);

    ui->tableWidget_comparison->setColumnCount(2);
    ui->tableWidget_comparison->setHorizontalHeaderLabels(QStringList() << tr("Reference Directory") << tr("Target Directory"));
    ui->tableWidget_comparison->setColumnWidth(0, settings.value("adaptTimesAssistant_column0Width", 200).toInt());
    ui->tableWidget_comparison->setColumnWidth(1, settings.value("adaptTimesAssistant_column1Width", 200).toInt());
    ui->tableWidget_comparison->setSelectionMode(QAbstractItemView::SingleSelection);
    ui->tableWidget_comparison->setSelectionBehavior (QAbstractItemView::SelectRows);
    ui->tableWidget_comparison->setIconSize(QSize(thumbSize, thumbSize));
}


AdaptTimesDialog::~AdaptTimesDialog()
{
    delete ui;
}

void AdaptTimesDialog::display()
{
    if (ui->comboBox_referenceDir->currentIndex() == 0) {
        ui->label_timediff->setText(tr("Add or remove time of selected dir: "));
    } else {
        ui->label_timediff->setText(tr("Time difference of selected directories: "));
    }

    if (ui->comboBox_referenceDir->count() > 0 && ui->comboBox_referenceDir->currentText()
            == ui->comboBox_targetDir->currentText()) {
        QMessageBox::critical(this,
                             tr("Same Directories"),
                             tr("Please select different sub directories as reference and target values."));
        return;
    }

    _targetPos = -1;
    _anchorPos = -1;
    _adaptTimesDelegate.setAnchorPos(-1);
    _adaptTimesDelegate.setTargetPos(-1);
    _displayList();
}

void AdaptTimesDialog::on_pushButton_reset_clicked()
{
    _resetTimes();
}

bool AdaptTimesDialog::_resetTimes()
{
    if (QMessageBox::question(this,
                         tr("Reset"),
                         tr("This will reset the image times of all subdirectories to their original times."
                            "\nContinue ?"),
                          QMessageBox::Ok | QMessageBox::Cancel)
            != QMessageBox::Ok) {
        return false;
    }

    for (int i=0; i< (int)_images.size(); i++) {
        static_cast<ImageDataExtended*>(_images[i].get())->setTimeShift(0);
    }

    ui->tableWidget_comparison->viewport()->update();
    ui->spinBox_hours->setValue(0);
    ui->spinBox_min->setValue(0);
    ui->spinBox_sec->setValue(0);
    return true;
}


void AdaptTimesDialog::closeEvent(QCloseEvent *)
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.setValue("adaptTimesAssistant_dialog_geometry", this->saveGeometry());
    settings.setValue("adaptTimesAssistant_column0Width", ui->tableWidget_comparison->columnWidth(0));
    settings.setValue("adaptTimesAssistant_column1Width", ui->tableWidget_comparison->columnWidth(1));
    settings.setValue("adaptTimesAssistant_thumbnailsize", ui->spinBox_thumb->value());
}

// global
bool lessThanMixedList(const SortHelp& left,
                       const SortHelp& right)
{
    const ImageDataExtended* leftP = static_cast<ImageDataExtended*>(left.imagedata.get());
    const ImageDataExtended* rightP = static_cast<ImageDataExtended*>(right.imagedata.get());
    if (!leftP->getTime().isValid()) {
        return false;
    } else if (!rightP->getTime().isValid()) {
        return true;
    } else {
        return leftP->getTime() < rightP->getTime();
    }

}

void AdaptTimesDialog::_displayList()
{
    if (_init)
        return;

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    const QString refText = ui->comboBox_referenceDir->currentText();
    const QString targetText = ui->comboBox_targetDir->currentText();

    _currentList.clear();
    for (int i=0; i<(int)_images.size(); i++) {
        const QString path = _images[i]->getFileInfo().canonicalPath();
        const QString subDir = path.length() -_baseLength < 0 ? "." : path.right(path.length() -_baseLength);
        if ((ui->comboBox_referenceDir->currentIndex() > 0 && subDir == refText)
                || subDir == targetText) {
            SortHelp tmp;
            tmp.imagedata = _images[i];
            tmp.subDir = subDir;
            _currentList.push_back(tmp);
        }
    }

    if (_currentList.size() == 0 ) {
        qApp->restoreOverrideCursor();
        return;
    }

    qStableSort(_currentList.begin(), _currentList.end(), lessThanMixedList);

    ui->tableWidget_comparison->clear();
    ui->tableWidget_comparison->setHorizontalHeaderLabels(QStringList() << tr("Reference Directory Images") << tr("Target Directory Images"));
    ui->tableWidget_comparison->setRowCount(_currentList.size());
    const int thumbSize = ui->spinBox_thumb->value();
    for (int i=0; i < (int)_currentList.size(); i++) {
        QTableWidgetItem* newItem=NULL, *date=NULL;
        const ImageDataExtended* tmp = static_cast<ImageDataExtended*>(_currentList[i].imagedata.get());

        int posIm = 1;
        if (_currentList[i].subDir == refText) {
            posIm = 0;
        }
        _setTableWidget(tmp, &date, &newItem, posIm);

        ui->tableWidget_comparison->setItem(i, posIm, newItem);
        ui->tableWidget_comparison->setItem(i, (posIm+1)%2, date);

        ui->tableWidget_comparison->setRowHeight(i, thumbSize);
    }
    ui->tableWidget_comparison->viewport()->update();
    _updateFields();
    qApp->restoreOverrideCursor();
}

void AdaptTimesDialog::_setTableWidget(const ImageDataExtended* tmp,
                                       QTableWidgetItem** date,
                                       QTableWidgetItem** newItem,
                                       const int posIm)
{
    const MiscTools::pQImage thumb = tmp->getThumbnail();

    const QString dateStr = tmp->getFileInfo().fileName() + "\n" +
            (tmp->getTime().isValid() ?
                tmp->getTime().toString(Qt::DefaultLocaleLongDate)
              : "no time" );
    if (!thumb->isNull() && thumb->width() > 0 && thumb->height() > 0) {
        MiscTools::pQPixmap newPixmap = QPixmap::fromImage(thumb->scaled(QSize(256,256), Qt::KeepAspectRatio));
        *newItem = new QTableWidgetItem(QIcon(*newPixmap), dateStr);
    } else {
        MiscTools::pQPixmap newPixmap;
        if (tmp->hasThumbnail() == true) {
            *newPixmap = QPixmap::fromImage(ImageViewModel::errorThumb->scaled(QSize(256,256), Qt::KeepAspectRatio));
        } else {
            *newPixmap = QPixmap::fromImage(ImageViewModel::noThumb->scaled(QSize(256,256), Qt::KeepAspectRatio));
        }
        *newItem = new QTableWidgetItem(QIcon(*newPixmap), dateStr);
        //hier nothumb und rescale :-)
    }


    (*newItem)->setTextAlignment(Qt::AlignVCenter | (posIm == 0 ? Qt::AlignRight : Qt::AlignLeft));

    *date = new QTableWidgetItem("");

    (*date)->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
    (*newItem)->setFlags((*date)->flags());
}


void AdaptTimesDialog::setDirectories(const QStringList& directories,
                                      const QList<ImageDataPtr>& fileList,
                                      const int baseLength)
{
    _images = fileList;
    _baseLength = baseLength;

    const bool oldInit = _init;
    _init = true;

    if (_directories == directories
            && _images == fileList)
    {
        _displayList();
        _init = oldInit;
        return;
    }

    _directories = directories;
    ui->comboBox_referenceDir->clear();
    ui->comboBox_targetDir->clear();
    _timeDiff.clear();

    // add this, so the users can edit directory times
    // without a reference list
    ui->comboBox_referenceDir->addItem(tr("No reference (manually change time)"));

    for (int i=0; i< _directories.size(); i++) {
        ui->comboBox_referenceDir->addItem(_directories[i]);
        ui->comboBox_targetDir->addItem(_directories[i]);
        _timeDiff.push_back(QVector<int>(_directories.size()));
        for (int k=0; k<_timeDiff.last().size(); k++ ) {
            _timeDiff.last()[k] = 0;
        }
    }
    _init = oldInit;

    _displayList();
}

void AdaptTimesDialog::on_spinBox_thumb_editingFinished()
{
    const int thumbSize = ui->spinBox_thumb->value();
    for (int i=0; i<(int)ui->tableWidget_comparison->rowCount(); i++) {
        ui->tableWidget_comparison->setRowHeight(i, thumbSize);
    }
    _adaptTimesDelegate.setThumbnailSize(thumbSize);
}

void AdaptTimesDialog::on_tableWidget_comparison_itemSelectionChanged()
{
    QList<QTableWidgetSelectionRange> selection = ui->tableWidget_comparison->selectedRanges();
    const int row = selection.size() == 1? selection[0].topRow() : -1;

    if (row >= 0) {
        if (ui->tableWidget_comparison->item(row, 1)->data(Qt::DisplayRole) == "") {
            ui->toolButton_setAnchor->setEnabled(true);
            ui->toolButton_justBefore->setEnabled(false);
            ui->toolButton_justAfter->setEnabled(false);
        } else {
            ui->toolButton_setAnchor->setEnabled(false);
            ui->toolButton_justBefore->setEnabled(true);
            ui->toolButton_justAfter->setEnabled(true);
        }
    } else {
        ui->toolButton_setAnchor->setEnabled(false);
        ui->toolButton_justBefore->setEnabled(false);
        ui->toolButton_justAfter->setEnabled(false);
    }
}

void AdaptTimesDialog::on_spinBox_thumb_valueChanged(int)
{
    on_spinBox_thumb_editingFinished();
}

void AdaptTimesDialog::on_toolButton_setAnchor_clicked()
{
    QList<QTableWidgetSelectionRange> selection = ui->tableWidget_comparison->selectedRanges();
    const int row = selection.size() == 1? selection[0].topRow() : -1;

    if (row >= 0) {

        QDateTime reference = static_cast<ImageDataExtended*>(_currentList[row].imagedata.get())->getTime();
        if (!reference.isValid()) {
                QMessageBox::warning(this,
                                     tr("Please select image with valid date/time"),
                                     tr ("You have selected a file without  timestamp.\n"
                                         "Date/Time adaptation can only be done on files with date/time information."));
                return;
        }

        _anchorPos = row;
        _adaptTimesDelegate.setAnchorPos(row);
        _targetPos = -1;
        _adaptTimesDelegate.setTargetPos(-1);

        ui->tableWidget_comparison->viewport()->update();
    }
}

void AdaptTimesDialog::_computeTimeDifference()
{
    // find anchor
    if (_anchorPos <0 || _targetPos < 0
            || _anchorPos>= _currentList.size()
            || _targetPos >= _currentList.size()) {
        return;
    }
    QDateTime reference = static_cast<ImageDataExtended*>(_currentList[_anchorPos].imagedata.get())->getTime();
    QDateTime target = static_cast<ImageDataExtended*>(_currentList[_targetPos].imagedata.get())->getTime(false);
    if (!target.isValid() || !reference.isValid()) {
        return;
    }

    qint64 diff =  target.secsTo(reference);

    const int ind1 = ui->comboBox_referenceDir->currentIndex(),
            ind2 = ui->comboBox_targetDir->currentIndex();
    _timeDiff[ind1][ind2] = diff;

    _updateFields();
}

void AdaptTimesDialog::_updateFields()
{
    const qint64 diff = _timeDiff[ui->comboBox_referenceDir->currentIndex()][ui->comboBox_targetDir->currentIndex()];
    const int hours = diff / 3600;
    const int minutes =  (diff - hours*3600) / 60;
    const int seconds = (diff - hours*3600 - minutes*60 );

    ui->spinBox_hours->setValue(hours);
    ui->spinBox_min->setValue(minutes);
    ui->spinBox_sec->setValue(seconds);
}

void AdaptTimesDialog::_updateAnchorsAndTimes(const int before)
{
    if (_anchorPos <0) {
        QMessageBox::warning(this,
                             tr("Please select anchor first"),
                             tr ("Please select an anchor image first.\n"
                                 "Anchor images can be any of the reference images."));
        return;
    }
    QList<QTableWidgetSelectionRange> selection = ui->tableWidget_comparison->selectedRanges();
    const int row = selection.size() == 1 ? selection[0].topRow() : -1;

    if (row >= 0) {
        QDateTime target = static_cast<ImageDataExtended*>(_currentList[row].imagedata.get())->getTime(false);
        if (!target.isValid()) {
                QMessageBox::warning(this,
                                     tr("Please select image with valid date/time"),
                                     tr ("You have selected a file without timestamp.\n"
                                         "Date/Time adaptation can only be done on files with date/time information."));
                return;
        }

        _targetPos = row;
        _computeTimeDifference();
        _updateImagesTimes(before);
        _updateKeepAnchors();
    }

    ui->tableWidget_comparison->viewport()->update();
}

void AdaptTimesDialog::newThumbnail(const QString& filepath)
{
    const QString refText = ui->comboBox_referenceDir->currentText();
    const int thumbSize = ui->spinBox_thumb->value();
    for (int i=0; i<(int)_currentList.size(); i++) {
        ImageDataExtended * tmp = static_cast<ImageDataExtended*>(_currentList[i].imagedata.get());
        if (tmp->getFileInfo().canonicalFilePath() == filepath) {
            QTableWidgetItem* newItem=NULL, *date = NULL;
            const ImageDataExtended* tmp = static_cast<ImageDataExtended*>(_currentList[i].imagedata.get());


            int posIm = 1;
            if (_currentList[i].subDir == refText) {
                posIm = 0;
            }
            _setTableWidget(tmp, &date, &newItem, posIm);

            ui->tableWidget_comparison->setItem(i, posIm, newItem);
            ui->tableWidget_comparison->setItem(i, (posIm+1)%2, date);

            ui->tableWidget_comparison->setRowHeight(i, thumbSize);
        }
    }
    ui->tableWidget_comparison->viewport()->update();
}

void AdaptTimesDialog::on_toolButton_justBefore_clicked()
{
    _updateAnchorsAndTimes(-1);
}

void AdaptTimesDialog::on_toolButton_justAfter_clicked()
{
    _updateAnchorsAndTimes(1);
}


void AdaptTimesDialog::_updateKeepAnchors()
{
    const QString oldAnchor = _anchorPos < 0 ? "" : _currentList[_anchorPos].imagedata->getFileInfo().canonicalFilePath();
    const QString oldTarget = _targetPos < 0 ? "" : _currentList[_targetPos].imagedata->getFileInfo().canonicalFilePath();
    _displayList();

    _targetPos = -1;
    _anchorPos = -1;

    for (int i=0; i<(int) _currentList.size(); i++) {
        if (_currentList[i].imagedata->getFileInfo().canonicalFilePath() == oldAnchor) {
            _anchorPos = i;
            _adaptTimesDelegate.setAnchorPos(i);
        } else if (_currentList[i].imagedata->getFileInfo().canonicalFilePath() == oldTarget) {
            _targetPos = i;
            _adaptTimesDelegate.setTargetPos(i);
        }
    }

    if (_targetPos >=0) {
        ui->tableWidget_comparison->selectRow(_targetPos);
    }
}

void AdaptTimesDialog::_updateImagesTimes(const int before)
{
    const int indLeft = ui->comboBox_referenceDir->currentIndex(),
            indRight = ui->comboBox_targetDir->currentIndex();

    _timeDiff[indLeft][indRight] += before;
    const qint64 diff = _timeDiff[indLeft][indRight];

    for (int i=0; i<(int) _currentList.size(); i++) {
        if (_currentList[i].subDir == ui->comboBox_targetDir->currentText()) {
            static_cast<ImageDataExtended*>(_currentList[i].imagedata.get())->setTimeShift(diff);
        }
    }
}

void AdaptTimesDialog::_recomputeInputDiff()
{
    if (ui->comboBox_referenceDir->count() > 0) {
        const qint64 diff = ui->spinBox_hours->value()*3600
                + ui->spinBox_min->value() * 60
                + ui->spinBox_sec->value();
        _timeDiff[ui->comboBox_referenceDir->currentIndex()][ui->comboBox_targetDir->currentIndex()] = diff;
        _updateImagesTimes(0);
        _updateKeepAnchors();

        ui->tableWidget_comparison->viewport()->update();
    }
}


void AdaptTimesDialog::_contextMenuEvent(const QPoint &pos)
{
    QMenu menu(this);
    menu.addAction(_viewReference);
    menu.addAction(_viewTarget);
    menu.addAction(_viewAll);
    QAction *a = menu.exec(ui->tableWidget_comparison->viewport()->mapToGlobal(pos));
    menu.close();

    if (a ==_viewAll){
        _onViewAll(ui->tableWidget_comparison->rowAt(pos.y()));
    } else if (a == _viewReference) {
        _onViewReference(ui->tableWidget_comparison->rowAt(pos.y()));
    } else if (a == _viewTarget) {
        _onViewTarget(ui->tableWidget_comparison->rowAt(pos.y()));
    }
}

void AdaptTimesDialog::_onViewAll(const int row)
{
    QList<ImageDataPtr> showImages;
    for (int i=0; i<(int)_currentList.size(); i++) {
        showImages.push_back(_currentList[i].imagedata);
    }
    ViewImageDialog * viewImages = ViewImageDialog::instance(this->parentWidget());
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();
    viewImages->setImages(showImages,
                          row >=0 && row <_currentList.size() ? row : 0,
                          _baseLength);
}

void AdaptTimesDialog::_onViewReference(const int row)
{
    QList<ImageDataPtr> showImages;
    int useStart = 0;
    for (int i=0; i<(int)_currentList.size(); i++) {
        if (_currentList[i].subDir == ui->comboBox_referenceDir->currentText()) {
            showImages.push_back(_currentList[i].imagedata);
        }
        if (i == row) {
            useStart = showImages.size()-1;
        }
    }
    ViewImageDialog * viewImages = ViewImageDialog::instance(this->parentWidget());
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();
    viewImages->setImages(showImages,
                          useStart,
                          _baseLength);
}

void AdaptTimesDialog::_onViewTarget(const int row)
{
    QList<ImageDataPtr> showImages;
    int useStart = 0;
    for (int i=0; i<(int)_currentList.size(); i++) {
        if (_currentList[i].subDir == ui->comboBox_targetDir->currentText()) {
            showImages.push_back(_currentList[i].imagedata);
        }
        if (i == row) {
            useStart = showImages.size()-1;
        }
    }

    ViewImageDialog * viewImages = ViewImageDialog::instance(this->parentWidget());
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();
    viewImages->setImages(showImages,
                          useStart,
                          _baseLength);
}

void AdaptTimesDialog::_on_pushButton_finish_clicked()
{
    emit doneEditing();
}

void AdaptTimesDialog::clear()
{
    ui->tableWidget_comparison->clear();
    ui->tableWidget_comparison->setRowCount(0);
}

void AdaptTimesDialog::showInfo()
{
    if (!_infoHasBeenShown) {
        if (!MiscTools::MiscTools::lockDialog()) {
            QTimer::singleShot(10000, this, SLOT(showInfo()));
            return;
        }
        _infoHasBeenShown = true;
        Tools::provideInfo(this,
                           tr("Times Adaptation"),
                           tr("<i>This is the time adaptation dialog.</i><br><br>"
                              "If you have images from different cameras,"
                              " they are very likely to have different time settings. "
                              "Here you can adapt these times by adding a time span to"
                              " images.<br>"
                              "The procedure is sub-directory based:<ol>"
                              "<li>Select a reference directory and a directory to be edited</li>"
                              "<li>Locate an image pair which has been shot almost at the same time</li>"
                              "<li>For this pair, select the reference image as anchor and the"
                              " second as target</li>"
                              "</ol>If necessary repeat for different sub-directories."),
                           tr("adapttimesview"));
        MiscTools::MiscTools::unlockDialog();
    }
}
