/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <cmath>
#include <QDebug>
#include <QDir>
#include <QMessageBox>
#include <QApplication>
#include <QSqlDatabase>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QFuture>
#include <QtConcurrent/QtConcurrent>

#include "copyfiles.h"
#include "filenamenumbering.h"
#include "misctools.h"
#include "progressdialog.h"

QVector<PictureTag> CopyFiles::tags;// = _initializeTags();

void CopyFiles::initializeTags()
{
    QVector<PictureTag> out;

    out.push_back(PictureTag(tr("[Y]"),
                             tr("File's year"),
                             "year",
                             "yyyy"));
    out.push_back(PictureTag(tr("[M]"),
                             tr("File's month"),
                             "month",
                             "M"));
    out.push_back(PictureTag(tr("[D]"),
                             tr("File's day"),
                             "day",
                             "d"));
    out.push_back(PictureTag(tr("[H]"),
                             tr("File's hour"),
                             "hour",
                             "h"));
    out.push_back(PictureTag(tr("[MI]"),
                             tr("File's minute"),
                             "minute",
                             "m"));
    out.push_back(PictureTag(tr("[S]"),
                             tr("File's second"),
                             "second",
                             "s"));

    out.push_back(PictureTag(tr("[YE]"),
                             tr("Year of earliest file in album"),
                             "yearEarly",
                             "yyyy"));
    out.push_back(PictureTag(tr("[ME]"),
                             tr("Month of earliest file in album"),
                             "monthEarly",
                             "M"));
    out.push_back(PictureTag(tr("[DE]"),
                             tr("Day of earliest file in album"),
                             "dayEarly",
                             "d"));

    out.push_back(PictureTag(tr("[YL]"),
                             tr("Year of latest file in album"),
                             "yearLate",
                             "yyyy"));
    out.push_back(PictureTag(tr("[ML]"),
                             tr("Month of latest file in album"),
                             "monthLate",
                             "M"));
    out.push_back(PictureTag(tr("[DL]"),
                             tr("Day of latest file in album"),
                             "dayLate",
                             "d"));


    out.push_back(PictureTag(tr("[A]"),
                             tr("File's album"),
                             "album"));

    out.push_back(PictureTag(tr("[N]"),
                             tr("File's number within album (start from 1)"),
                             "number"));

    out.push_back(PictureTag(tr("[CURRDIR]"),
                             tr("Loaded directory"),
                             "currdir"));

    out.push_back(PictureTag(tr("[HOME]"),
                             tr("User directory; "
                                "<font color=\"blue\">\"[HOME]/Pictures/\"</font>"
                                " maps<br>to"
                                " <font color=\"blue\">\"My Pictures\"</font>"
                                "on most platforms"),
                             "home"));

    out.push_back(PictureTag(tr("[SUBDIR]"),
                             tr("File's mapped subdirectory"),
                             "subdir"));

    out.push_back(PictureTag(tr("[UnknownDay]"),
                             tr("See 'Date/Time' tab for details"),
                             "unknownday"));
    out.push_back(PictureTag(tr("[AlbumUnknownDay]"),
                             tr("See 'Date/Time' tab for details"),
                             "albumunknownday"));

    out.push_back(PictureTag(tr("[COMMENT]"),
                             tr("File comment (see misc settings)"),
                             "comment"));

//    out.push_back(PictureTag(tr(""),
//                             tr(""),
//                             ""));

    tags = out;
}


QString CopyFiles::getTag(const QString& key)
{
    for (int i=0; i<(int)tags.size(); i++) {
        if (key == tags[i].key) {
            return tags[i].tag;
        }
    }
    throw "tag not found, please check implementation";
}

QString CopyFiles::getTagDesc(const QString& key)
{

    for (int i=0; i<(int)tags.size(); i++) {
        if (key == tags[i].key) {
            return tags[i].description;
        }
    }
    return "";
}

CopyItem CopyFiles::applyTags(const QStringList& destinationDescriptions,
                              const QString& currentDir,
                              const QVector<PhotoAlbum>& albums,
                              const QList<QPair<QString, QString> >& subDirs,
                              const QMap<int, QDateTime> earliestPerAlbum,
                              const QMap<int, QDateTime> latestPerAlbum,
                              const bool albumHasNoDayTime,
                              const QString& albumKnownDayTags,
                              const QString& knownDayTag,
                              const QString& albumUnknownDayTags,
                              const QString& unknownDayTag,
                              ImageDataExtended* image,
                              QList<ImageDataExtended*>& noAlbumError,
                              QList<ImageDataExtended*>& noTimeError)
{
    CopyItem out;
    out.image = image;

    for (int i=0; i<(int)destinationDescriptions.size(); i++) {
        QString desc = destinationDescriptions[i];
        /**
          * keep the numeric tag values,
          * they will be used for detecting existing files.
          * So descNoDatesFilled will have the tags replaced
          * properly, except for the numeric tags
          */
        QString descNoDatesFilled = desc;

        int album = -1;
        for (int j=0; j<albums.size(); j++) {
            if (albums[j].ID == image->getAlbumID()) {
                album = j;
                break;
            }
        }

        bool good = true;

        if (!albumHasNoDayTime) {
            desc = desc.replace(getTag("albumunknownday"), albumKnownDayTags, Qt::CaseInsensitive);
        } else {
            desc = desc.replace(getTag("albumunknownday"), albumUnknownDayTags, Qt::CaseInsensitive);
        }

        if (image->getDayIsKnown()) {
            desc = desc.replace(getTag("unknownday"), knownDayTag, Qt::CaseInsensitive);
        } else {
            desc = desc.replace(getTag("unknownday"), unknownDayTag, Qt::CaseInsensitive);
        }

        QString tmpComment = image->getComment().remove("[/\\|*>,])");
        desc = desc.replace(getTag("comment"), tmpComment, Qt::CaseInsensitive);
        descNoDatesFilled = desc;

        const QDateTime time = image->getTime();
        if (    (  (!time.isValid() || time.isNull())
                   && ( desc.indexOf(getTag("year"), Qt::CaseInsensitive) >= 0
                        || desc.indexOf(getTag("month"), Qt::CaseInsensitive) >= 0
                        || desc.indexOf(getTag("day"), Qt::CaseInsensitive) >= 0
                        || desc.indexOf(getTag("hour"), Qt::CaseInsensitive) >= 0
                        || desc.indexOf(getTag("minute"), Qt::CaseInsensitive) >= 0
                        || desc.indexOf(getTag("second"), Qt::CaseInsensitive) >= 0
                        )
                   )
                ||
                ( earliestPerAlbum.find(album) == earliestPerAlbum.end()
                  && ( desc.indexOf(getTag("yearEarly"), Qt::CaseInsensitive) >= 0
                       || desc.indexOf("monthEarly", Qt::CaseInsensitive) >= 0
                       || desc.indexOf("dayEarly", Qt::CaseInsensitive)  >= 0
                       )
                  )
                ||
                ( latestPerAlbum.find(album) == latestPerAlbum.end()
                  && ( desc.indexOf(getTag("yearLate"), Qt::CaseInsensitive) >= 0
                       || desc.indexOf("monthLate", Qt::CaseInsensitive) >= 0
                       || desc.indexOf("dayLate", Qt::CaseInsensitive)  >= 0
                       )
                  )
                )
        {
            good = false;
            noTimeError.push_back(image);
            image->setCopyStatus(CopyStatusNoTime);
        } else {
            desc = desc.replace(getTag("currdir"), QDir::toNativeSeparators(currentDir), Qt::CaseInsensitive);
            desc = desc.replace(getTag("home"), QDir::toNativeSeparators(QDir::homePath()), Qt::CaseInsensitive);
            desc = desc.replace(getTag("year"), time.toString("yyyy"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("month"), time.toString("MM"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("day"), time.toString("dd"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("hour"), time.toString("hh"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("minute"), time.toString("mm"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("second"), time.toString("ss"), Qt::CaseInsensitive);

            desc = desc.replace(getTag("yearEarly"), earliestPerAlbum[album].toString("yyyy"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("monthEarly"), earliestPerAlbum[album].toString("MM"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("dayEarly"), earliestPerAlbum[album].toString("dd"), Qt::CaseInsensitive);

            desc = desc.replace(getTag("yearLate"), latestPerAlbum[album].toString("yyyy"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("monthLate"), latestPerAlbum[album].toString("MM"), Qt::CaseInsensitive);
            desc = desc.replace(getTag("dayLate"), latestPerAlbum[album].toString("dd"), Qt::CaseInsensitive);
        }

        if (album == -1
                && desc.indexOf(getTag("album"), Qt::CaseInsensitive) >= 0) {
            noAlbumError.push_back(image);
            image->setCopyStatus(CopyStatusNoAlbum);
            good = false;
        } else if (album >= 0){
            desc = desc.replace(getTag("album"), albums[album].name, Qt::CaseInsensitive);
            descNoDatesFilled = descNoDatesFilled.replace(getTag("album"), albums[album].name, Qt::CaseInsensitive);
        }

        if (good) {
            int subDir = -1;
            const QString path = image->getFileInfo().canonicalPath();
            const QString imageSubDir = path.length() - 1 - currentDir.length() < 0 ? "." : path.right(path.length() - 1 - currentDir.length());

            for (int j=0; j<(int)subDirs.size(); j++) {
                if (subDirs[j].first == imageSubDir) {
                    subDir = j;
                    break;
                }
            }

            const QString suffix = "." + image->getFileInfo().suffix();
            QString tmp = desc;
            QString tmpNoDatesFilled = descNoDatesFilled;
            const QString withoutSubdir = tmp.replace(getTag("subdir"), "", Qt::CaseInsensitive).trimmed() + suffix;
            const QString withoutSubdirNoDatesFilled = tmpNoDatesFilled.replace(getTag("subdir"), "", Qt::CaseInsensitive).trimmed() + suffix;
            if (subDir < 0) {
                desc = withoutSubdir;
                descNoDatesFilled = withoutSubdirNoDatesFilled;
            } else {
                out.subdir = subDirs[subDir].second;
                desc = desc.replace(getTag("subdir"), subDirs[subDir].second +QDir::separator(), Qt::CaseInsensitive).trimmed()  + suffix;
                descNoDatesFilled = descNoDatesFilled.replace(getTag("subdir"), subDirs[subDir].second +QDir::separator(), Qt::CaseInsensitive).trimmed()  + suffix;
            }
            const int pathPos = withoutSubdir.lastIndexOf(QDir::separator());
            QString withoutSubdirPath = QDir::cleanPath(withoutSubdir.left(pathPos));
            out.destinations.push_back(QPair<QString, QString>(desc, withoutSubdirPath));
            out.destinationsNoDatesFilled.push_back(descNoDatesFilled);
        }
    }

    out.serialNumbers.resize(out.destinations.size());
    return out;
}

bool lessThanTimeSort(const ImageDataPtr t1, const ImageDataPtr t2)
{
    return static_cast<ImageDataExtended*>(t1.get())->getTime() < static_cast<ImageDataExtended*>(t2.get())->getTime();
}

bool lessThanTimeSortItem(const CopyItem& t1,
                          const CopyItem& t2)
{
    //return static_cast<ImageDataExtended*>(t1.image.get())->getTime() < static_cast<ImageDataExtended*>(t2.image.get())->getTime();
    return t1.image->getTime() < t2.image->getTime();
}
int CopyFiles::_getNewRunnNumber(const CopyItem& currCopyItem,
                                 const QVector<CopyItem>& out,
                                 const bool forceSameNumber,
                                 const QVector<QPair<QString, QString> >& tagsAndDateMap,
                                 const HandleExistingItems existingFiles,
                                 const int k,
                                 QStringList& filenamesExisting,
                                 QVector<DestinationInfo>& destinationRunningNumbers)
{

    int newRunningNum = 1;
    // find existing files with the same pattern,
    QString pattern = currCopyItem.destinationsNoDatesFilled[k];
    const int pSep = pattern.lastIndexOf(QDir::separator())+1;
    pattern = pattern.mid(pSep, pattern.lastIndexOf(".") - pSep);
    QString searchPattern = pattern + ".*";
    for (int j=0; j<CopyFiles::tags.size(); j++) {
        searchPattern.replace(CopyFiles::tags[j].tag, "*");
    }

    filenamesExisting.clear();

    const QString directory = QFileInfo(currCopyItem.destinations[k].first).path();
    QDir scanDir(directory);
    QVector< int > runningFilenumbers;
    QVector<QVector<QPair<int, int> > > startAndLengthOfRunnNumbers;
    if (scanDir.exists()) {
        filenamesExisting = scanDir.entryList(QStringList() << searchPattern);

        MiscTools::FilenameNumbering::getTagsBackMapAndNumbers(
                    filenamesExisting,
                    pattern,
                    tagsAndDateMap,
                    getTag("number"),
                    runningFilenumbers,
                    startAndLengthOfRunnNumbers);

        if (existingFiles == HandleExistingItems_Continue
                || existingFiles == HandleExistingItems_ContinueAndRenameExisting) {
            for (int j=0; j<runningFilenumbers.size(); j++) {
                newRunningNum = qMax(newRunningNum, runningFilenumbers[j]);
            }

            newRunningNum += 1;
        }
    } else {
        qCritical() << directory << " is no dir";
    }

    if (forceSameNumber) {
        // check out for same source dir -> use and increase this number
        for (int j=0; j<out.size(); j++) {
            if (/*out[j].image->getFileInfo().canonicalPath()
                    == currCopyItem.image->getFileInfo().canonicalPath()
                    && */
                    out[j].image->getAlbumID() == currCopyItem.image->getAlbumID()) {
                // choose fist serial numb, as they are update to  be the same, anyways
                newRunningNum = qMax(out[j].serialNumbers[0] + 1, newRunningNum);
            }
        }
    }

    destinationRunningNumbers.push_back(DestinationInfo(currCopyItem.destinations[k].second,
                                                        newRunningNum,
                                                        filenamesExisting,
                                                        runningFilenumbers,
                                                        startAndLengthOfRunnNumbers,
                                                        pattern
                                                        )
                                        );

    return newRunningNum;
}

QVector<CopyItem> CopyFiles::applyForAllFiles(
        const QStringList& descriptions,
        const QString & currentDir,
        const QVector<PhotoAlbum>& albums,
        const QList<QPair<QString, QString> >& subDirs,
        const QList<ImageDataPtr> imagesOrig,
        const bool jpgAndRAW,
        const bool unique,
        const bool copyOnlyOnce,
        const QSqlDatabase* db,
        const QString& albumKnownDayTags,
        const QString& knownDayTag,
        const QString& albumUnknownDayTags,
        const QString& unknownDayTag,
        const bool forceSameNumber,
        const HandleExistingItems existingFiles,
        QList<QPair<QString, QString> >& moveExistingFiles,
        QList<ImageDataExtended*>& noAlbumError,
        QList<ImageDataExtended*>& noTimeError,
        QList<ImageDataExtended*>& removeUnique,
        QList<ImageDataExtended*>& alreadyCopied,
        QList<ImageDataExtended*>& notSelected)
{
    noAlbumError.clear();
    noTimeError.clear();
    removeUnique.clear();
    alreadyCopied.clear();
    notSelected.clear();

    // build a tag vs qdatestring format map
    const int nTags = CopyFiles::tags.size();
    QVector<QPair<QString, QString> > tagsAndDateMap(nTags);
    for (int i=0; i<nTags; i++) {
        tagsAndDateMap[i] = QPair<QString, QString>(CopyFiles::tags[i].tag,
                                                    CopyFiles::tags[i].QDateFormatEquivalent);
    }


    QList<ImageDataPtr> images = imagesOrig;

    ProgressDialog* progress = new ProgressDialog();
    progress->setModal(true);
    progress->setMinMax(1, images.size());
    progress->setValue(1);
    progress->setLabel("");
    progress->show();
    progress->setLabel(tr("Computing copy list...") + (unique ? tr("\nThis may take a while if there are many duplicates") : "") );

    // sort by time
    QVector<CopyItem> out;

    QVector<DestinationInfo> destinationRunningNumbers;

    // copy to list of imagedataextended structure and determine album based minimum and maximum
    // also determine if there are albums containing files without day or time

    QVector<bool> foundNoDayTimeperAlbum(albums.size());
    for (int i=0; i<foundNoDayTimeperAlbum.size(); i++) {
        foundNoDayTimeperAlbum[i] = false;
    }

    QMap<int, QDateTime> earliestDatesPerAlbum, latestDatesPerAlbum;
    for (int i=0; i<(int)images.size(); i++) {
        ImageDataExtended* tmp = static_cast<ImageDataExtended*>(images[i].get());
        const int albumIdGlobal = tmp->getAlbumID();
        int albumId = -1;
        for (int j=0; j<albums.size(); j++) {
            if (albums[j].ID == albumIdGlobal) {
                albumId =j;
                break;
            }
        }
        if (albumId == -1) {
            continue;
        }
        if (!tmp->getDayIsKnown()) {
            foundNoDayTimeperAlbum[albumId] = true;
        }

        if (tmp->getTime().isValid()) {
            if (earliestDatesPerAlbum.find(albumId) == earliestDatesPerAlbum.end()) {
                earliestDatesPerAlbum.insert(albumId, tmp->getTime());
            } else {
                earliestDatesPerAlbum[albumId] = qMin(earliestDatesPerAlbum[albumId], tmp->getTime());
            }

            if (latestDatesPerAlbum.find(albumId) == latestDatesPerAlbum.end()) {
                latestDatesPerAlbum.insert(albumId, tmp->getTime());
            } else {
                latestDatesPerAlbum[albumId] = qMin(latestDatesPerAlbum[albumId], tmp->getTime());
            }
        }
    }

    qStableSort(images.begin(), images.end(), lessThanTimeSort);

    // check all images for running number
    // tags, etc...
    for (int i=0; i<(int)images.size(); i++) {
        if (progress->isCanceled() == true) {
            out.clear();
            noAlbumError.clear();
            noTimeError.clear();
            removeUnique.clear();
            alreadyCopied.clear();
            notSelected.clear();

            break;
        }
        if (progress->checkUpdateWanted()) {
            progress->setValue(i+1);
            qApp->processEvents();
        }

        ImageDataExtended* image = static_cast<ImageDataExtended*>(images[i].get());

        int albumId = -1;
        for (int j=0; j<albums.size(); j++) {
            if (albums[j].ID == image->getAlbumID()) {
                albumId =j;
                break;
            }
        }

        if ( !image->getUsed() ) {
            notSelected.push_back(image);
            continue;
        }
        // find alreadyCopied
        if ( copyOnlyOnce &&
             (image->getCopyStatus() == CopyStatusAlreadyCopied
              || checkHasBeenCopied(image, db))
             ) {

            image->setCopyStatus(CopyStatusAlreadyCopied);
            alreadyCopied.push_back(image);
            continue;
        }

        // now find other images, with possibly same size and MD5
        if (unique) {
            bool uniqueFound = true;
            for (int j=0; j<(int)out.size(); j++) {
                if (out[j].image->getFileInfo().size() > 0 &&
                        out[j].image->getFileInfo().size() == image->getFileInfo().size())
                {
                    //QString md1 = static_cast<ImageDataExtended*>(out[j].image.get())->getMD5();
                    QString md1 = out[j].image->getMD5();
                    QString md2 = image->getMD5();
                    if (md1 == md2 && md1.length() > 0) {
                        uniqueFound = false;
                        break;
                    }
                }
            }
            if (!uniqueFound) {
                image->setCopyStatus(CopyStatusRemovedUnique);
                removeUnique.push_back(image);
                continue;
            }
        }


        CopyItem currCopyItem = applyTags(descriptions,
                                 currentDir,
                                 albums,
                                 subDirs,
                                 earliestDatesPerAlbum,
                                 latestDatesPerAlbum,
                                 (albumId >= 0? foundNoDayTimeperAlbum[albumId] : false),
                                 albumKnownDayTags,
                                 knownDayTag,
                                 albumUnknownDayTags,
                                 unknownDayTag,
                                 static_cast<ImageDataExtended*>(images[i].get()),
                                 noAlbumError,
                                 noTimeError);

        if (currCopyItem.destinations.size() == 0) {
            // already has been added to noalbum or notime
            continue;
        }

        // now find images which have same basename (different extensions)
        int sameBasename = -1;
        if (jpgAndRAW) {
            for (int j=0; j<(int)out.size(); j++) {
                if (out[j].image->getFileInfo().canonicalPath() == currCopyItem.image->getFileInfo().canonicalPath()
                        && out[j].image->getFileInfo().completeBaseName() == currCopyItem.image->getFileInfo().completeBaseName()) {
                    sameBasename = j;
                    break;
                }
            }
        }


        if (sameBasename >= 0) {
            for (int k=0; k<(int)out[sameBasename].serialNumbers.size(); k++) {
                currCopyItem.serialNumbers[k] = out[sameBasename].serialNumbers[k];
                bool found = false;
                for (int j=0; j<(int)destinationRunningNumbers.size(); j++) {
                    if (destinationRunningNumbers[j].path == currCopyItem.destinations[k].second) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    QStringList filenamesExisting;
                    const int newRunningNum = _getNewRunnNumber(currCopyItem,
                                                                out,
                                                                forceSameNumber,
                                                                tagsAndDateMap,
                                                                existingFiles,
                                                                k,
                                                                filenamesExisting,
                                                                destinationRunningNumbers);
                    currCopyItem.serialNumbers[k] = newRunningNum;
                }
            }
        } else {
            // find existing destinations and increase their running numbers
            int ma = -1;
            for (int k=0; k<(int)currCopyItem.destinations.size(); k++ ) {
                // the file doesn't want to be copied to this particular destination ?
                if (!currCopyItem.image->getCopyDestinationStatus(k)) {
                    continue;
                }

                bool found = false;
                for (int j=0; j<(int)destinationRunningNumbers.size(); j++) {
                    if (destinationRunningNumbers[j].path == currCopyItem.destinations[k].second) {

                        destinationRunningNumbers[j].runningNumber++;
                        currCopyItem.serialNumbers[k] = destinationRunningNumbers[j].runningNumber;
                        found = true;
                        break;
                    }
                }
                if (!found) {QStringList filenamesExisting;
                    const int newRunningNum = _getNewRunnNumber(currCopyItem,
                                                                out,
                                                                forceSameNumber,
                                                                tagsAndDateMap,
                                                                existingFiles,
                                                                k,
                                                                filenamesExisting,
                                                                destinationRunningNumbers);
                    currCopyItem.serialNumbers[k] = newRunningNum;
                }
                ma = qMax(ma, currCopyItem.serialNumbers[k]);
            }
            // again run through all destinations of this and make sure
            // the running number ist set to the maximum number having occured
            // in the different destinations. this ensures same numbering
            // for all destinations of this file.
            // Also make sure that all the running numbers of the affected
            //destinations is set to this
            if (forceSameNumber) {
                for (int k=0; k<(int)currCopyItem.destinations.size(); k++ ) {
                    currCopyItem.serialNumbers[k] = ma;
                    for (int j=0; j<(int)destinationRunningNumbers.size(); j++) {
                        if (destinationRunningNumbers[j].path == currCopyItem.destinations[k].second) {
                            if (currCopyItem.image->getCopyDestinationStatus(k)) {
                                destinationRunningNumbers[j].runningNumber = ma;
                            } else {
                                // make sure the next image in the dest folder
                                // without curr is increased
                                destinationRunningNumbers[j].runningNumber = ma+1;
                            }
                            break;
                        }
                    }
                }
            }
        }
        out.push_back(currCopyItem);
    }

    // now care for existing files.
    // since the user may want to integrate the new files into
    // the old ones
    moveExistingFiles.clear();
    if (existingFiles == HandleExistingItems_Merge) {
        // vielleicht den code vom unteren else zweig hier benutzen, oder auslagern

        for (int i=0; i<(int)destinationRunningNumbers.size(); i++) {
            QVector<QDateTime> newFileTimes;
            QVector<QPair<int, int> > indicesNewFiles;
            for (int k=0; k<(int)out.size(); k++) {
                int found = -1;
                for (int j=0; j<out[k].destinations.size(); j++) {
                    if (destinationRunningNumbers[i].path == out[k].destinations[j].second) {
                        found = j;
                        break;
                    }
                }
                if (found >= 0) {
                    indicesNewFiles.push_back(QPair<int, int> (k, found) );
                    newFileTimes.push_back(out[k].image->getTime());
                }
            }
            QVector<QPair<int, int> > newlySortedList;
            QVector<MiscTools::FilenameNumbering::DT> fileDatesExisting;
            const int n = destinationRunningNumbers[i].existingFilesWithSamePattern.size();
            QStringList filenames;
            for (int j=0; j < n; j++) {
                filenames.push_back(destinationRunningNumbers[i].path + QDir::separator() + destinationRunningNumbers[i].existingFilesWithSamePattern[j]);
            }
            if (MiscTools::FilenameNumbering::mergeAndReorderFiles(
                        filenames,
                        newFileTimes,
                        destinationRunningNumbers[i].filepattern,
                        tagsAndDateMap,
                        newlySortedList,
                        fileDatesExisting)) {

                const int numberOfDigits = (int)std::ceil((float)log10((float)newlySortedList.size()+1));
                destinationRunningNumbers[i].runningNumber = newlySortedList.size();
                // for alle filenames in the current directories
                for (int j=0; j<n; j++) {

                    // determine new number of existing file
                    int found = -1;
                    for (int k=0; k < newlySortedList.size(); k++) {
                        if (newlySortedList[k].first == j) {
                            found = k;
                            break;
                        }
                    }
                    if (found == -1) {
                        qCritical() << "Could not find existing file in merged list, aborting...";
                        out.clear();
                        delete progress;
                        progress = 0;
                        return out;
                    }
                    _updateMoveExisting(destinationRunningNumbers[i].existingFilesWithSamePattern[j],
                                        destinationRunningNumbers[i].startAndLengthOfRunnNumbers[j],
                                        found+1,
                                        numberOfDigits,
                                        destinationRunningNumbers[i].path,
                                        moveExistingFiles);
                }
                // determine new numbers for existing files
                for (int j=0; j < indicesNewFiles.size(); j++) {
                    // determine new number of existing file
                    int found = -1;
                    for (int k=0; k < newlySortedList.size(); k++) {
                        if (newlySortedList[k].second == j) {
                            found = k;
                            break;
                        }
                    }
                    if (found == -1) {
                        qCritical() << "Could not find new file in merged list, aborting...";
                        out.clear();
                        delete progress;
                        progress = 0;
                        return out;
                    }
                    out[indicesNewFiles[j].first].serialNumbers[indicesNewFiles[j].second] = found+1;
                }

            } else {
                qCritical() << "could not merge existing and new files";
                out.clear();
                delete progress;
                progress = 0;
                return out;
            }
        }


    } else if (existingFiles == HandleExistingItems_ContinueAndRenameExisting) {
        // for all destination directories
        for (int i=0; i<destinationRunningNumbers.size(); i++) {
            const int numberOfDigits = (int)std::ceil((float)log10((float)destinationRunningNumbers[i].runningNumber+1));
            const int n = destinationRunningNumbers[i].existingFilesWithSamePattern.size();
            // for alle filenames in the current directories
            for (int j=0; j<n; j++) {
                _updateMoveExisting(destinationRunningNumbers[i].existingFilesWithSamePattern[j],
                                    destinationRunningNumbers[i].startAndLengthOfRunnNumbers[j],
                                    destinationRunningNumbers[i].existingRunningNumbers[j],
                                    numberOfDigits,
                                    destinationRunningNumbers[i].path,
                                    moveExistingFiles);
            }
        }
    }

    // now set the numbers, according to the overall count in each directory
    for (int i=0; i<(int)out.size(); i++) {
        int highestNumber = 1;
        if (forceSameNumber) {
            // choose highest running number of directory
            for (int j=0; j<out[i].destinations.size(); j++) {
                for (int k=0; k<(int)destinationRunningNumbers.size(); k++) {
                    if (destinationRunningNumbers[k].path == out[i].destinations[j].second) {
                        highestNumber = qMax(destinationRunningNumbers[k].runningNumber, highestNumber);
                        break;
                    }
                }
            }
        }
        if (existingFiles != HandleExistingItems_PerformAction) {
            // choose highest running number of directory
            for (int j=0; j<out[i].destinations.size(); j++) {
                for (int k=0; k<(int)destinationRunningNumbers.size(); k++) {
                    if (destinationRunningNumbers[k].path == out[i].destinations[j].second) {
                        highestNumber = qMax(destinationRunningNumbers[k].runningNumber, highestNumber);
                        break;
                    }
                }
            }
        }
        for (int j=0; j<out[i].destinations.size(); j++) {
            // let's check if the files goes to any destination
            // and get the index for the overall number
            int found = -1;
            for (int k=0; k<(int)destinationRunningNumbers.size(); k++) {
                if (destinationRunningNumbers[k].path == out[i].destinations[j].second) {
                    found = k;
                    break;
                }
            }
            // if yes, replace tag
            if (found >=0 ) {

                int numDigits = 1;
                if (forceSameNumber) {
                    // make sure the number of digits is the same for all destinations
                    numDigits = (int)std::ceil((float)log10((float)highestNumber+1));
                } else {
                    numDigits = (int)std::ceil((float)log10((float)destinationRunningNumbers[found].runningNumber+1));
                }
                out[i].destinations[j].first =
                        out[i].destinations[j].first.replace(getTag("number"),
                                                        QString("%1").arg(
                                                            (int)out[i].serialNumbers[j],
                                                            numDigits,
                                                            (int)10,
                                                            QChar('0')),
                                                        Qt::CaseInsensitive);

            }
        }
    }

//circular dependency check hier  einbauen, also schauen, ob es möglich ist, die datei zu bewegen
//entsprechend auch die fehlerhaften zurückgeben.

    delete progress;
    progress = 0;
    return out;
}

bool CopyFiles::doCopy(const QStringList& descriptions,
                       const QString & currentDir,
                       const QVector<PhotoAlbum>& albums,
                       const QList<QPair<QString, QString> >& subDirs,
                       const QList<ImageDataPtr> imagesOrig,
                       const bool jpgAndRAW,
                       const bool unique,
                       const bool updateMetadata,
                       const bool saveTextFileIfFailure,
                       const bool copyOnlyOnce,
                       const bool createOwnerShipTextFile,
                       const QSqlDatabase* db,
                       const QString& albumKnownDayTags,
                       const QString& knownDayTag,
                       const QString& albumUnknownDayTags,
                       const QString& unknownDayTag,
                       const CopyOverwriteMode overwriteMode,
                       const bool forceSameNumbers,
                       const HandleExistingItems existingFiles,
                       QList<QPair<QString, QString> >& moveExistingFiles,
                       QWidget* parent,
                       QList<ImageDataExtended*>& noAlbumError,
                       QList<ImageDataExtended*>& noTimeError,
                       QList<ImageDataExtended*>& removeUnique,
                       QList<ImageDataExtended*>& alreadyCopied,
                       QList<ImageDataExtended*>& notSelected,
                       QList<ImageDataExtended*>& copyError
                       )
{
    copyError.clear();
    QVector<CopyItem> items = CopyFiles::applyForAllFiles(
                descriptions,
                currentDir,
                albums,
                subDirs,
                imagesOrig,
                jpgAndRAW,
                unique,
                copyOnlyOnce,
                db,
                albumKnownDayTags,
                knownDayTag,
                albumUnknownDayTags,
                unknownDayTag,
                forceSameNumbers,
                existingFiles,
                moveExistingFiles,
                noAlbumError,
                noTimeError,
                removeUnique,
                alreadyCopied,
                notSelected);
    if (items.size() > 0) {
        // check for files to be moved
        const int nM = moveExistingFiles.size();
        if (nM > 0) {
            QVector<int> succesful;
            const QVector<int> order = MiscTools::FilenameNumbering::findMoveOrdering(moveExistingFiles);
            if (order.size() != moveExistingFiles.size()) {
                qCritical("Could not determine movement order for preview, circular dependency ?");
                return false;
            } else {
                for (int i=0; i<moveExistingFiles.size(); i++) {
                    bool retry = true;
                    while (retry) {
                        if (!QDir().rename(moveExistingFiles[order[i]].first,
                                          moveExistingFiles[order[i]].second)) {
                            const QMessageBox::StandardButton choice =
                                    QMessageBox::critical(
                                        parent,
                                        tr("Moving existing file not successful"),
                                        tr("Already existing target file %1 cannot be moved to %2"
                                           "This may happen, if the destination file is currently opened in a different program"
                                           ", or due to other mounting problems.\n\n"
                                           "Do you want to skip to the next files ?"
                                           " Beware that this might affect following files as well.")
                                        .arg(moveExistingFiles[order[i]].first)
                                        .arg(moveExistingFiles[order[i]].second),
                                        QMessageBox::Yes |
                                        QMessageBox::Retry
                                        );
                            if (choice == QMessageBox::Yes) {
                                retry = false;
                            }
                        } else {
                            succesful.push_back(order[i]);
                            retry = false;
                        }
                    }
                }
                qSort(succesful.begin(), succesful.end());
                for (int i=succesful.size()-1; i>=0; i--) {
                    moveExistingFiles.removeAt(succesful[i]);
                }
            }
        }

        bool aborted = false;
        ProgressDialog* progress = new ProgressDialog();
        progress->setModal(true);
        progress->setMinMax(1, imagesOrig.size());
        progress->setValue(1);
        progress->setLabel(tr("Copying ... ") );
        progress->show();

        bool ignoreAll = false;

        CopyOverwriteMode overWriteAll = CopyAsk;
        for (int i=0; i<(int)items.size(); i++) {
            if (progress->isCanceled() == true) {
                aborted = true;
                break;
            }
            progress->setValue(i+1);

            ImageDataExtended* data = items[i].image; //static_cast<ImageDataExtended*>(items[i].image.get());
            bool allSuccess = true;

            for (int j=0; j<(int)items[i].destinations.size(); j++) {
                // extract destination directory
                QString destinationDir = items[i].destinations[j].first.left(items[i].destinations[j].first.lastIndexOf(QDir::separator()));

                if (!items[i].image->getCopyDestinationStatus(j)) {
                    // destination was specifially and manually chosen not to be copied to
                    continue;
                }
                progress->setLabel(tr("Copying ... (%1)").arg(QFileInfo(items[i].destinations[j].first).fileName()) );

                bool done = false;
                int success = 1;
                while (!done) {
                    if (!QDir().exists(destinationDir)) {
                        QDir().mkpath(destinationDir);
                    }

                    if (QDir().exists(destinationDir)) {
                        if (QDir().exists(items[i].destinations[j].first)) {
                            bool doOverwriteForNow = false;
                            if (overwriteMode == CopyOverwrite
                                    || overWriteAll == CopyOverwrite) {
                                doOverwriteForNow = true;
                            } else if (overwriteMode == CopyAsk
                                       && overWriteAll == CopyAsk) {
                                QMessageBox::StandardButton choice =
                                        QMessageBox::critical(
                                            parent,
                                            tr("Destination file does exist"),
                                            tr("Do you want to overwrite the destination file %1 ?")
                                            .arg(items[i].destinations[j].first),
                                            QMessageBox::Yes |
                                            QMessageBox::No |
                                            QMessageBox::YesAll |
                                            QMessageBox::NoToAll
                                            );
                                if (choice == QMessageBox::No
                                        || choice == QMessageBox::NoToAll) {
                                    success = -1;
                                } else if (choice == QMessageBox::Yes
                                           || choice == QMessageBox::YesToAll) {
                                    doOverwriteForNow = true;
                                }

                                overWriteAll = choice == QMessageBox::NoToAll ?
                                            CopyNoOverwrite : ( choice == QMessageBox::YesToAll ?
                                                                    CopyOverwrite : CopyAsk);
                            } else if (overWriteAll == CopyNoOverwrite
                                       || overwriteMode == CopyNoOverwrite){
                                success = -1;
                            } else {
                                qCritical   () << "implementation error (overwrite copy)";
                                throw "not implemented";
                            }

                            if (doOverwriteForNow) {
                                QMessageBox::StandardButton choice = QMessageBox::Retry;
                                while (!QFile::remove(items[i].destinations[j].first)
                                       && choice == QMessageBox::Retry) {
                                    choice = QMessageBox::critical(
                                                parent,
                                                tr("Could not delete existing file"),
                                                tr("Could not delete existing file '%1' for overwriting." )
                                                .arg(items[i].destinations[j].first),
                                                QMessageBox::Abort |
                                                QMessageBox::Retry
                                                );
                                }
                                // if choice is retry, then it was successful
                                if (choice != QMessageBox::Retry) {
                                    success = -1;
                                }
                            }
                        }
                        if (success >= 0) {
                            QFile file(items[i].image->getFileInfo().canonicalFilePath());

                            // make GUI responsive. Put copying into separate thread
                            QFuture<bool> copyWait = QtConcurrent::run(&file, &QFile::copy, items[i].destinations[j].first );
                            while( !copyWait.isFinished() ) {
                                qApp->processEvents();
                            }
                            success = copyWait.result() ? 1 : 0;

                        }
                    } else {
                        success = 0;
                    }

                    bool ignore = false;
                    if (success == 0 && !ignoreAll) {
                        QMessageBox::StandardButton choice =
                                QMessageBox::critical(
                                    parent,
                                    tr("Could not copy file"),
                                    tr("Could not copy file \"%1\" to \"%2\".\n"
                                       "Please make sure there is enough space left one the device, "
                                       " the device is mounted properly, the source file exists "
                                       " and that you have the privileges to write this path.\n\n"
                                       "Retry ?").arg(items[i].image->getFileInfo().canonicalFilePath())
                                    .arg(items[i].destinations[j].first),
                                    QMessageBox::Abort |
                                    QMessageBox::No |
                                    QMessageBox::NoToAll |
                                    QMessageBox::Retry
                                    );
                        if (choice == QMessageBox::No) {
                            ignore = true;
                        } else if (choice == QMessageBox::NoToAll) {
                            ignoreAll = true;
                        } else if (choice == QMessageBox::Abort) {
                            data->setCopyStatus(CopyStatusCopyError);
                            return false;
                        }
                    }
                    done = (success != 0) | ignore | ignoreAll;
                }

                // now add the ownership to the text file, if given a specific directory
                if (success && createOwnerShipTextFile) {
                    if (items[i].subdir != "" ) {
                        const QString fnOwner = QFileInfo(items[i].destinations[j].second).absoluteFilePath() + QDir::separator() + tr("authors.txt");
                        const bool existed = QFile::exists(fnOwner);
                        QFile ownership(fnOwner);
                        ownership.open(QIODevice::Append | QIODevice::Text);
                        if (ownership.isOpen()) {
                            QTextStream out(&ownership);
                            if (!existed) {
                                out << tr("Files with specific authors (created by pivot, www.dominik-ruess.de/pivot):") << "\n\n";
                            }
                            out << QFileInfo(items[i].destinations[j].second).fileName() << "\t" << items[i].subdir << "\n";
                            ownership.close();
                        }
                    }
                }


                allSuccess &= (success >= 0 && done);

                if (success > 0
                        && updateMetadata
                        //&& static_cast<ImageDataExtended*>(items[i].image.get())->getAutomatedTime() != data->getTime()) {
                        && items[i].image->getAutomatedTime() != data->getTime()) {

                    if (!MiscTools::MiscTools::changeTimeStamp(items[i].destinations[j].first,
                                                               items[i].destinations[j].first,
                                                               data->getTime(),
                                                               saveTextFileIfFailure))
                    {
                        if (data->getFileInfo().suffix().toLower() == "jpg") {
                            qWarning() << "...failed to adapt metadata for "
                                     << items[i].destinations[j].first;
                        }
                    }
                }
            }

            if (allSuccess) {
                QMutexLocker l(&FastImageThumbnail::databaseMutex);
                QSqlTableModel model(0, *db);
                model.setTable(DATABASE_MD5_TABLE);
                int row=0;
                model.insertRows(row, 1);
                model.setData(model.index(row, 0), data->getFileInfo().fileName());
                model.setData(model.index(row, 1), QString(data->getMD5()));
                model.setData(model.index(row, 2), data->getFileInfo().size());
                if (!model.submitAll()) {
                    qCritical() << "error submitting md5 to database";
                }
                //update alreadyCopied
                data->setCopyStatus(CopyStatusSuccess);
            } else {
                data->setCopyStatus(CopyStatusCopyError);
                copyError.push_back(items[i].image);
            }
        }

        delete progress;
        progress = 0;
        return !aborted;
    }
    return true;
}

bool CopyFiles::checkHasBeenCopied(
    const ImageDataExtended* data,
    const QSqlDatabase* db)
{
    QMutexLocker l(&FastImageThumbnail::databaseMutex);
    QSqlTableModel model(0, *db);
    model.setTable(DATABASE_MD5_TABLE);
    model.setFilter(QString("filesize='%1'").arg(data->getFileInfo().size() ));
    model.select();
    bool found = false;
    for (int i=0; i<(int)model.rowCount(); i++) {
        QSqlRecord record = model.record(i);
        QString dbMD5(record.value("md5").toString()) ;
        if (dbMD5 == data->getMD5()) {
            found = true;
            break;
        }
    }
    return found;
}

void CopyFiles::_updateMoveExisting(const QString& newFilename,
                                    const QVector<QPair<int, int> > & startAndLengthOfRunnNumbers,
                                    const int desiredNumber,
                                    const int numberOfDigits,
                                    const QString& path,
                                    QList<QPair<QString, QString> >& moveExistingFiles)
{
    QString newFilenameOut = QString(newFilename);
    // for all appearances of the running number tag
    for (int k=startAndLengthOfRunnNumbers.size()-1; k>=0; k--) {
        const int start = startAndLengthOfRunnNumbers[k].first;
        const int len = startAndLengthOfRunnNumbers[k].second;
        const QString newNumber = QString("%1").arg(desiredNumber,
                                                    numberOfDigits,
                                                    (int)10,
                                                    QChar('0'));
        newFilenameOut.replace(start, len, newNumber);
    }
    if (newFilenameOut != newFilename) {
        // rename existing file
        const QString dir = path + QDir::separator();
        moveExistingFiles.push_back(
                    QPair<QString, QString>(
                        dir + newFilename,
                        dir + newFilenameOut));
    }
}
