/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QSettings>
#include <QDir>
#include <QString>
#include <QDebug>
#include <QSpacerItem>
#include <QMessageBox>
#include <cmath>
#include <QWhatsThis>
#include <QStyle>

#include "mainsettings.h"
#include "ui_mainsettings.h"

#include <thumbnailloader.h>

#include "settings.h"


#include "imagedata.h"

#include "copyfiles.h"

MainSettings::MainSettings(QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::MainSettings)
    , _spacerDestinations(NULL)
    , _init (true)
{
    ui->setupUi(this);

    qRegisterMetaType<DestinationSettingsOption>("DestinationSettingsOption");

    _loadSettings();

    this->setModal(true);

    const int numCols = 2;
    ui->gridLayout_tags->addWidget(new QLabel(tr("<b>Possible tags to use:<b> ")), 0, 0, 1, numCols*2);
    const int numPerColMax = std::ceil((float)CopyFiles::tags.size()/(float)numCols);
    for (int i=0; i<(int)CopyFiles::tags.size(); i++ ) {
        ui->gridLayout_tags->addWidget(new QLabel(CopyFiles::tags[i].tag),
                                       1 + i % numPerColMax,
                                       2*(i / numPerColMax));
        ui->gridLayout_tags->addWidget(new QLabel(CopyFiles::tags[i].description),
                                       1 + i % numPerColMax,
                                       2*(i / numPerColMax) +1);
    }
    for (int i=0; i<numCols; i++) {
        ui->gridLayout_tags->setColumnStretch(i*2, 1);
        ui->gridLayout_tags->setColumnStretch(i*2+1, 5);
    }


    QStyle * style = this->style();
    ui->toolButton_whatsThis->setIcon(style->standardPixmap(QStyle::SP_MessageBoxQuestion));

}

MainSettings::~MainSettings()
{
    delete ui;
}

void MainSettings::_loadSettings()
{
    _init = true;

    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    this->restoreGeometry(settings.value("settings_dialog_geometry").toByteArray());

    QStringList normalExtensions;
    normalExtensions
            << "jpg"
            << "jpeg"
            << "png"
            << "bmp"
            << "cr2"
            << "tiff"
            << "tif"
            << "gif"
            << "pbm"
            << "pgm"
            << "ppm"
            << "xbm"
            << "xpm"
            << "mpeg"
            << "svg"
            << "mpg"
            << "mpeg"
            << "divx"
            << "avi"
            << "mov"
            << "mts"
            << "mp4"
            << "wmv";

    QStringList rawExtensions;
    rawExtensions = ThumbnailLoader::returnRawFormats();

    ui->lineEdit_images->setText(settings.value("settings_extensions", normalExtensions).toStringList().join(", "));
    ui->lineEdit_rawImages->setText(settings.value("raw_extensions", rawExtensions).toStringList().join(", "));

    ui->spinBox_thumbnailSize->setValue(settings.value("settings_thumbsize",64).toInt());
    ui->checkBox_basename->setChecked(settings.value("settings_tryBasename", true).toBool());
    ui->checkBox_neighbours->setChecked(settings.value("settings_tryNeighbours", true).toBool());
    ui->checkBox_creationDate->setChecked(settings.value("settings_tryFileTime", true).toBool());
    ui->checkBox_metaData->setChecked(settings.value("settings_tryMetadata", true).toBool());
    ui->checkBox_updateMetadata->setChecked(settings.value("settings_tryUpdateMetaData", true).toBool());    
    ui->checkBox_createTextFile->setChecked(settings.value("settings_tryUpdateMetaDataTextFile", true).toBool());
    ui->checkBox_unique->setChecked(settings.value("settings_copyUnique", true).toBool());
    ui->checkBox_copyonce->setChecked(settings.value("settings_copyOnce", true).toBool());
    ui->checkBox_sameNumberRAW->setChecked(settings.value("settings_sameNumberRaw", true).toBool());    
    ui->checkBox_saveAlbumDate->setChecked(settings.value("settings_saveAlbumDate", true).toBool());
    ui->comboBox_overwrite->setCurrentIndex(settings.value("settings_overwriteMode", 0).toInt());
    ui->checkBox_differentTags->setChecked(settings.value("settings_differentTags", false).toBool());
    ui->checkBox_forceImageNumber->setChecked(settings.value("settings_forceSameNumber", false).toBool());
    ui->lineEdit_knownDay->setEnabled(ui->checkBox_differentTags->isChecked());
    ui->lineEdit_unknownDay->setEnabled(ui->checkBox_differentTags->isChecked());
    ui->lineEdit_knownDayAlbum->setEnabled(ui->checkBox_differentTags->isChecked());
    ui->lineEdit_unknownDayAlbum->setEnabled(ui->checkBox_differentTags->isChecked());
    ui->lineEdit_knownDay->setText(settings.value(QString("settings_tagsKnownDay_%1").arg(tr("en")),
                                                  (CopyFiles::getTag("year") + "-"
                                                   + CopyFiles::getTag("month") + "-"
                                                   + CopyFiles::getTag("day") + " "
                                                   + CopyFiles::getTag("album") + " "
                                                   + CopyFiles::getTag("number"))).toString());
    ui->lineEdit_unknownDay->setText(settings.value(QString("settings_tagsUnknownDay_%1").arg(tr("en")),
                                                    (CopyFiles::getTag("year") + "-"
                                                     + CopyFiles::getTag("month") + " "
                                                     + CopyFiles::getTag("album") + " "
                                                     + CopyFiles::getTag("number"))).toString());

    ui->lineEdit_knownDayAlbum->setText(settings.value(QString("settings_tagsKnownDayAlbum_%1").arg(tr("en")),
                                                       (CopyFiles::getTag("yearEarly") + "-"
                                                        + CopyFiles::getTag("monthEarly") + "-"
                                                        + CopyFiles::getTag("dayEarly") + " "
                                                        + CopyFiles::getTag("album"))).toString());
    ui->lineEdit_unknownDayAlbum->setText(settings.value(QString("settings_tagsUnknownDayAlbum_%1").arg(tr("en")),
                                                         (CopyFiles::getTag("yearEarly") + "-"
                                                          + CopyFiles::getTag("monthEarly") + " "
                                                          + CopyFiles::getTag("album"))).toString());
    ui->checkBox_fileComments->setChecked(settings.value("settings_enableFileComments", false).toBool());
    ui->checkBox_autosave->setChecked(settings.value("settings_autosave", true).toBool());
    ui->spinBox_interval->setValue(settings.value("settings_autosaveinterval", 10).toInt());
    ui->spinBox_interval->setEnabled(ui->checkBox_autosave->isChecked());

    ui->radioButton_action->setChecked(settings.value("settings_existingFilesPerformAction", false).toBool());
    ui->radioButton_continue->setChecked(settings.value("settings_existingFilesContinueNumbering", true).toBool());
    ui->radioButton_merge->setChecked(settings.value("existingFilesMergeNumbering", false).toBool());
    ui->checkBox_renameExisting->setChecked(settings.value("existingFilesRenamingNumDigits", true).toBool());
    ui->checkBox_renameExisting->setEnabled(ui->radioButton_continue->isChecked());
    ui->comboBox_overwrite->setEnabled(ui->radioButton_action->isChecked());

    ui->checkBox_createTextFile->setEnabled(ui->checkBox_updateMetadata->isChecked());


    ui->checkBox_ownershipFile->setChecked(settings.value("settings_createOwnerShip", true).toBool());


    QStringList destinations;
    if (!settings.contains(QString("settings_destinations_%1").arg(tr("en")))) {
        destinations.push_back(CopyFiles::getTag("home") + QDir::separator()
                               + tr("Pictures") + QDir::separator()
                               + CopyFiles::getTag("yearEarly") + QDir::separator()
                               + (CopyFiles::getTag("yearEarly") + "-"
                                  + CopyFiles::getTag("monthEarly") + "-"
                                  + CopyFiles::getTag("dayEarly") + " "
                                  + CopyFiles::getTag("album")
                                  + QDir::separator())
                               + (CopyFiles::getTag("year") + "-"
                                  + CopyFiles::getTag("month") + "-"
                                  + CopyFiles::getTag("day") + " "
                                  + CopyFiles::getTag("album") + " "
                                  + CopyFiles::getTag("number"))
                               );

        destinations.push_back(CopyFiles::getTag("home") + QDir::separator()
                               + tr("Original Pictures") + QDir::separator()
                               + CopyFiles::getTag("yearEarly") + QDir::separator()
                               + ("O_" + CopyFiles::getTag("yearEarly") + "-"
                                  + CopyFiles::getTag("monthEarly") + "-"
                                  + CopyFiles::getTag("dayEarly") + " "
                                  + CopyFiles::getTag("album")
                                  + QDir::separator())
                               + (CopyFiles::getTag("subdir"))
                               + ("O_" + CopyFiles::getTag("year") + "-"
                                  + CopyFiles::getTag("month") + "-"
                                  + CopyFiles::getTag("day") + " "
                                  + CopyFiles::getTag("album") + " "
                                  + CopyFiles::getTag("number"))
                               );
    } else {
        destinations = settings.value(QString("settings_destinations_%1").arg(tr("en"))).toStringList();
    }

    QGridLayout* layout = static_cast<QGridLayout*>(ui->scrollAreaWidgetContents->layout());
    for (int i=0; i<(int)_destinations.size(); i++) {
        layout->removeWidget(_destinations[i].add);
        layout->removeWidget(_destinations[i].remove);
        layout->removeWidget(_destinations[i].destinationDescription);
        delete _destinations[i].add;
        delete _destinations[i].remove;
        delete _destinations[i].destinationDescription;
    }
    _destinations.clear();
    for (int i=0; i<(int)destinations.size(); i++) {
        _destinations.push_back(_insertDestination(destinations[i], i, layout));
    }
    layout->setRowStretch(_destinations.size(), 1);

    _init = false;
}

DestinationSettingsOption MainSettings::_insertDestination(const QString& text,
                                                           const int row,
                                                           QGridLayout* layout)
{
    DestinationSettingsOption tmp;
    tmp.destinationDescription = new QLineEdit(text);
    tmp.add = new QToolButton();
    tmp.add->setText("+");
    tmp.remove = new QToolButton();
    tmp.remove->setText("-");
    connect(tmp.add, SIGNAL(clicked()), this, SLOT(_addDestinationRow()));
    connect(tmp.remove, SIGNAL(clicked()), this, SLOT(_removeDestinationRow()));
    layout->addWidget(tmp.destinationDescription, row, 0);
    layout->addWidget(tmp.remove, row, 1);
    layout->addWidget(tmp.add, row, 2);
    layout->setRowStretch(row, 0);
    return tmp;
}

QStringList MainSettings::getExtensions() const
{
    QString extensions1 = ui->lineEdit_images->text(),
            extensions2 = ui->lineEdit_rawImages->text();
    QStringList out;
    out += extensions1.split(",", QString::SkipEmptyParts);
    out += extensions2.split(",", QString::SkipEmptyParts);

    for (int i=0; i<(int)out.size(); i++) {
        out[i] = out[i].trimmed().toLower();
    }

    return out;
}

QStringList MainSettings::getRawExtensions() const
{
    QString extensions = ui->lineEdit_rawImages->text();
    QStringList out = extensions.split(",", QString::SkipEmptyParts);
    for (int i=0; i<out.size(); i++) {
        out[i] = out[i].trimmed();
    }
    return out;
}

void MainSettings::_saveSettings()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.setValue("settings_dialog_geometry", this->saveGeometry());

    settings.setValue("settings_extensions",ui->lineEdit_images->text());
    settings.setValue("raw_extensions", ui->lineEdit_rawImages->text());

    settings.setValue("settings_thumbsize",ui->spinBox_thumbnailSize->value());
    settings.setValue("settings_tryBasename",ui->checkBox_basename->isChecked());
    settings.setValue("settings_tryNeighbours",ui->checkBox_neighbours->isChecked());
    settings.setValue("settings_tryFileTime",ui->checkBox_creationDate->isChecked());
    settings.setValue("settings_tryMetadata", ui->checkBox_metaData->isChecked());
    settings.setValue("settings_tryUpdateMetaData",ui->checkBox_updateMetadata->isChecked());
    settings.setValue("settings_tryUpdateMetaDataTextFile", ui->checkBox_createTextFile->isChecked());
    settings.setValue("settings_copyUnique",ui->checkBox_unique->isChecked());
    settings.setValue("settings_copyOnce",ui->checkBox_copyonce->isChecked());
    settings.setValue("settings_overwriteMode", ui->comboBox_overwrite->currentIndex());
    settings.setValue("settings_sameNumberRaw", ui->checkBox_sameNumberRAW->isChecked());
    settings.setValue("settings_saveAlbumDate", ui->checkBox_saveAlbumDate->isChecked());
    settings.setValue("settings_differentTags", ui->checkBox_differentTags->isChecked());
    settings.setValue("settings_forceSameNumber", ui->checkBox_forceImageNumber->isChecked());
    settings.setValue(QString("settings_tagsKnownDay_%1").arg(tr("en")), ui->lineEdit_knownDay->text());
    settings.setValue(QString("settings_tagsUnknownDay_%1").arg(tr("en")), ui->lineEdit_unknownDay->text());
    settings.setValue(QString("settings_tagsKnownDayAlbum_%1").arg(tr("en")), ui->lineEdit_knownDayAlbum->text());
    settings.setValue(QString("settings_tagsUnknownDayAlbum_%1").arg(tr("en")), ui->lineEdit_unknownDayAlbum->text());
    settings.setValue("settings_enableFileComments", ui->checkBox_fileComments->isChecked());
    settings.setValue("settings_autosave", ui->checkBox_autosave->isChecked());
    settings.setValue("settings_autosaveinterval", ui->spinBox_interval->value());
    settings.setValue("settings_existingFilesPerformAction", ui->radioButton_action->isChecked());
    settings.setValue("settings_existingFilesContinueNumbering", ui->radioButton_continue->isChecked());
    settings.setValue("existingFilesMergeNumbering", ui->radioButton_merge->isChecked());
    settings.setValue("existingFilesRenamingNumDigits", ui->checkBox_renameExisting->isChecked());
    settings.setValue("settings_createOwnerShip", ui->checkBox_ownershipFile->isChecked());

    QStringList destinations;
    for (int i=0; i< (int)_destinations.size(); i++) {
        destinations.push_back(_destinations[i].destinationDescription->text());
    }
    settings.setValue(QString("settings_destinations_%1").arg(tr("en")), destinations);
}

void MainSettings::on_buttonBox_rejected()
{
    _loadSettings();
}

void MainSettings::on_buttonBox_accepted()
{
    _saveSettings();
}

int MainSettings::getThumbnailSize() const
{
    return ui->spinBox_thumbnailSize->value();
}

bool MainSettings::tryToObtainByBasename() const
{
    return ui->checkBox_basename->isChecked();
}

bool MainSettings::tryToObtainByNeighbour() const
{
    return ui->checkBox_neighbours->isChecked();
}

bool MainSettings::tryToObtainFileTime() const
{
    return ui->checkBox_creationDate->isChecked();
}

bool MainSettings::tryUseMetaData() const
{
    return ui->checkBox_metaData->isChecked();
}

bool MainSettings::tryUpdateTimeMetaData() const
{
    return ui->checkBox_updateMetadata->isChecked();
}

bool MainSettings::saveToTextFileIfFailure() const
{
    return ui->checkBox_createTextFile->isChecked();
}


void MainSettings::_removeDestinationRow()
{

    if (QMessageBox::question(this,
                              tr("Really remove ?"),
                              tr("Really remove destination ?"),
                              QMessageBox::Yes | QMessageBox::No)
            == QMessageBox::No) {
        return;
    }
    int row = -1;
    QGridLayout * layout = static_cast<QGridLayout*>(ui->scrollAreaWidgetContents->layout());

    for (int i=0; i<(int)_destinations.size(); i++) {
        if (row < 0 && sender() == _destinations[i].remove) {
            layout->removeWidget(_destinations[i].add);
            layout->removeWidget(_destinations[i].remove);
            layout->removeWidget(_destinations[i].destinationDescription);
#ifndef _WIN32
            /* TODO: these lines cause WINDOWS CORE exception
             * maybe qt5 does the deletion somewhere else, itself ?
             */
            delete _destinations[i].add;
            delete _destinations[i].remove;
            delete _destinations[i].destinationDescription;
            _destinations[i].add = 0;
            _destinations[i].remove = 0;
            _destinations[i].destinationDescription = 0;
#else
            // instead, for now (mem leak):
            _destinations[i].add->setVisible(false);
            _destinations[i].remove->setVisible(false);
            _destinations[i].destinationDescription->setVisible(false);
#endif
            row = i;
        } else if (row >=0) {
            layout->removeWidget(_destinations[i].add);
            layout->removeWidget(_destinations[i].remove);
            layout->removeWidget(_destinations[i].destinationDescription);
            layout->addWidget(_destinations[i].add, i-1, 2);
            layout->addWidget(_destinations[i].remove, i-1, 1);
            layout->addWidget(_destinations[i].destinationDescription, i-1, 0);
        }
    }

    if (row >=0 ) {

        // now delete afterwards, since elements after current row had to be processed, as well
        _destinations.erase(_destinations.begin() + row);

    }
    layout->setRowStretch(_destinations.size(), 1);
}

void MainSettings::_addDestinationRow()
{
    if (_destinations.size() >= MAX_NUM_COPY_DESTINATIONS) {
        // only MAX_NUM_COPY_DESTINATIONS possible, due to keyboard shortcut reasons in image viewer
        QMessageBox::warning(this, tr("Too many destination targets"), tr("Currently, only %1 different destinations are possible.").arg(MAX_NUM_COPY_DESTINATIONS));
        return;
    }
    QGridLayout * layout = static_cast<QGridLayout*>(ui->scrollAreaWidgetContents->layout());
    for (int i=(int)_destinations.size()-1; i>=0; i--) {
        if (sender() != _destinations[i].add) {
            layout->removeWidget(_destinations[i].add);
            layout->removeWidget(_destinations[i].remove);
            layout->removeWidget(_destinations[i].destinationDescription);
            layout->addWidget(_destinations[i].add, i+1, 2);
            layout->addWidget(_destinations[i].remove, i+1, 1);
            layout->addWidget(_destinations[i].destinationDescription, i+1, 0);
            layout->setRowStretch(i+1, 0);
        } else if (sender() == _destinations[i].add) {
            _destinations.insert(_destinations.begin()+i+1,
                                 _insertDestination("", i+1, layout));
            break;
        }
    }
    layout->setRowStretch(_destinations.size(), 1);
}


QStringList MainSettings::getDestinationDescriptions() const
{
    QStringList out;
    for (int i=0; i<(int)_destinations.size(); i++) {
        const QString text = _destinations[i].destinationDescription->text().trimmed();
        if (text.length() > 0)
            out.push_back(text);
    }
    return out;
}


bool MainSettings::getUseUniqueFiles() const
{
    return ui->checkBox_unique->isChecked();
}

bool MainSettings::getCombineRAWAndImage() const
{
    return ui->checkBox_sameNumberRAW->isChecked();
}

bool MainSettings::getCopyOnlyOnce() const
{
    return ui->checkBox_copyonce->isChecked();
}


bool MainSettings::getSaveAlbumDate() const
{
    return ui->checkBox_saveAlbumDate->isChecked();
}


bool MainSettings::getDifferentTags() const
{
    return ui->checkBox_differentTags->isChecked();
}

QString MainSettings::getTagsKnownDay() const
{
    return ui->lineEdit_knownDay->text();
}

QString MainSettings::getTagsUnknownDay() const
{
    return ui->lineEdit_unknownDay->text();
}

QString MainSettings::getTagsKnownDayAlbum() const
{
    return ui->lineEdit_knownDayAlbum->text();
}

QString MainSettings::getTagsUnknownDayAlbum() const
{
    return ui->lineEdit_unknownDayAlbum->text();
}

void MainSettings::on_pushButton_clicked()
{
    if (QMessageBox::question(this,
                              tr("Delete all settings ?"),
                              tr("Do you really want to reset all settings ?"),
                              QMessageBox::Yes | QMessageBox::No)
            == QMessageBox::No)
    {
        return;
    }

    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.clear();
    _loadSettings();

    emit resetSettings();
}

void MainSettings::on_pushButton_2_clicked(const bool preventDialog)
{
    if (!preventDialog) {
        if (QMessageBox::question(this,
                                  tr("Delete all user data ?"),
                                  tr("Do you really want to reset all user data, including albums, thumbnails and information about copied files ?"),
                                  QMessageBox::Yes | QMessageBox::No)
                == QMessageBox::No)
        {
            return;
        }
    }

    emit resetUserData();
}

bool MainSettings::getUseComments() const
{
    return ui->checkBox_fileComments->isChecked();
}



bool MainSettings::getDoAutosave() const
{
    return ui->checkBox_autosave->isChecked();
}

int MainSettings::getAutosaveInterval() const
{
    return ui->spinBox_interval->value();
}


CopyOverwriteMode MainSettings::getOverwriteMode() const
{
    switch(ui->comboBox_overwrite->currentIndex())
    {
    case 1:
            return CopyOverwrite;
    case 2:
            return CopyNoOverwrite;
            break;
    case 0:
    default:
        return CopyAsk;
        break;
    }
}

void MainSettings::on_toolButton_whatsThis_clicked()
{
    if (!QWhatsThis::inWhatsThisMode()) {
        QWhatsThis::enterWhatsThisMode();
    }
}

bool MainSettings::getForceSameNumber() const
{
    return ui->checkBox_forceImageNumber->isChecked();
}

void MainSettings::_duplicateMessage()
{
    if (!_init) {
        QMessageBox::warning(this,
                             tr("Possible Duplication of Files ?"),
                             tr("You have not checked the option to not "
                                "copy files which have already been copied in the past.\n\n"
                                "This may lead to files being copied twice or more.\n\n"
                                "Beware of already copied files up to this point may"
                                " not have been marked as already copied."));
    }
}

void MainSettings::on_radioButton_merge_toggled(bool checked)
{
    if (!_init) {
        if (checked) {
            QMessageBox::warning(this,
                                 tr("Experimental"),
                                 tr("This mode is still considered experimental!\n"
                                    "However, it has already been successfully tested in quite some setups,"
                                    " but it may involve massive renaming of existing files in "
                                    " the destination folders. So please be careful using this option and backup "
                                    " existing files in the destination folders, before you start copying.\n"
                                    "Also use the preview to make sure results will look like you extect them to.\n\n"
                                    " If you encounter bugs, don't hesitate to report them (see menu->help)"));
        }
        if (checked && !ui->checkBox_copyonce->isChecked()) {
            _duplicateMessage();

            if (ui->checkBox_forceImageNumber->isChecked()) {
                ui->checkBox_forceImageNumber->setChecked(false);
                QMessageBox::warning(this,
                                     tr("Not same number for different destinations"),
                                     tr("The option for forcing the same running number of"
                                        " one file in different destinations has been actived,"
                                        " since it is not implemented up to now, to merge "
                                        " new files with the existing files and retain the "
                                        " same running number at the same time."));
            }
        }
    }
}

void MainSettings::on_radioButton_continue_toggled(bool checked)
{
    if (checked && !ui->checkBox_copyonce->isChecked()) {
        _duplicateMessage();
    }
}

bool MainSettings::getExistingFilesMerge() const
{
    return ui->radioButton_merge->isChecked();
}

bool MainSettings::getExistingFilesContinue() const
{
    return ui->radioButton_continue->isChecked() & !ui->checkBox_renameExisting->isChecked();
}

bool MainSettings::getExistingFilesContinueRename() const
{
    return ui->radioButton_continue->isChecked() & ui->checkBox_renameExisting->isChecked();
}

bool MainSettings::getCreateOwnerShipText() const
{
    return ui->checkBox_ownershipFile->isChecked();
}


void MainSettings::on_checkBox_copyonce_toggled(bool checked)
{
    if (!checked && (ui->radioButton_merge->isChecked()
                        || ui->radioButton_continue->isChecked()) ) {
        _duplicateMessage();
    }
}

void MainSettings::checkNewVersionAndDeleteSettings(const int major,
                                                    const int minor,
                                                    const int patch)
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    const int lastMajor = settings.value("settings_major", -1).toInt();
    const int lastMinor = settings.value("settings_minor", -1).toInt();
    const int lastPatch = settings.value("settings_patch", -1).toInt();

    Q_UNUSED(lastPatch);
    Q_UNUSED(patch);
    if (lastMajor != major || lastMinor != minor) {
        qDebug() << "resetting the cache, thumbnails etc due to new version";
        on_pushButton_2_clicked(true);
    }


    settings.setValue("settings_major", major);
    settings.setValue("settings_minor", minor);
    settings.setValue("settings_patch", patch);
}
