/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QSettings>
#include <QFileDialog>
#include <QDebug>
#include <QMessageBox>
#include <QTimer>
#include <QSqlQuery>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QSqlError>
#include <QWhatsThis>
#include <QProgressDialog>
#include <QStyle>
#include <QHBoxLayout>
#include <QSpacerItem>
#include <QTextBrowser>
#include <QDesktopServices>
#include <QTimer>

#ifdef FOUND_LIB_SOLID
#include <solid/solidnamespace.h>
#include <solid/camera.h>
#include <solid/device.h>
#include <solid/deviceinterface.h>
#include <solid/devicenotifier.h>
#include <solid/predicate.h>
#include <solid/storageaccess.h>
#include <solid/storagedrive.h>
#include <solid/storagevolume.h>
#endif

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "settings.h"
#include "mainsettings.h"
#include "tools.h"
#include "misctools.h"
#include "filenamenumbering.h"
#include "imagebutton.h"
#include "version_pivot.h"

#include "viewimagedialogextended.h"
#include "copyfiles.h"


MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
    , _eventAssistant(this)
    , _loader(NULL)
    , _db(NULL)
    , _dbFotoSettings(NULL)
    , _infoBoxMode(0)
    , _lastRow(-1)
    , _assistantButtonLoadDirectory(NULL)
    , _assistantButtonViewImages(NULL)
    , _assistantButtonReviewFiles(NULL)
    , _assistantButtonAdaptTimes(NULL)
    , _assistantButtonAlbums(NULL)
    , _assistantButtonKeepDirectories(NULL)
    //, _assistantButtonSettings(NULL)
    , _assistantButtonPreview(NULL)
    , _assistantButtonStart(NULL)
    , _assistantButtonMinHeight(10)
    , _assistantButtonMargin(3)
    , _randomImageList(NULL)
    , _currentTab(NULL)
{
	//QTimer::singleShot(PRELOAD_IMAGE_VIEWER_TIME, IMAGE)

    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    // this call is not necessary, since it only runs static assertions
    // nevertheless it is placed here to make sure code is generated
    // as well as to give readers a hint on requirements
    version_pivot.checkRequirements();
		
    // clear cache in case of new version
    _settings.checkNewVersionAndDeleteSettings(version_pivot.getVersionMajor(),
                                               version_pivot.getVersionMinor(),
                                               version_pivot.getVersionPatch());

    ui->setupUi(this);
    this->setWindowTitle(tr("pivot - Photo, Image and Video Organization Tool"));

    // setup table view
    //_currentFiles.setShowDetails(settings.getShowDetails());
    ui->tableView_images->setShowGrid(true);
    ui->tableView_images->setAlternatingRowColors(true);
    ui->tableView_images->setModel(&_currentFiles);
    ui->tableView_images->setItemDelegate(&_delegate);
    ui->tableView_images->setTextElideMode(Qt::ElideLeft);
    ui->tableView_images->setSelectionMode(QAbstractItemView::MultiSelection);
    ui->tableView_images->setSelectionBehavior(QAbstractItemView::SelectRows);

    QObject::connect (ui->tableView_images, SIGNAL(customContextMenuRequested(const QPoint &)),
                      this, SLOT(_contextMenuEvent(const QPoint&)));
    ui->tableView_images->setContextMenuPolicy(Qt::CustomContextMenu);

    _viewSelected = new QAction(tr("View this with other &selected images"), this);
    _viewDir = new QAction(tr("View this with images from same &directories"), this);
    _viewAll = new QAction(tr("View &all images"), this);
    MiscTools::setAllToolTips(_viewSelected, tr("View all selected image, with image under cursor as starting image"));
    MiscTools::setAllToolTips(_viewDir, tr("View image under cursors, with all images of the same directory"));
    MiscTools::setAllToolTips(_viewAll, tr("View all images, if called from context menu, use image under cursor"));
    connect(_viewSelected, SIGNAL(triggered()), this, SLOT(_onViewSelected()));
    connect(_viewDir, SIGNAL(triggered()), this, SLOT(_onViewDir()));
    connect(_viewAll, SIGNAL(triggered()), this, SLOT(_onViewAll()));

    _albumReset = new QAction(tr("Reset &manual album changes"), this);
    _albumReset->setIcon(QIcon(":images/album_reset.png"));
    _albumReset->setIconVisibleInMenu(true);
    _albumUp = new QAction(tr("Set album of last selected item to other selected items"), this);
    _albumUp->setIcon(QIcon(":images/album_up.png"));
    _albumUp->setIconVisibleInMenu(true);
    _albumDown = new QAction(tr("Set album of first selected item to other selected items"), this);
    _albumDown->setIconVisibleInMenu(true);
    _albumDown->setIcon(QIcon(":images/album_down.png"));
    MiscTools::setAllToolTips(_albumReset, tr("Remove all manual changes concerning albums"));
    MiscTools::setAllToolTips(_albumUp, tr("Manually change album from first selected image to all other selected images"));
    MiscTools::setAllToolTips(_albumDown, tr("Manually change album from last selected image to all other selected images"));
    connect(_albumReset, SIGNAL(triggered()), this, SLOT(_on_albumReset()));
    connect(_albumUp, SIGNAL(triggered()), this, SLOT(_on_albumUp()));
    connect(_albumDown, SIGNAL(triggered()), this, SLOT(_on_albumDown()));

    _timeReset = new QAction(tr("Reset manual &time changes"), this);
    _timeReset->setIconVisibleInMenu(true);
    _timeReset->setIcon(QIcon(":images/time_reset.png"));
    _timeUp = new QAction(tr("Set time of &last selected item to other selected items"), this);
    _timeUp->setIconVisibleInMenu(true);
    _timeUp->setIcon(QIcon(":images/time_up.png"));
    _timeDown = new QAction(tr("Set time of &first selected item to other selected items"), this);
    _timeDown->setIconVisibleInMenu(true);
    _timeDown->setIcon(QIcon(":images/time_down.png"));
    MiscTools::setAllToolTips(_timeReset, tr("Remove all manual changes concerning file times"));
    MiscTools::setAllToolTips(_timeUp, tr("Manually change time from first selected image to all other selected images"));
    MiscTools::setAllToolTips(_timeDown, tr("Manually change time from last selected image to all other selected images"));
    connect(_timeReset, SIGNAL(triggered()), this, SLOT(_on_timeReset()));
    connect(_timeUp, SIGNAL(triggered()), this, SLOT(_on_timeUp()));
    connect(_timeDown, SIGNAL(triggered()), this, SLOT(_on_timeDown()));

    _destinationsReset = new QAction(tr("Reset manual &destination changes"), this);
    _destinationsReset->setIconVisibleInMenu(true);
    _destinationsReset->setIcon(QIcon(":images/XXX_reset.png"));
    _destinationsUp = new QAction(tr("Set file destinations of last selected item to other selected items"), this);
    _destinationsUp->setIconVisibleInMenu(true);
    _destinationsUp->setIcon(QIcon(":images/XXX_up.png"));
    _destinationsDown= new QAction(tr("Set file destinations of first selected item to other selected items"), this);
    _destinationsDown->setIconVisibleInMenu(true);
    _destinationsDown->setIcon(QIcon(":images/XXX_down.png"));
    MiscTools::setAllToolTips(_destinationsReset, tr("Remove all manual changes concerning file copy destinations (i.e. set to copy to all destinations)"));
    MiscTools::setAllToolTips(_destinationsUp, tr("Manually change copy destinations from first selected image to all other selected images"));
    MiscTools::setAllToolTips(_destinationsDown, tr("Manually change copy destinations from last selected image to all other selected images"));
    connect(_destinationsReset, SIGNAL(triggered()), this, SLOT(_on_destinationsReset()));
    connect(_destinationsUp, SIGNAL(triggered()), this, SLOT(_on_destinationsUp()));
    connect(_destinationsDown, SIGNAL(triggered()), this, SLOT(_on_destinationsDown()));


    _checkAll = new QAction(tr("Check all files (i.e. copy all files)"), this);
    _checkAll->setIconVisibleInMenu(true);
    _checkAll->setIcon(QIcon(":images/check_all.png"));
    _uncheckAll = new QAction(tr("Uncheck all files"), this);
    _uncheckAll->setIconVisibleInMenu(true);
    _uncheckAll->setIcon(QIcon(":images/uncheck_all.png"));
//    _invertCheck = new QAction(tr("Invert Selection"), this);
//    _invertCheck->setIconVisibleInMenu(true);
//    _invertCheck->setIcon(QIcon(":images/check_invert.png"));
    MiscTools::setAllToolTips(_checkAll, tr("Check all files. Meaning all files will be copied based on their destination images"));
    MiscTools::setAllToolTips(_uncheckAll, tr("Un-check all files. Meaning none of the files will be copied"));
    connect(_checkAll, SIGNAL(triggered()), &_currentFiles, SLOT(selectAll()));
    connect(_uncheckAll, SIGNAL(triggered()), &_currentFiles, SLOT(deselectAll()));

    _selectAll = new QAction(tr("Select all files"), this);
    _selectAll->setIconVisibleInMenu(true);
    _selectAll->setIcon(QIcon(":images/selectall.png"));
    _deselectAll = new QAction(tr("De-select all files"), this);
    _deselectAll->setIconVisibleInMenu(true);
    _deselectAll->setIcon(QIcon(":images/deselectall.png"));
    MiscTools::setAllToolTips(_selectAll, tr("Select all files (not to be confused with check/uncheck)"));
    MiscTools::setAllToolTips(_deselectAll, tr("De-select all files (not to be confused with check/uncheck)"));
    connect(_selectAll, SIGNAL(triggered()), SLOT(_on_selectAll()));
    connect(_deselectAll, SIGNAL(triggered()), SLOT(_on_deselectAll()));


    _toggleError = new QAction(tr("Toggle files with missing information (album, time, ...)"), this);
    _toggleError->setCheckable(true);
    _toggleError->setIconVisibleInMenu(true);
    _toggleError->setIcon(QIcon(":images/only_error.png"));
    MiscTools::setAllToolTips(_toggleError, tr("Toggle display mode - show only files with missing information (album, time, ...)"));
    connect(_toggleError, SIGNAL(triggered(bool)),
            &_currentFiles, SLOT(toggleNoTimeOrAlbum(bool)));
    connect(_toggleError, SIGNAL(triggered()),
            ui->tableView_images, SLOT(clearSelection()));

    _delegate.setThumbnailSize(_settings.getThumbnailSize());
    _delegate.setParent(ui->tableView_images);

    QMenu * mainEditMenu;
    _getEditMenu(&mainEditMenu, false );
    mainEditMenu->setTitle(("&Edit"));
    menuBar()->insertMenu(ui->menuHelp->menuAction(),
                          mainEditMenu);

    // connect actions
    connect(ui->action_Load_directory, SIGNAL(triggered()),
            this, SLOT(_loadDirectory()));
    connect(ui->action_Settings, SIGNAL(triggered()),
            &_settings, SLOT(show()));
    connect(ui->actionManage_Albums, SIGNAL(triggered()),
            this, SLOT(_onManageAlbums()));
//    connect(ui->actionAdapt_Dates_Times, SIGNAL(triggered()),
//            this, SLOT(_onAdaptTimes()));
    connect(ui->actionDirectories, SIGNAL(triggered()),
            this, SLOT(_onSubDirMap()));

//    connect(ui->toolButton_selectInv, SIGNAL(clicked()),
//            &_currentFiles, SLOT(invertSelection()));

    connect(&_currentFiles, SIGNAL(errorMessage(QString,QString)),
            this, SLOT(onErrorMessage(QString,QString)));

    connect(&_eventAssistant, SIGNAL(finishedAlbums()),
            this, SLOT(_doneAlbums()));

    connect (&_settings, SIGNAL(resetSettings()),
             this, SLOT(_resetSettings()));
    connect(&_settings, SIGNAL(resetUserData()),
            this, SLOT(_removeUserData()));

    connect(&_settings, SIGNAL(accepted()),
            this, SLOT(_onSettingsDone()));

    connect(&_currentFiles, SIGNAL(layoutChanged()),
            ui->tableView_images, SLOT(resizeRowsToContents()));

    connect(ui->action_Save_manual_settings, SIGNAL(triggered()),
            this, SLOT(_saveAlbumDate()));

    // connect assistant
    connect(ui->toolButton_whatsThis, SIGNAL(clicked()), this, SLOT(on_toolButton_whatsThis_clicked()));

    // create and set the path for the local image thumbnail database
    QString location = MiscTools::MiscTools::appDataPath();
    QDir locationDir(location);
    _settingsLocation = QDir::toNativeSeparators(location + QDir::separator() + SETTINGS_FOLDER);
    locationDir.mkpath(_settingsLocation);

    // also cause album dialog to load albums
    _eventAssistant.loadAlbumsFromFile(_settingsLocation);

    // set thumbnails file
    _databaseLocation = QDir::toNativeSeparators(_settingsLocation + QDir::separator() + "thumbnails.sqlite.db");
    // set already copied file
    _alreadyCopiedLocation = QDir::toNativeSeparators(_settingsLocation + QDir::separator() + ALREADY_COPIED_DB);
    _savedAlbumDateLocation = QDir::toNativeSeparators(_settingsLocation + QDir::separator() + SAVED_PROPERTIES_DB);

//    if (qApp->arguments().size() > 1) {
//        if (QFileInfo(qApp->arguments().at(1)).isDir()) {
//            _imageDir = qApp->arguments().at(1);
//            ui->label_currDir->setText(_imageDir);
//            settings.setValue("lastDirectory", _imageDir);

//            QTimer::singleShot(500, this, SLOT(_findImageFiles()));
//        }
//    }
    const QIcon icon(":/images/PhotoOrganizer.png");
    this->setWindowIcon(icon);
    ViewImageDialogExtended::instance(this)->setWindowIcon(this->windowIcon());

    connect(&_autosaveTimer, SIGNAL(timeout()),
            this, SLOT(_saveAlbumDate()));


    ui->tab_album->setLayout(new QHBoxLayout());
    ui->tab_album->layout()->addWidget(&_eventAssistant);
    connect(&_eventAssistant, SIGNAL(finishedAlbums()),
            this, SLOT(_doneAlbums()));

    ui->tab_adapttimes->setLayout(new QHBoxLayout());
    ui->tab_adapttimes->layout()->addWidget(&_adaptTimeAssistant);
//    connect(&_adaptTimeAssistant, SIGNAL(doneEditing()),
//            this, SLOT(_doneAdaptTimes()));

    ui->tab_keepdirectory->setLayout(new QHBoxLayout());
    ui->tab_keepdirectory->layout()->addWidget(&_directoryAssistant);
    connect(&_directoryAssistant, SIGNAL(doneEditing()),
            this, SLOT(_doneDirectoryAssistant()));

    ui->tab_preview->setLayout(new QHBoxLayout());
    ui->tab_preview->layout()->addWidget(&_preview);
//    connect(&_preview, SIGNAL(doneEditing()),
//            this, SLOT(_donePreviewDialog()));

    QStyle * style = this->style();
    ui->toolButton_whatsThis->setIcon(style->standardPixmap(QStyle::SP_MessageBoxQuestion));

    _assistantButtonLoadDirectory = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonLoadDirectory->setText(tr("&Load source directory"));
    _assistantButtonLoadDirectory->setPixmap(":/images/assistant_no1.svg", _assistantButtonMargin);
    _assistantButtonLoadDirectory->setStyleSheet("font:bold;");
    MiscTools::setAllToolTips(_assistantButtonLoadDirectory, tr("Select a directory. All images/videos in this directory and sub directories will be scanned"));
    connect(_assistantButtonLoadDirectory, SIGNAL(clicked()), SLOT(_loadDirectory())    );

    _assistantButtonReviewFiles = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonReviewFiles->setText(tr("Review dates, times and albums"));
    MiscTools::setAllToolTips(_assistantButtonReviewFiles, tr("Review files for missing dates or albums. You can adapt all files' necessary meta data."));
    connect(_assistantButtonReviewFiles, SIGNAL(clicked()), SLOT(_on_reviewFiles_clicked()));


    _assistantButtonViewImages = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonViewImages->setText(tr("View media and choose destination targets"));
    MiscTools::setAllToolTips(_assistantButtonViewImages, tr("View all images/videos one by one. Select copy destinations for each image."));
    connect(_assistantButtonViewImages, SIGNAL(clicked()), SLOT(_on_viewImages_clicked()));


    _assistantButtonAdaptTimes = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonAdaptTimes->setText(tr("Adapt times from different cameras"));
    MiscTools::setAllToolTips(_assistantButtonAdaptTimes, tr("When using images/videos from different cameras, you can minimize existing time offsets of these."));
    connect(_assistantButtonAdaptTimes, SIGNAL(clicked()), SLOT(_on_adaptTimes_clicked()));

    _assistantButtonAlbums = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonAlbums->setText(tr("Manage albums"));
    _assistantButtonAlbums->setPixmap(":/images/assistant_no2.svg", _assistantButtonMargin);
    _assistantButtonAlbums->setStyleSheet("font:bold;");
    MiscTools::setAllToolTips(_assistantButtonAlbums, tr("Create albums and enter their descriptions. This is the single most important part of coping the image/video files."));
    connect(_assistantButtonAlbums, SIGNAL(clicked()), SLOT(_onManageAlbums()));

    _assistantButtonKeepDirectories = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonKeepDirectories->setText(tr("Different directories for different photographers"));
    MiscTools::setAllToolTips(_assistantButtonKeepDirectories, tr("If you want (for one or more destinations) to keep different subdirectories, e.g. from different photographers, use this dialog."));
    connect(_assistantButtonKeepDirectories, SIGNAL(clicked()), SLOT(_onSubDirMap()));

    //_assistantButtonSettings = new MiscTools::ImageButton(_assistantButtonMinHeight);
    //_assistantButtonSettings->setText(tr("Check (destination) settings"));
    //setAllToolTips(_assistantButtonSettings, tr("Make sure, the program (i.e. the copy destinations) is setup to you favor."));
    //connect(_assistantButtonSettings, SIGNAL(clicked()), &_settings, SLOT(show()));

    _assistantButtonPreview = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonPreview->setText(tr("Preview copy results"));
    _assistantButtonPreview->setPixmap(":/images/assistant_no3.svg", _assistantButtonMargin);
    _assistantButtonPreview->setStyleSheet("font:bold;");
    MiscTools::setAllToolTips(_assistantButtonPreview, tr("Preview the process of copying"));
    connect(_assistantButtonPreview, SIGNAL(clicked()), SLOT(on_actionPreview_copy_results_triggered()));

    _assistantButtonStart = new MiscTools::ImageButton(_assistantButtonMinHeight);
    _assistantButtonStart->setText(tr("Start copying"));
    _assistantButtonStart->setPixmap(":/images/assistant_no4.svg", _assistantButtonMargin);
    _assistantButtonStart->setStyleSheet("font:bold;");
    MiscTools::setAllToolTips(_assistantButtonStart, tr("Start the copying process. All files with an album and a given date and time will be copied."));
    connect(_assistantButtonStart, SIGNAL(clicked()), SLOT(on_actionStart_copying_triggered()));

    _assistantButtons = QList<MiscTools::ImageButton*>()
            << _assistantButtonLoadDirectory
            << _assistantButtonReviewFiles
            << _assistantButtonAdaptTimes
            << _assistantButtonViewImages
            << _assistantButtonAlbums
            << _assistantButtonKeepDirectories
            //<< _assistantButtonSettings
            << _assistantButtonPreview
            << _assistantButtonStart;

    //on_pushButton_assAdvanced_toggled(ui->pushButton_assAdvanced->isChecked());

    ui->horizontalLayout_loadDir->addWidget(_assistantButtonLoadDirectory);
    ui->gridLayout_review->addWidget(_assistantButtonReviewFiles, 0, 1);
    ui->gridLayout_review->addWidget(_assistantButtonAdaptTimes, 1, 1);
    ui->gridLayout_review->addWidget(_assistantButtonViewImages, 2, 1);
    ui->gridLayout_review->addItem(new QSpacerItem(60, 1, QSizePolicy::Fixed), 0, 2);
    ui->horizontalLayout_albums->addWidget(_assistantButtonAlbums);
    ui->gridLayout_settings->addWidget(_assistantButtonKeepDirectories, 0, 1);
    //ui->gridLayout_settings->addWidget(_assistantButtonSettings, 1, 1);
    ui->gridLayout_settings->addItem(new QSpacerItem(60, 1, QSizePolicy::Fixed), 0, 2);
    ui->horizontalLayout_preview->addWidget(_assistantButtonPreview);
    ui->horizontalLayout_start->addWidget(_assistantButtonStart);

    const bool checked = settings.value("assistant_advanced", false).toBool();
    ui->pushButton_assAdvanced->setChecked(checked);
    on_pushButton_assAdvanced_toggled(checked);

    _randomImageList = new RandomImageList();
    _randomImageList->setFrameShadow(QFrame::Raised);
    _randomImageList->setFrameShape(QFrame::Panel);
    _randomImageList->setVisible(false);
    ui->tab_assistant->layout()->addWidget(_randomImageList);

    _onSettingsDone();
    _resetAssistant();
    ui->tabWidget->setCurrentWidget(ui->tab_assistant);
    _currentTab = ui->tabWidget->currentWidget();

    QTimer::singleShot(WAIT_FOR_INFO_BOX, this, SLOT(_showInfoBox()));

    // check for update after two minutes
    QTimer::singleShot( WAIT_FOR_DONATE_HINT_MS, this, SLOT(on_actionSupport_by_donation_triggered()) );
    QTimer::singleShot( WAIT_FOR_CHECK_FOR_UPDATE, this, SLOT(on_actionCheck_for_new_version_triggered()) );
    //QTimer::singleShot(2000, this, SLOT(on_actionCheck_for_new_version_triggered()));

    restoreGeometryNow();
    //QTimer::singleShot(0, this, SLOT(restoreGeometryNow()));
}

void MainWindow::restoreGeometryNow()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    restoreGeometry(settings.value("main_geometry").toByteArray());
    restoreState(settings.value("main_state").toByteArray());
    this->show();
}

MainWindow::~MainWindow()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.setValue("main_geometry", saveGeometry());
    settings.setValue("main_state", saveState());
    settings.setValue("assistant_advanced", ui->pushButton_assAdvanced->isChecked());

    _saveAlbumDate(false);

    if (_loader != NULL) {
        _loader->stop();
        _loader->deleteLater();
        _loader = 0;
    }

    delete _viewSelected;
    delete _viewDir;
    delete _viewAll;
    _viewSelected = 0;
    _viewDir = 0;
    _viewAll = 0;

    delete ui;
}


void MainWindow::closeEvent(QCloseEvent *)
{
    ViewImageDialogExtended::closeInstance();
}


void MainWindow::_onNewAlbums()
{
    _delegate.setAlbums(_eventAssistant.getAlbums());
    _currentFiles.updateFilesAlbums(_eventAssistant.getAlbums());
}


void MainWindow::_onSettingsDone()
{
    qint32 l = _settings.getDestinationDescriptions().length();
    _delegate.setThumbnailSize(_settings.getThumbnailSize());
    ui->tableView_images->setColumnWidth(IMAGE_VIEW_PREVIEW_IDX, _settings.getThumbnailSize());
    //_randomImageList->setThumbnailsize(_settings.getThumbnailSize());

    for (int i=0; i<IMAGE_VIEW_NUM_COLS; i++) {
        ui->tableView_images->resizeColumnToContents(i);
    }
    ui->tableView_images->resizeColumnsToContents();
    ui->tableView_images->resizeRowsToContents();

    _currentFiles.setUseDifferentTags(_settings.getDifferentTags());
    _currentFiles.setUseComments(_settings.getUseComments());
    _currentFiles.setNumDestinations(l);
    _delegate.setUseComments(_settings.getUseComments());


    _autosaveTimer.start(_settings.getAutosaveInterval() * 60 * 1000); //minutes

    // update number of copy destinations
    ViewImageDialogExtended::instance(this)->setNumberOfDestinations(_settings.getDestinationDescriptions());
    ViewImageDialogExtended::instance(this)->setRawFormats(_settings.getRawExtensions());

    //_setDoneAssistant(_assistantButtonSettings);
}

void MainWindow::loadDirectory(const QString& imageDirOrig)
{
    QString imageDir = imageDirOrig;
    _infoBoxMode = _infoBoxMode <= 1 ? 1 : _infoBoxMode;

    const QString magicSolid = "/org/";
    if (imageDir.left(magicSolid.length()).toLower() == magicSolid
            && imageDir.contains("device", Qt::CaseInsensitive)) {
#ifdef FOUND_LIB_SOLID
        // the following block is mainly based on © digikam
        Solid::Device device(imageDir);

        if ( device.isValid() )
        {
            Solid::StorageAccess* access = device.as<Solid::StorageAccess>();

            if (!access)
            {
                qCritical() << "no browsing access to camera " << imageDir;
                QMessageBox::critical(this,
                                      tr("Camera Error"),
                                      tr("Camera %1 is not a browsing camera (in KDE that is).\n"
                                         "Until a proper support has been implemented"
                                         " copy the files to a directory on the hard drive"
                                         " or mount the camera onto a drive and manually"
                                         " have pivot read that directory our mount, respectively").arg(imageDir));
                return;
            }

            if (!access->isAccessible())
            {
                QApplication::setOverrideCursor(Qt::WaitCursor);

                if (!access->setup())
                {
                    qCritical() << "camera setup not successful: " << imageDir;
                    QMessageBox::critical(this,
                                          tr("Camera Error"),
                                          tr("Libsolid camera setup for %1 not successful").arg(imageDir));
                    QApplication::restoreOverrideCursor();
                    return;
                }

                QEventLoop* eventLoop = new QEventLoop(this);

                int returnCode = eventLoop->exec(QEventLoop::ExcludeUserInputEvents);

                delete eventLoop;
                eventLoop = 0;
                QApplication::restoreOverrideCursor();

                if (returnCode == 1)
                {
                    qCritical() << "camera returned an error with code " << returnCode;
                    QMessageBox::critical(this,
                                          tr("Camera Error"),
                                          tr("Camera setup returned an error: %1").arg(returnCode));
                    return;
                }
            }

            // Create Camera UI
            imageDir = access->filePath();

            qDebug() << "libsolid mounted camera to " << imageDir ;
        } else {
            qCritical() << "invalid camera device: " << imageDir;
            QMessageBox::critical(this,
                                  tr("Camera Error"),
                                  tr("Libsolid said %1 was an invalid device").arg(imageDir));
            return;
        }
#else
        QMessageBox::warning(this, tr("Libsolid error"),
                             tr("Can not read contents of the camera. "
                                "Pivot has been compiled without libsolid"
                                " - a KDE library. Please recompile and"
                                " make sure CMake finds libsolid."));
        return;
#endif //FOUND_LIB_SOLID
    }


    if (imageDir.trimmed().length() == 0 || !QDir(imageDir).exists()) {
        return;
    }    

    _saveAlbumDate(false);

    _imageDir = QFileInfo(imageDir).canonicalFilePath();

    ui->label_currDir->setText(_imageDir);

    _findImageFiles();

    ui->tableView_images->resizeColumnsToContents();
    ui->tableView_images->resizeRowsToContents();

    raise();
    activateWindow();

    if (_infoBoxMode == 1) {
        _infoBoxMode = 0;
        _showInfoBox();
    }
    _eventAssistant.resetLimitOutside();

    _resetAssistant();
    if (_currentFiles.getList().size() > 0) {
        _setDoneAssistant(_assistantButtonLoadDirectory);
    }
    _randomImageList->clearImages();
    _randomImageList->startAnimation();
    _randomImageList->setVisible(true);
    MiscTools::MiscTools::unlockDialog();
}

void MainWindow::_loadDirectory()
{
    MiscTools::MiscTools::lockDialog();
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    QString lastDirecotry = settings.value("lastDirectory").toString();

    QString imageDir = QFileDialog::getExistingDirectory(
                this,
                tr("Open Directory"),
                lastDirecotry,
                QFileDialog::ShowDirsOnly);

    if (imageDir.length() == 0) {
        MiscTools::MiscTools::unlockDialog();
        return;
    }

    settings.setValue("lastDirectory", imageDir);

    ui->actionRescan_current_directory->setEnabled(true);
    loadDirectory(imageDir);

}

void MainWindow::_findImageFiles()
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    // find alreadyCopied status
    _connectToLocalDB(_alreadyCopiedLocation, CREATE_ALREADY_COPIED, &_db);
    QList<ImageDataPtr> newList = Tools::findFiles(_imageDir,
                                               _settings.getCopyOnlyOnce(),
                                               _db);
    _disconnectDB(&_db);

    QList<ImageDataPtr> discarded;
    QList<QDir> dirs;

    // load saved albums / times
    if (_settings.getSaveAlbumDate()) {

        _connectToLocalDB(_savedAlbumDateLocation, CREATE_SAVED_ITEMS, &_dbFotoSettings);
        // it seems a lot faster to obtain all data at once, and search
        // rather than to query every single file
        QSqlTableModel model(0, *_dbFotoSettings);
        model.setEditStrategy(QSqlTableModel::OnManualSubmit);
        model.setTable(DATABASE_DATE_SETTINGS);
        model.select();

        _disconnectDB(&_dbFotoSettings);

        // using a map will save iterations, since
        // once encountered and updated files won't
        // be searched again
        QVector<int> map(newList.size()) ;
        for (int i=0; i<(int)map.size(); i++){
            map[i]=i;
        }

        ProgressDialog* progress = new ProgressDialog();
        progress->setModal(true);
        progress->setMinMax(1, model.rowCount());
        progress->setValue(1);
        progress->show();
        progress->setLabel(tr("checking database for saved manual data... ") );

        // for every db entry search for corresponding file
        // since we can remove found files, this might be the
        // faster option (still O^2)
        // Todo: use heap or sorted lists to improve searching ?!
        for (int j=0; j<(int) model.rowCount(); j++) {
            if (progress->checkUpdateWanted()) {
                progress->setValue(j);
            }
            if (progress->isCanceled()) {
                newList.clear();
                break;
            }
            QSqlRecord record = model.record(j);
            const QString filename = record.value("filename").toString();
            //qDebug() << filename << static_cast<ImageDataExtended*>(newList[map[0]];
            // use the map to iterate over remaining files
            for (int i=(int)map.size()-1; i>=0; i--) {
                ImageDataExtended* tmp = static_cast<ImageDataExtended*>(newList[map[i]].get());
                if (tmp->getFileInfo().canonicalFilePath() == filename) {
                    map.remove(i);
                    const QString time = record.value("datetime").toString();
                    int albumId = record.value("albumid").toInt();
                    bool knownDay = record.value("dateknown").toInt() > 0;
                    quint64 useDestinations = record.value("useCopyDestination").toUInt();


                    if (time.length() > 0) {
                        QDateTime savedDate = QDateTime::fromString(time, DATE_TIME_CACHE_FORMAT);
                        tmp->setManualTime(savedDate);
                    }

                    tmp->setDayIsKnown(knownDay);

                    // is it automated album or not ?
                    bool isManualAlbum = false;

                    for (int j=0; j<(int)_eventAssistant.getAlbums().size(); j++) {
                        if (tmp->getTime().isValid()
                                && _eventAssistant.getAlbums()[j].from <= tmp->getTime()
                                && _eventAssistant.getAlbums()[j].to > tmp->getTime()) {
                            isManualAlbum =  _eventAssistant.getAlbums()[j].ID != albumId;
                        }
                    }

                    tmp->setAlbumID(albumId, isManualAlbum);

                    // set copy status
                    tmp->setCopyDestinationStatusAll(useDestinations);

                    // already saved in DB, so do not resave later on
                    tmp->setChangedStatus(false);
                    break;
                }
            }
        }
        delete progress;
        progress = 0;
    }

    newList = Tools::sortItems(newList,
                               _settings.getExtensions(),                               
                               _settings.tryUseMetaData(),
                               _settings.tryToObtainByBasename(),
                               _settings.tryToObtainByNeighbour(),
                               _settings.tryToObtainFileTime(),
                               discarded,
                               dirs);


    // find directories
    _subDirectories.clear();
    for (int i=0; i<(int)newList.size(); i++) {
        const QString path = newList[i]->getFileInfo().canonicalPath();
        const QString subDir = path.length() - 1 - _imageDir.length() < 0 ? "." : path.right(path.length() - 1 - _imageDir.length());
        if (!_subDirectories.contains(subDir)) {
            _subDirectories.push_back(subDir);
        }

    }

    _currentFiles.setRootDirectory(_imageDir);
    _currentFiles.setList(newList);
    _delegate.setAlbums(_eventAssistant.getAlbums());
    _currentFiles.updateFilesAlbums(_eventAssistant.getAlbums());

    if (_loader != NULL) {
        _loader->stop();
        delete _loader;
        _loader = 0;
    }
    _loader = new ThumbnailLoader;
    _loader->setLocationPath(_databaseLocation);

    connect(_loader, SIGNAL(loadedThumbnails(const QFileInfo&, const MiscTools::pQImage&)),
                    SLOT(onloadedThumbnails(const QFileInfo&, const MiscTools::pQImage&)),
            Qt::QueuedConnection
                    );
    connect(_loader, SIGNAL(errorLoadingThumb(const QFileInfo&)),
                    SLOT(onErrorThumb(const QFileInfo&)),
            Qt::QueuedConnection
                    );

    connect(_loader, SIGNAL(loadedThumbnails(QFileInfo, MiscTools::pQImage)),
            &_eventAssistant, SLOT(onLoadedThumbnails(QFileInfo, MiscTools::pQImage)),
            Qt::QueuedConnection);
    connect(_loader, SIGNAL(errorLoadingThumb(QFileInfo)),
            &_eventAssistant, SLOT(onErrorThumbnails(QFileInfo)),
            Qt::QueuedConnection);

    connect(_loader, SIGNAL(doneFast()), SLOT(onDoneFastThumbnails()),
            Qt::QueuedConnection);

    _loader->computeThumbnails(_currentFiles.getList(), true);

    qApp->restoreOverrideCursor();
}

void MainWindow::onloadedThumbnails(const QFileInfo& file,
                                    const MiscTools::pQImage& thumbnail)
{
    // now search corresponding listwidget item
    const QList<ImageDataPtr>& list = _currentFiles.getList();
    for (int i=0; i < (int)list.size(); i++) {
        // for some reason, the list seems to change while in this method
        if (i>=list.size() || i<0) {
            continue;
        }
        QFileInfo listItemPath(list.at(i)->getFileInfo());
        if (file.canonicalFilePath() == listItemPath.canonicalFilePath()) {
            ImageDataExtended tmp = _currentFiles.getImageDataExtended(i);
            tmp.setHasThumbnail( true );
            tmp.setThumbnail(thumbnail);
            _currentFiles.updateList(i, tmp);
            _adaptTimeAssistant.newThumbnail(file.canonicalFilePath());
            ui->tableView_images->resizeRowToContents(i);
            _randomImageList->addImage(list.at(i));
        }
    }

}

void MainWindow::onErrorThumb(const QFileInfo& file)
{
    // now search corresponding listwidget item
    const QList<ImageDataPtr>& list = _currentFiles.getList();
    for (int i=0; i < (int)list.size(); i++) {
        if (i>=list.size() || i<0) {
            continue;
        }
        QFileInfo listItemPath(list.at(i)->getFileInfo());
        if (file.canonicalFilePath() == listItemPath.canonicalFilePath()) {
            // qWarning() << "error thumb " << file.canonicalFilePath();
            ImageDataExtended tmp = _currentFiles.getImageDataExtended(i);
            tmp.setHasThumbnail(true);
            tmp.setThumbnail(QImage());
            _currentFiles.updateList(i, tmp);
            _adaptTimeAssistant.newThumbnail(file.canonicalFilePath());
            break;
        }
    }
}

void MainWindow::onDoneFastThumbnails()
{
    QList<ImageDataPtr> notYetComputed;
    // now search corresponding listwidget item
    for (int i=0; i < (int)_currentFiles.getList().count(); i++) {
        if (!_currentFiles.getList().at(i)->hasThumbnail()) {
            notYetComputed.push_back(_currentFiles.getList().at(i));
        }
    }

    if (notYetComputed.count() > 0) {
        _loader->computeThumbnails(notYetComputed, false);
    }
    statusBar()->showMessage(tr("done extracting META-DATA thumbnails. Now rescaling remaining files"), 5);
}

static inline QVector<int> getSortedSelected(QTableView* view, QString error,
                                             QWidget* parent,
                                             int minRows=2,
                                             bool clear=false)
{
    QModelIndexList indexes = view->selectionModel()->selectedIndexes();
    QModelIndex index;

    QVector<int> rows;
    foreach(index, indexes) {
        if (!rows.contains(index.row())) {
            rows.push_back(index.row());
        }
    }

    qSort(rows.begin(), rows.end());


    if (rows.size() < minRows) {
        QMessageBox::warning ( parent,
                               qApp->translate("not enough files title","Not enough files selected"),
                               error);
        return QVector<int>();
    }


    if (clear) view->clearSelection();
    return rows;
}



void MainWindow::onErrorMessage(const QString& headline, const QString& errorMsg)
{
    QMessageBox::warning ( this, headline, errorMsg);
}

void MainWindow::_onManageAlbums()
{
    _eventAssistant.setImageList(_currentFiles.getList(),
                                 _imageDir.length() + 1);
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_eventAssistant, SLOT(showInfo()));
    ui->tabWidget->setCurrentWidget(ui->tab_album);
}


QList<ImageDataPtr> MainWindow::_getTimeSorted(int& currIndex,
                                               const QList<ImageDataPtr>& currentList,
                                               QVector<int> indexes)
{
    if (indexes.size() == 0) {
        indexes.resize(currentList.size());
        for (int i=0; i<indexes.size(); i++) {
            indexes[i] = i;
        }
    }
    int indOut = -1;
    QList<ImageDataPtr> out;
    // get sorted list by time and update the current index to match the same image in the new list
    QVector<int> indexesSortedTime = _currentFiles.getImageTimeSortIndexes(indexes);
    for (int i=0; i<indexesSortedTime.size(); i++) {
        out.push_back(currentList[indexesSortedTime[i]]);
        if (currIndex > 0 && currIndex < indexes.size() && indexes[currIndex] == indexes[indexesSortedTime[i]]) {
            indOut = i;
        }
    }
    currIndex = indOut;
    return out;
}

void MainWindow::_onViewAll()
{
    int row = (sender() == _assistantButtonViewImages) ? -1 : _lastRow;
    ViewImageDialogExtended * viewImages = ViewImageDialogExtended::instance(this);
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();

    QList<ImageDataPtr> sorted = _getTimeSorted(row, _currentFiles.getList());

    viewImages->setImages(sorted,
                          row,
                          _imageDir.length() + 1);
}

void MainWindow::_onViewSelected()
{
    const int row = _lastRow;
    const QList<ImageDataPtr> l = _currentFiles.getList();

    QList<ImageDataPtr> showImages;
    QVector<int> rows = getSortedSelected(ui->tableView_images, "", this, 0, false);

    if (row >=0 && !rows.contains(row)) {
        rows.push_back(row);
        qSort(rows);
    }

    int useStart = 0;
    for (int i=0; i<(int)rows.size(); i++) {
        if (rows[i] == row) {
            useStart = i;
        }
        showImages.push_back(_currentFiles.getList()[rows[i]]);
    }

    QList<ImageDataPtr> sorted = _getTimeSorted(useStart, showImages, rows);

    ViewImageDialogExtended * viewImages = ViewImageDialogExtended::instance(this);
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();
    viewImages->setImages(sorted,
                          useStart,
                          _imageDir.length() + 1);
}

static bool _compareByFilename(const ImageDataPtr &s1, const ImageDataPtr &s2)
{
    return s1->getFileInfo().canonicalFilePath() < s2->getFileInfo().canonicalFilePath();
}


void MainWindow::_onViewDir()
{
    const int row = _lastRow;
    const QList<ImageDataPtr> l = _currentFiles.getList();
    QList<ImageDataPtr> showImages;

    int useStart = 0;
    // removed: show images of all selected dirs, if given
    // otherwise all images from row underneath
//    QVector<int> rows = getSortedSelected(ui->tableView_images, "", this, 0, false);
//    if (rows.size() > 0) {
//        // get all selected images
//        for (int i=0; i<(int)rows.size(); i++) {
//            if (rows[i] == row) {
//                useStart = i;
//            }
//            showImages.push_back(l[rows[i]]);
//        }
//    } else {
//        showImages.push_back(l[row]);
//    }

    if (row >=0)
        showImages.push_back(l[row]);
    const ImageDataPtr ref = showImages[useStart];
    QVector<int> samedirIndexes;

    for (int i=0; i<(int)l.size(); i++) {
        bool sameDir = false;
        for (int j=0; j<(int)showImages.size(); j++) {
            if (l[i]->getFileInfo().canonicalPath() == showImages[j]->getFileInfo().canonicalPath()) {
               sameDir = true;
               break;
            }
        }
        if (sameDir) {
            samedirIndexes.push_back(i);
            showImages.push_back(_currentFiles.getList()[i]);
        }
    }

    // sort by directory
    qSort(showImages.begin(), showImages.end(), _compareByFilename);

    // refind index
    for (int i=0; i < (int)showImages.size(); i++) {
        if (showImages[i]->getFileInfo().canonicalFilePath()
                == ref->getFileInfo().canonicalFilePath()) {
            useStart = i;
            break;
        }
    }

    ViewImageDialogExtended * viewImages = ViewImageDialogExtended::instance(this);
    viewImages->setModal(false);
    viewImages->show();
    viewImages->activateWindow();
    viewImages->setImages(showImages,
                          useStart,
                          _imageDir.length() + 1);
}

void MainWindow::_onSubDirMap()
{
    _directoryAssistant.setDirectories(_subDirectories);
    ui->tabWidget->setCurrentWidget(ui->tab_keepdirectory);
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_directoryAssistant, SLOT(showInfo()));
}

void MainWindow::_onAdaptTimes()
{    
    _adaptTimeAssistant.setDirectories(_subDirectories,
                                       _currentFiles.getList(),
                                       _imageDir.length() + 1);
    _lastImageDirUsed = _imageDir;
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_adaptTimeAssistant, SLOT(showInfo()));
}

void MainWindow::_doneDirectoryAssistant()
{
    _setDoneAssistant(_assistantButtonKeepDirectories );
}

void MainWindow::_getEditMenu(QMenu** menu,
                              const bool showView)
{
    *menu = new QMenu(this);
    (*menu)->addAction(_viewAll);
    if (showView) {
        (*menu)->addAction(_viewSelected);
        (*menu)->addAction(_viewDir);
    }
    (*menu)->addSeparator();
    (*menu)->addAction(_albumReset);
    (*menu)->addAction(_albumDown);
    (*menu)->addAction(_albumUp);
    (*menu)->addSeparator();
    (*menu)->addAction(_timeReset);
    (*menu)->addAction(_timeDown);
    (*menu)->addAction(_timeUp);
    (*menu)->addSeparator();
    (*menu)->addAction(_destinationsReset);
    (*menu)->addAction(_destinationsDown);
    (*menu)->addAction(_destinationsUp);
    (*menu)->addSeparator();
    (*menu)->addAction(_checkAll);
    (*menu)->addAction(_uncheckAll);
    //(*menu)->addAction(_invertCheck);
    (*menu)->addSeparator();
    (*menu)->addAction(_selectAll);
    (*menu)->addAction(_deselectAll);
    (*menu)->addSeparator();
    (*menu)->addAction(_toggleError);
}


void MainWindow::_contextMenuEvent(const QPoint &pos)
{
    QMenu* menu;
    MainWindow::_getEditMenu(&menu, true);
    _lastRow = ui->tableView_images->rowAt(pos.y());
    menu->exec(ui->tableView_images->viewport()->mapToGlobal(pos));
    menu->close();
    delete menu;
    menu = 0;
}

void MainWindow::_doneAlbums()
{
    _delegate.setAlbums(_eventAssistant.getAlbums());
    _currentFiles.updateFilesAlbums(_eventAssistant.getAlbums());
    //ui->tabWidget->setCurrentIndex(0);
    //ui->tab_assistant->setFocus();
    _setDoneAssistant(_assistantButtonAlbums );
}

namespace
{

QString _addHeader(const QString headline)
{
    QString out = "#######################################################################\n"
            + QString("# ")
            + headline
//            + QString(" ").repeated( (72 - headline.length() )-2 > 0 ? (72 - headline.length()) -2 : 0) + QString("#")
            + QString("\n")
            +    "#######################################################################\n\n";
    return out;
}

}

HandleExistingItems MainWindow::_getExistingFilesChoice()
{
    HandleExistingItems existingFiles = HandleExistingItems_PerformAction;
    if (_settings.getExistingFilesMerge()) {
        existingFiles = HandleExistingItems_Merge;
    } else if (_settings.getExistingFilesContinue()) {
        existingFiles = HandleExistingItems_Continue;
    } else if (_settings.getExistingFilesContinueRename()) {
        existingFiles = HandleExistingItems_ContinueAndRenameExisting;
    }
    return existingFiles;
}


void MainWindow::on_actionPreview_copy_results_triggered()
{
    if (sender() == _assistantButtonPreview) {
        ui->tabWidget->setCurrentWidget(ui->tab_preview);
        return;
    }
    MiscTools::MiscTools::lockDialog();

    QList<ImageDataExtended*> noAlbumError;
    QList<ImageDataExtended*> noTimeError;
    QList<ImageDataExtended*> removeUnique;
    QList<ImageDataExtended*> alreadyCopied;
    QList<ImageDataExtended*> notSelected;

    QList<ImageDataPtr> listCopy = _currentFiles.getList();

    _connectToLocalDB(_alreadyCopiedLocation, CREATE_ALREADY_COPIED, &_db);

    QList<QPair<QString, QString> > moveExistingFiles;
    QVector<CopyItem> items = CopyFiles::applyForAllFiles(
                _settings.getDestinationDescriptions(),
                _imageDir,
                _eventAssistant.getAlbums(),
                _directoryAssistant.getDirectoryMap(),
                listCopy,
                _settings.getCombineRAWAndImage(),
                _settings.getUseUniqueFiles(),
                _settings.getCopyOnlyOnce(),
                _db,
                _settings.getTagsKnownDayAlbum(),
                _settings.getTagsKnownDay(),
                _settings.getTagsUnknownDayAlbum(),
                _settings.getTagsUnknownDay(),
                _settings.getForceSameNumber(),
                _getExistingFilesChoice(),
                moveExistingFiles,
                noAlbumError,
                noTimeError,
                removeUnique,
                alreadyCopied,
                notSelected);

    _disconnectDB(&_db);

    QString plaintext;

    bool noErrs = true;
    if (noAlbumError.size() >0
            || noTimeError.size() > 0
            || removeUnique.size() > 0
            || alreadyCopied.size() > 0
            || notSelected.size() > 0)
    {
        plaintext +=
                tr("WARNING: There are files which won't be copied, please see below \n\n");
        noErrs = false;
    }

    if (items.size() == 0 && noErrs) {
        plaintext += tr("no output generated (cancelled ?)");
    } else if (items.size() > 0) {
        if (moveExistingFiles.size() > 0) {
            const QVector<int> order = MiscTools::FilenameNumbering::findMoveOrdering(moveExistingFiles);
            if (order.size() != moveExistingFiles.size()) {
                qCritical("Could not determine movement order for preview, circular dependency ?");
                plaintext += _addHeader(tr("Critical: could not determine an order to move existing files"
                                           "- possibly due to a circular dependency ?"));
            } else {
                plaintext += _addHeader(tr("The following files, already existing in the destination folder(s), will be moved:"));
                plaintext += _addHeader(tr("(The order displayed here will avoid temporary clashes)"));
                for (int i=0; i<moveExistingFiles.size(); i++) {
                    const QString targetName = QFileInfo(moveExistingFiles[order[i]].second).fileName();
                    plaintext += moveExistingFiles[order[i]].first + tr(" -> ", "move source to target") + targetName;
                }
            }
        }
        QStringList targetFileNames;
        for (int i=0; i<(int)items[0].destinations.size(); i++) {
            for (int j=0; j<(int)items.size(); j++) {
                if (! items[j].image->getCopyDestinationStatus(i)) {
                    // targt was specifially and manually chosen not to be copied to
                    continue;
                }

                targetFileNames.push_back(items[j].destinations[i].first
                                          + tr("           (from '%1')").arg(items[j].image->getFileInfo().canonicalFilePath().right(items[j].image->getFileInfo().canonicalFilePath().length() -1 -_imageDir.length())));
            }
        }
        qStableSort(targetFileNames);

        plaintext += _addHeader(tr("The following files will be created (by copying):"));

        plaintext += targetFileNames.join(QString("\n"));
    }

    if (notSelected.size() > 0) {
        plaintext += "\n\n" + _addHeader(tr("The following files were not selected to be copied:"));
        for (int i=0; i<(int)notSelected.size(); i++) {
            plaintext += notSelected[i]->getFileInfo().canonicalFilePath().right(notSelected[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
        }
    }


    if (noTimeError.size() > 0) {
        plaintext += "\n\n" + _addHeader(tr("The following files don't have date/time assignment:"));
        for (int i=0; i<(int)noTimeError.size(); i++) {
            plaintext += noTimeError[i]->getFileInfo().canonicalFilePath().right(noTimeError[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
        }
    }

    if (noAlbumError.size() > 0) {
        plaintext += "\n\n" + _addHeader(tr("The following files don't have an album associated:"));
        for (int i=0; i<(int)noAlbumError.size(); i++) {
            plaintext += noAlbumError[i]->getFileInfo().canonicalFilePath().right(noAlbumError[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");
        }
    }


    if (removeUnique.size() > 0) {
        plaintext += "\n\n" + _addHeader(tr("The following files were not considered due to multiple appearance:"));
        for (int i=0; i<(int)removeUnique.size(); i++) {
            plaintext += removeUnique[i]->getFileInfo().canonicalFilePath().right(removeUnique[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
        }
    }

    if (alreadyCopied.size() > 0) {
        plaintext += "\n\n" + _addHeader(tr("The following files were already copied, in an earlier run:"));
        for (int i=0; i<(int)alreadyCopied.size(); i++) {
            plaintext += alreadyCopied[i]->getFileInfo().canonicalFilePath().right(alreadyCopied[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
        }
    }

    _preview.setText(plaintext);
    _setDoneAssistant(_assistantButtonPreview);
    MiscTools::MiscTools::unlockDialog();
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_preview, SLOT(showInfo()));

}

void MainWindow::on_actionStart_copying_triggered()
{
    MiscTools::MiscTools::lockDialog();

    QList<ImageDataExtended*> noAlbumError;
    QList<ImageDataExtended*> noTimeError;
    QList<ImageDataExtended*> removeUnique;
    QList<ImageDataExtended*> alreadyCopied;
    QList<ImageDataExtended*> notSelected;
    QList<ImageDataExtended*> error;


    QList<QPair<QString, QString> > moveExistingFiles;
    bool allSuccess = true;
    {
        _connectToLocalDB(_alreadyCopiedLocation, CREATE_ALREADY_COPIED,  &_db);

        // this lock causes thumbnail-loading threads to wait
        // (right after the current thumbnail is finished
        QMutexLocker l(&ThumbnailLoader::pauseMutex);

        allSuccess = CopyFiles::doCopy(
                    _settings.getDestinationDescriptions(),
                    _imageDir,
                    _eventAssistant.getAlbums(),
                    _directoryAssistant.getDirectoryMap(),
                    _currentFiles.getList(),
                    _settings.getCombineRAWAndImage(),
                    _settings.getUseUniqueFiles(),
                    _settings.tryUpdateTimeMetaData(),
                    _settings.saveToTextFileIfFailure(),
                    _settings.getCopyOnlyOnce(),
                    _settings.getCreateOwnerShipText(),
                    _db,
                    _settings.getTagsKnownDayAlbum(),
                    _settings.getTagsKnownDay(),
                    _settings.getTagsUnknownDayAlbum(),
                    _settings.getTagsUnknownDay(),
                    _settings.getOverwriteMode(),
                    _settings.getForceSameNumber(),
                    _getExistingFilesChoice(),
                    moveExistingFiles,
                    this,
                    noAlbumError,
                    noTimeError,
                    removeUnique,
                    alreadyCopied,
                    notSelected,
                    error);

        _disconnectDB(&_db);
    }

    qDebug() << "sucessfully copied: " << allSuccess;
    if (!allSuccess) {
        QMessageBox::warning(this,
                                 tr("Aborted"),
                                 tr("Some files have been copied. Process has been aborted."));
    }
    if ( (!allSuccess && moveExistingFiles.size() > 0)
         || (allSuccess && (
                 noAlbumError.size() >0
                 || noTimeError.size() > 0
                 || removeUnique.size() > 0
                 || alreadyCopied.size() > 0
                 || notSelected.size() > 0
                 || error.size() > 0
                 )
             )
         )
    {
        QString plaintext;

        if (moveExistingFiles.size() > 0) {
            plaintext +=
                    tr("WARNING: There were files existing in the destination folder, "
                       "which couldn't have been moved, please see below \n\n");
            for (int i=0; i<(int)moveExistingFiles.size(); i++) {
                plaintext += moveExistingFiles[i].first + QString(" -> ") + moveExistingFiles[i].second + QString("\n");;
            }
        }

        plaintext +=
                tr("WARNING: There are files which haven't been copied, please see below \n\n");

        if (error.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files returned an error on copying:"));
            for (int i=0; i<(int)error.size(); i++) {
                //plaintext += error[i]->getFileInfo().canonicalFilePath().right(error[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
                plaintext += error[i]->getFileInfo().canonicalFilePath() + QString("\n");;
            }
        }


        if (notSelected.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files were not selected to be copied:"));
            for (int i=0; i<(int)notSelected.size(); i++) {
                //plaintext += notSelected[i]->getFileInfo().canonicalFilePath().right(notSelected[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
                plaintext += notSelected[i]->getFileInfo().canonicalFilePath() + QString("\n");;
            }
        }

        if (noAlbumError.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files don't have an album associated:"));
            for (int i=0; i<(int)noAlbumError.size(); i++) {
                //plaintext += noAlbumError[i]->getFileInfo().canonicalFilePath().right(noAlbumError[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");
                plaintext += noAlbumError[i]->getFileInfo().canonicalFilePath() + QString("\n");
            }
        }

        if (noTimeError.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files don't have date/time assignment:"));
            for (int i=0; i<(int)noTimeError.size(); i++) {
                //plaintext += noTimeError[i]->getFileInfo().canonicalFilePath().right(noTimeError[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
                plaintext += noTimeError[i]->getFileInfo().canonicalFilePath() + QString("\n");;
            }
        }

        if (removeUnique.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files were not considered due to multiple appearance:"));
            for (int i=0; i<(int)removeUnique.size(); i++) {
                //plaintext += removeUnique[i]->getFileInfo().canonicalFilePath().right(removeUnique[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
                plaintext += removeUnique[i]->getFileInfo().canonicalFilePath() + QString("\n");;
            }
        }

        if (alreadyCopied.size() > 0) {
            plaintext += "\n\n" + _addHeader(tr("The following files were already copied, in an earlier run:"));
            for (int i=0; i<(int)alreadyCopied.size(); i++) {
                //plaintext += alreadyCopied[i]->getFileInfo().canonicalFilePath().right(alreadyCopied[i]->getFileInfo().canonicalFilePath().length()-1-_imageDir.length()) + QString("\n");;
                plaintext += alreadyCopied[i]->getFileInfo().canonicalFilePath() + QString("\n");;
            }
        }

        _preview.setText(plaintext);
		QMessageBox::information(this,
			tr("Files have been copied"),
			tr("Files have successfully been copied. Some where left out of the copying process - see the preview tab"));

		_setDoneAssistant(_assistantButtonStart);
    } else if (allSuccess) {
        QMessageBox::information(this,
                                 tr("All files been copied"),
                                 tr("All files have successfully been copied."));

        _setDoneAssistant(_assistantButtonStart);
    }
    MiscTools::MiscTools::unlockDialog();
}


bool MainWindow::_connectToLocalDB(const QString& locationPath,
                                   const QString& createString,
                                   QSqlDatabase** db)
{
    QMutexLocker l(&FastImageThumbnail::databaseMutex);
    QSqlDatabase::removeDatabase("qt_sql_default_connection");
    delete *db;
    *db = new QSqlDatabase;
    **db = QSqlDatabase::addDatabase("QSQLITE");
    (*db)->setDatabaseName(locationPath);
    if (!(*db)->open()) {
        // maybe the cache is corrupted
        if (QDir().exists(locationPath)) {
             qWarning() << "could not connect to db - now deleting file and retrying";
            QFile::remove(locationPath);
        }
        if (!(*db)->open()) {
            qCritical() << "failed to open DB connection for copied files";
            return false;
        }
    }

    {
        QSqlQuery query(**db);
        QString queryStr = createString;



        if (!query.exec(queryStr)) {
            qCritical() << "could not insert table";
        }
    }

    return true;
}


void MainWindow::_disconnectDB(QSqlDatabase** db)
{
    //db.removeDatabase();
    QMutexLocker l(&FastImageThumbnail::databaseMutex);
    if (db == NULL) return;
    (*db)->close();
    delete *db;
    *db = NULL;
    QSqlDatabase::removeDatabase("QSQLITE");
    QSqlDatabase::removeDatabase("qt_sql_default_connection");
}

void MainWindow::on_actionAbout_Qt_triggered()
{
    qApp->aboutQt();
}

void MainWindow::on_actionAbout_libraw_triggered()
{
    QMessageBox::information(this,
                             tr("About libraw"),
                             tr("This software uses the libraw open source image library. "
                                "See <a href=\"http://www.libraw.org/\">http://www.libraw.org/</a> for details. "
                                "L is used under the GPL, version 2.1."));
}

void MainWindow::on_actionAbout_libexif_triggered()
{
    QMessageBox::information(this,
                             tr("About libexif"),
                             tr("This software uses the libexif open source EXIF library. "
                                "See <a href=\"http://libexif.sourceforge.net/\">http://libexif.sourceforge.net/</a> for details. "
                                "libexif is used under the LGPL, version 2.1."));
}

void MainWindow::on_actionAbout_libjpg_triggered()
{
    QMessageBox::information(this,
                             tr("About libjpg"),
                             tr("This software uses the libjpg open source jpg library. "
                                "See <a href=\"http://http://www.ijg.org/\">http://http://www.ijg.org/</a> for details. "));
}

void MainWindow::on_actionAbout_pivot_triggered()
{
    _about.setModal(true);
    _about.setVersion(version_pivot.getVersionMajor(),
                      version_pivot.getVersionMinor(),
                      version_pivot.getVersionPatch(),
                      version_pivot.getCompilerBits());
    _about.show();
}

void MainWindow::_resetSettings()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.clear();

    _delegate.setThumbnailSize(_settings.getThumbnailSize());
    restoreGeometry(settings.value("main_geometry").toByteArray());
    restoreState(settings.value("main_state").toByteArray());

    // reset all settings

    _adaptTimeAssistant.loadSettings();
    _eventAssistant.loadSettings();
    _directoryAssistant.loadSettings();
    ViewImageDialogExtended * viewImages = ViewImageDialogExtended::instance(this);
    viewImages->resetSettings();
    viewImages->loadSettings();
}


void MainWindow::_removeUserData()
{
    _eventAssistant.resetData(_settingsLocation);
    QFile::remove(_alreadyCopiedLocation);
    if (_loader != NULL) {
        _loader->stop();
        delete _loader;
        _loader = 0;
    }
    QFile::remove(_databaseLocation);
    QFile::remove(_savedAlbumDateLocation);
}

void MainWindow::_saveAlbumDate(const bool dispMessage)
{
    if (sender() == &_autosaveTimer && !_settings.getDoAutosave()) {

        qDebug() << "autosaving not desired";
        return;
    }

    qDebug() << "saving album and image data";

    const int n = _currentFiles.getList().size();

    if (_settings.getSaveAlbumDate()) {        
        statusBar()->showMessage(tr("saving manually changed image data... please wait"), 5000);

        _connectToLocalDB(_savedAlbumDateLocation, CREATE_SAVED_ITEMS, &_dbFotoSettings);

        // first of all retrieve all data sets
        QSqlTableModel model(0, *_dbFotoSettings);
        typedef QPair<int, ImageDataExtended*> Pairt;
        QVector <Pairt> changes;
        model.setEditStrategy(QSqlTableModel::OnManualSubmit);
        model.setTable(DATABASE_DATE_SETTINGS);
        model.select();
        int numNew = 0;

        _dbFotoSettings->transaction(); // for fast writing

        // iterate to find already existing files
        for (int i=0; i < (int) n; i++) {
            ImageDataExtended* tmp = static_cast<ImageDataExtended*>(_currentFiles.getList()[i].get());
            if (!tmp->hasChanged()) {
                continue;
            }

            bool copyStatus = true;
            for (int j=0; j< MAX_NUM_COPY_DESTINATIONS; j++) {
                copyStatus &= tmp->getCopyDestinationStatus(j);
            }

            int alreadyInDB = -1;
            for (int j=0; j<model.rowCount(); j++) {
                QSqlRecord record = model.record(j);
                QString filename = record.value("filename").toString();
                if (filename == tmp->getFileInfo().canonicalFilePath()) {
                    alreadyInDB = j;
                    break;
                }
            }

            /* replace by haschanged behaviour
             *
            if (!tmp->getIsManualAlbum()
                    && !tmp->getHasComment()
                    && !tmp->getIsManualTime()
                    && !tmp->getIsShiftedTime()
                    && tmp->getDayIsKnown()
                    && copyStatus) {
                // if no data is known and no db relevant entries
                // continue working
                if (alreadyInDB < 0)
                    continue;
            }*/

            // does file have an existing row in database ?
            if (alreadyInDB < 0 ) {
                // code new entries negatively, e.g. 5th new > -6
                alreadyInDB = (-numNew) - 1;
                numNew++;
            }

            // push_back files with differences from automatically determined
            // settings
            changes.push_back(Pairt(alreadyInDB, tmp));
        }

        qDebug() << "saving changes: " << changes.size();


        if (numNew > 0) {
            model.insertRows(0, numNew);
        }

        // iterate over collected items with changes or new data
        for (int i=0; i<changes.size(); i++) {
            const int id = changes[i].first;
            const ImageDataExtended* tmp = changes[i].second;

            quint64 useDestinations = tmp->getCopyDestinationStatusAll();

            // new indexes are coded negative
            // so check and compute correct index
            // also beware of index shifts for existing rows
            const bool isNew = id < 0;
            const int ind = isNew ? (-id) -1  : id + numNew;

            model.setData(model.index(ind, 0), tmp->getFileInfo().canonicalFilePath());
            model.setData(model.index(ind, 1), tmp->getAlbumID());
            model.setData(model.index(ind, 2), (tmp->getIsManualTime() || tmp->getIsShiftedTime())
                          ? tmp->getTime().toString(DATE_TIME_CACHE_FORMAT) : ""
                                                 );
            model.setData(model.index(ind, 3), (tmp->getDayIsKnown() ? 1 : 0));
            model.setData(model.index(ind, 4), (tmp->getComment()));

            model.setData(model.index(ind, 5), useDestinations);
        }

        // submit and commit
        bool success = true;
        if (model.submitAll()) {
            if (_dbFotoSettings->commit()) { // for fast writing
                // reset changed flag to NOT changed
                for (int i=0; i<changes.size(); i++) {
                    changes[i].second->setChangedStatus(false);
                }
            } else {
                success = false;
            }
        } else {
            success = false;
        }
        if (!success) {
            statusBar()->showMessage(tr("Could NOT save manual changes - DB Error ?"), 15000);
            qCritical() << "could not save manual changes - db error ?" << model.lastError().text();
            if (dispMessage && sender() != &_autosaveTimer) {
                QMessageBox::warning ( this,
                                       tr("Error saving data"),
                                       tr("Error, could not save manual data changes to database:\n"
                                          " %1").arg(model.lastError().text()));
            }
        }

        _disconnectDB(&_dbFotoSettings);
    } else {
        const QString text = tr("I'm NOT saving data, since this"
                                " has been deactivated in the settings."
                                " This holds for autosaving, as well, even"
                                " if activated.");
        if (dispMessage && sender() != &_autosaveTimer) {
            QMessageBox::warning ( this,
                                   tr("Not saving data"),
                                   text);
        }
        statusBar()->showMessage(text, 15000);
    }
}


void MainWindow::on_toolButton_whatsThis_clicked()
{
    if (!QWhatsThis::inWhatsThisMode()) {
        QWhatsThis::enterWhatsThisMode();
    }
}


void MainWindow::_showInfoBox()
{
    if (_infoBoxMode == 0) {
        if (!MiscTools::MiscTools::lockDialog()) {
            QTimer::singleShot(10000, this, SLOT(_showInfoBox()));
            return;
        }
        _infoBoxMode = 2;
        Tools::provideInfo(this,
                           tr("Main View"),
                           tr("<i>This is your main image view.</i><br><br>"
                              "At first startup a simple copy assistant is presented. "
                              "It can be used to load a directory of image/video files,"
                              " enter album information and copy the files to your"
                              " destination(s).<br><br>"
                              "By pressing 'advanced' you will see a lot more options,"
                              " to make your copying procedure more flexible. Hover"
                              " over a button to get information about it in the status bar."
                              "<br><br>"
                              "Whenever you want to go back to the assistant, select the"
                              " top left tab of the tab bar."),
                           tr("mainview"));
        MiscTools::MiscTools::unlockDialog();
    }
}

void MainWindow::on_actionRescan_current_directory_triggered()
{
    loadDirectory(_imageDir);
}

void MainWindow::on_actionSupport_by_donation_triggered()
{
    const QString donateText = _about.getDonateText();
    const QString donateTitle = tr("Support pivot");

    if (sender() == ui->actionSupport_by_donation) {
        QMessageBox::information(this,
                                 donateTitle,
                                 donateText);
    } else {
        // if first time: save date:
        // otherwise: check if one week has passed and asked for donation
        QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
        const int firstTimeYear = 2000;
        QDateTime firstTime = settings.value("firsttimestarteddonation",
                                             QDateTime(QDate(firstTimeYear,1,1),
                                                       QTime::currentTime())).toDateTime();
        if (firstTime.date().year() == firstTimeYear)
        {
            settings.setValue("firsttimestarteddonation", QDateTime::currentDateTime());
        } else if (QDateTime::currentDateTime().addDays(DAYS_UNTIL_DONATE_HINT) > firstTime ){
            if (!MiscTools::MiscTools::lockDialog()) {
                QTimer::singleShot(10000, this, SLOT(on_actionSupport_by_donation_triggered()));
                return;
            }
            Tools::provideInfo(this,
                               donateTitle,
                               donateText,
                               tr("donatehint"));
            MiscTools::MiscTools::unlockDialog();
        }
    }
}

void MainWindow::_on_albumReset()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images, tr("Please select at least one file to reset its album"), this, 1);

    _currentFiles.albumReset(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_albumUp()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the album from the last selected one, please select at least two files"),
                this);

    _currentFiles.albumUp(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_albumDown()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the album from the first selected one, please select at least two files"),
                this);
    _currentFiles.albumDown(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_timeReset()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images, tr("Please select at least one file to reset its time"), this, 1);

    _currentFiles.timeReset(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_timeUp()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the time from the last selected one, please select at least two files"),
                this);
    _currentFiles.timeUp(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_timeDown()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the time from the first selected one, please select at least two files"),
                this);

    _currentFiles.timeDown(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_destinationsReset()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the destinations from the first selected one, please select at least two files"),
                this);
    _currentFiles.destinationReset(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_destinationsUp()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the destinations from the first selected one, please select at least two files"),
                this);

    _currentFiles.destinationUp(rows);
    ui->tableView_images->clearSelection();
}

void MainWindow::_on_destinationsDown()
{
    QVector<int> rows = getSortedSelected(
                ui->tableView_images,
                tr("If you want to copy the destinations from the first selected one, please select at least two files"),
                this);

    _currentFiles.destinationDown(rows);
    ui->tableView_images->clearSelection();
}

void  MainWindow::_on_selectAll()
{
    ui->tableView_images->selectAll();
}

void  MainWindow::_on_deselectAll()
{
    ui->tableView_images->clearSelection();
}

void MainWindow::_resetAssistant()
{
    for (int i=0; i<_assistantButtons.size(); i++) {
        if (_assistantButtons[i] != NULL) {
            _assistantButtons[i]->setDone(":/images/empty.svg", _assistantButtonMargin);
        }
        if (_assistantButtons[i] != _assistantButtonLoadDirectory) {
            _assistantButtons[i]->setEnabled(false);
        }
    }
    _randomImageList->setVisible(false);
}

void MainWindow::_setDoneAssistant(MiscTools::ImageButton* button)
{
    if (button != NULL && button->isEnabled()) {
        button->setDone(":/images/assistant_done.svg", _assistantButtonMargin);
    }

    if (button == _assistantButtonLoadDirectory) {
        for (int i=0; i<_assistantButtons.size(); i++) {
            _assistantButtons[i]->setEnabled(true);
        }
    }
}

void MainWindow::on_pushButton_assAdvanced_toggled(bool checked)
{
    if (_assistantButtonReviewFiles != NULL) {
        _assistantButtonReviewFiles->setVisible(checked);
    }
    if (_assistantButtonAdaptTimes != NULL) {
        _assistantButtonAdaptTimes->setVisible(checked);
    }
//    if (_assistantButtonSettings != NULL) {
//        _assistantButtonSettings->setVisible(checked);
//    }
    if (_assistantButtonKeepDirectories != NULL) {
        _assistantButtonKeepDirectories->setVisible(checked);
    }
    if (_assistantButtonViewImages != NULL) {
        _assistantButtonViewImages->setVisible(checked);
    }
}

void MainWindow::_on_reviewFiles_clicked()
{
    ui->tabWidget->setCurrentWidget(ui->tab_overview);
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_delegate, SLOT(showInfo()));
    _setDoneAssistant(_assistantButtonReviewFiles);
}

void MainWindow::_on_viewImages_clicked()
{
    _lastRow = 0;
    _onViewAll();
    ViewImageDialogExtended * viewImages = ViewImageDialogExtended::instance(this);
    QTimer::singleShot(WAIT_FOR_INFO_BOX, viewImages, SLOT(showInfo()));
    _setDoneAssistant(_assistantButtonViewImages);
}

void MainWindow::_on_adaptTimes_clicked()
{
    ui->tabWidget->setCurrentWidget(ui->tab_adapttimes);
    QTimer::singleShot(WAIT_FOR_INFO_BOX, &_adaptTimeAssistant, SLOT(showInfo()));
    _setDoneAssistant(_assistantButtonAdaptTimes);
}

void MainWindow::on_actionCheck_for_new_version_triggered()
{
    qint32 major, minor, patch;
    QString link, text;
    bool connection;

    QProgressDialog* bar = NULL;
    if (sender() == ui->actionCheck_for_new_version) {
        bar = new QProgressDialog(tr("Retrieving version data"), tr("asdf"), 0, 0, this);
        bar->setCancelButton(0);
        bar->show();
    }

    //QDataStream data(reply);

    const bool result = version_pivot.checkForUpdateVersion(NEW_VERSION_LOCATION_PIVOT,
                                                            NEW_VERSION_MAGIC_NUMBER,
                                                            major,
                                                            minor,
                                                            patch,
                                                            link,
                                                            text,
                                                            connection);
    if (bar != NULL) {
        bar->close();
        bar->deleteLater();
    }

    // only show once a week
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.setValue("main_geometry", saveGeometry());
    QDate lastHint = settings.value("lastShowUpdate", QDate(1950,1,1)).toDate();
    QDate oneWeekBack = QDate::currentDate().addDays(-7);
    const bool autoCheck = lastHint < oneWeekBack || sender() == ui->actionCheck_for_new_version;

    if (result && autoCheck)
    {
        QDialog *diag = new QDialog(this);
        diag->setModal(true);
        QVBoxLayout* layH = new QVBoxLayout();
        diag->setLayout(layH);
        diag->layout()->addWidget(new QLabel(tr("New version of <i>pivot</i> available:  <b>%1.%2.%3</b> (current: %4.%5.%6)").arg(major).arg(minor).arg(patch).arg(version_pivot.getVersionMajor()).arg(version_pivot.getVersionMinor()).arg(version_pivot.getVersionPatch())));
        QLabel* dlLabel = new QLabel(QString(tr("Download: <a href=\"%1\">%2</a>")).arg(link).arg(link));
        dlLabel->setTextInteractionFlags(Qt::LinksAccessibleByKeyboard | Qt::LinksAccessibleByMouse);
        dlLabel->setOpenExternalLinks(true);
        diag->layout()->addWidget(dlLabel);
        diag->layout()->addWidget(new QLabel(tr("<b>Changelog:</b>")));
        QTextBrowser* tb = new QTextBrowser();
        tb->setText(text);
        tb->setReadOnly(true);
        diag->layout()->addWidget(tb);
        QPushButton *okay = new QPushButton(tr("close"));
        connect(okay, SIGNAL(clicked()), diag, SLOT(deleteLater()));
        QHBoxLayout *lay = new QHBoxLayout();
        lay->addStretch();
        lay->addWidget(okay);
        lay->addStretch();
        layH->addLayout(lay);
        diag->show();
        settings.setValue("lastShowUpdate", QDate::currentDate());
    } else {
        if (!connection && sender() == ui->actionCheck_for_new_version) {
            QMessageBox::warning(this, tr("no connection"), tr("No connection to update information found. Please check manually"), QMessageBox::Ok);
        } else if (sender() == ui->actionCheck_for_new_version) {
            QMessageBox::information(this, tr("no update necessary"), tr("There is no new version of pivot!"), QMessageBox::Ok);
        }
    }
}

void MainWindow::on_action_Help_triggered()
{
    QDesktopServices::openUrl(tr("http://www.dominik-ruess.de/pivot/pivot_help_en.html"));
}

void MainWindow::on_tabWidget_currentChanged(int index)
{
    QWidget* newTab = ui->tabWidget->widget(index);
    if (_currentTab == ui->tab_album) {
        _eventAssistant._on_pushButton_finish_clicked();
    } else if (_currentTab == ui->tab_keepdirectory) {
        _directoryAssistant._on_pushButton_finish_clicked();
    } else if (_currentTab == ui->tab_adapttimes) {
        _adaptTimeAssistant._on_pushButton_finish_clicked();
    }

    if (newTab == ui->tab_adapttimes) {
       _onAdaptTimes();
    } else if (newTab == ui->tab_keepdirectory) {
        _onSubDirMap();
    } else if (newTab == ui->tab_album) {
        _onManageAlbums();
    } else if (newTab == ui->tab_preview) {
        on_actionPreview_copy_results_triggered();
    }

    _currentTab = newTab;
}
