/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QCheckBox>
#include <QPainter>
#include <QDebug>
#include <QProgressBar>
#include <QStylePainter>
#include <QDateTimeEdit>
#include <QComboBox>
#include <QLineEdit>
#include <QTimer>

#include <imageviewmodel.h>
#include <tools.h>
#include "viewdelegate.h"
#include "misctools.h"


ViewDelegate::ViewDelegate( QWidget *parent)
    : ImageViewDelegate(parent)
    , _infoHasBeenShown(false)
{

}

void ViewDelegate::paint ( QPainter * painter,
                           const QStyleOptionViewItem & option,
                           const QModelIndex & index ) const
{
    if (index.column() < IMAGE_VIEW_COLUMNS) {
        return ImageViewDelegate::paint(painter, option, index);
    } else if (index.column() == DELEGATE_EVENT_COLUMN) {
        const int ID = index.model()->data(index, Qt::DisplayRole).toInt();
        QString text = "";
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].ID == ID) {
                text = _albums[i].name;
                break;
            }
        }

        QItemDelegate::drawBackground(painter, option, index);
        painter->drawText(option.rect, Qt::AlignCenter, text);
        QItemDelegate::drawFocus(painter, option, option.rect);
    } else {
        return QItemDelegate::paint(painter, option, index);
    }
}


QSize ViewDelegate::sizeHint ( const QStyleOptionViewItem & option, const QModelIndex & index ) const
{
    if (index.column() <  IMAGE_VIEW_COLUMNS) {
        return ImageViewDelegate::sizeHint(option, index);
    } else if (index.column() == DELEGATE_EVENT_COLUMN) {
        const int ID = index.model()->data(index, Qt::EditRole).toInt();
        QString text = "No Album";
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].ID == ID) {
                text = _albums[i].name;
            }
        }
        QFontMetrics fm (_parent->font());
        //return QSize(fm.width(text), qMax(fm.height(), _thumbnailsize));
        return QSize(fm.width(text), fm.height());
    } else {
        QSize sizehint = QItemDelegate::sizeHint(option, index);
        //sizehint.setHeight( qMax(sizehint.height(), _thumbnailsize));
        sizehint.setHeight( sizehint.height());
        return sizehint;
    }
}

QWidget* ViewDelegate::createEditor(QWidget *parent,
    const QStyleOptionViewItem &option,
    const QModelIndex& index) const
{
    if (index.column() == DELEGATE_DATE_COLUMN) {
        QDateTimeEdit* editor = new QDateTimeEdit(parent);
        //editor->setDisplayFormat(Qt::DefaultLocaleLongDate);
        editor->setCalendarPopup(true);

        return editor;
    } else if (index.column() == DELEGATE_EVENT_COLUMN) {
        QComboBox* editor = new QComboBox(parent);
        editor->addItem(tr("None"), -1);
        for (int i=0; i<(int)_albums.size(); i++) {
            editor->addItem(_albums[i].name, _albums[i].ID);
        }
        return editor;
    } else if (_useComments && index.column() == 5) {
        QLineEdit* editor = new QLineEdit(parent);
        editor->setText(index.model()->data(index, Qt::EditRole).toString());
        return editor;
    } else {
        return QItemDelegate::createEditor(parent, option, index);
    }
}

void ViewDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
{
    if (index.column() == DELEGATE_DATE_COLUMN) {
        QDateTime value = index.model()->data(index, Qt::EditRole).toDateTime();
        QDateTimeEdit *dateTime = static_cast<QDateTimeEdit*>(editor);
        dateTime->setDateTime(value);
    } else if (index.column() == DELEGATE_EVENT_COLUMN) {
        QComboBox *box = static_cast<QComboBox*>(editor);
        const int currId = index.model()->data(index, Qt::EditRole).toInt();
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].ID == currId) {
                box->setCurrentIndex(i);
            }
        }
    } else if (_useComments && index.column() == 5) {
        QLineEdit* line = static_cast<QLineEdit*>(editor);
        line->setText(index.model()->data(index, Qt::EditRole).toString());
    } else {
        QItemDelegate::setEditorData(editor, index);
    }
}

void ViewDelegate::setModelData(QWidget *editor, QAbstractItemModel *model,
                  const QModelIndex &index) const
{

    if (index.column() == DELEGATE_DATE_COLUMN) {
        QDateTimeEdit *dateTime = static_cast<QDateTimeEdit*>(editor);
        dateTime->interpretText();
        QDateTime value = dateTime->dateTime();

        model->setData(index, value, Qt::EditRole);
    } else if (index.column() == DELEGATE_EVENT_COLUMN) {
        QComboBox *box = static_cast<QComboBox*>(editor);
        const int ID = box->itemData(box->currentIndex(), Qt::UserRole).toInt();
        model->setData(index, ID, Qt::EditRole);
    } else if (_useComments && index.column() == 5) {
        QLineEdit *line = dynamic_cast<QLineEdit*>(editor);
        const QString comment = line->text();
        model->setData(index, comment, Qt::EditRole);
    } else {
        return QItemDelegate::setModelData(editor, model, index);
    }
}

void ViewDelegate::updateEditorGeometry(QWidget *editor,
    const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    if (index.column() == DELEGATE_DATE_COLUMN
            || index.column() == DELEGATE_EVENT_COLUMN) {
        editor->setGeometry(option.rect);    
    } else {
        QItemDelegate::updateEditorGeometry(editor, option, index);
    }
}


void ViewDelegate::showInfo()
{
    if (!_infoHasBeenShown) {
        if (!MiscTools::MiscTools::lockDialog()) {
            QTimer::singleShot(10000, this, SLOT(showInfo()));
            return;
        }
        _infoHasBeenShown = true;
        Tools::provideInfo(0,
                           tr("Review Image Meta Data"),
                           tr("<i>This is the image meta data editing tab.</i><br><br>"
                              "Here you can adapt the files' times and albums:"
                              "<ul>"
                              "<li>You can limit the display to only show files with"
                              " missing data</li>"
                              "<li>If you wish, add a comment to a file (you may want "
                              " to change settings, to activate this option)</li>"
                              "<li>Copy meta data from a certain file to others</li>"
                              "<li>Select certain files to be viewed in the image viewer</li>"
                              "<li>If you uncheck a file, it won't be copied at all. Using the image viewer"
                              " you can distinguish copying to certain destinations</li>"
                              "</ul>Most of the options are available through the (context) menu."),
                           tr("viewimagedata"));
        MiscTools::MiscTools::unlockDialog();
    }

}
