/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include "viewimagedialogextended.h"

#include <QTimer>
#include <QSizePolicy>
#include <QAction>
#include <QGraphicsDropShadowEffect>
#include <QGraphicsBlurEffect>
#include "ui_viewimagedialog.h"

#include "tools.h"

// singleton static pointer
ViewImageDialogExtended* ViewImageDialogExtended::_instance = NULL;


ViewImageDialogExtended* ViewImageDialogExtended::instance(QWidget* parent)
{
  if(_instance == NULL){
      _instance = new ViewImageDialogExtended(parent);
      _instance->setVisible(false);
  }
  return _instance;
}

void ViewImageDialogExtended::closeInstance()
 {
     if(_instance == NULL) {
         return;
     } else {
         _instance->close();
     }
 }

ViewImageDialogExtended::~ViewImageDialogExtended()
{
}

ViewImageDialogExtended::ViewImageDialogExtended(QWidget *parent)
    : ViewImageDialog(parent)
    , _useNumTargets(0)
    , _infoHasBeenShown(false)
    ,_styleRed("QLabel { color : #FF6666; opacity:0.75;}")
    ,_styleGreen("QLabel { color : #66FF66; opacity:0.75;}")
{	
    //setVisible(false);
    //hide();
    _toggleCopyTarget.resize(MAX_NUM_COPY_DESTINATIONS);
    for (int i=0; i<qMin(9,MAX_NUM_COPY_DESTINATIONS); i++) {
        _toggleCopyTarget[i] = new QAction(tr("Toggle copy to target %1").arg(i+1), this);
        _toggleCopyTarget[i]->setShortcut(tr("%1").arg(i+1));
        _toggleCopyTarget[i]->setShortcutContext(Qt::ApplicationShortcut);
        addAction(_toggleCopyTarget[i]);

        connect(_toggleCopyTarget[i], SIGNAL(triggered()), this, SLOT(toggleCopyTargetAction()));
    }


    _copyTargetStatus = new QLabel(tr(""));
    _copyTargetStatus->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    _copyTargetStatus->setStyleSheet(tr("border: 1px solid green;"));


    QFont font("Monospace");
    font.setStyleHint(QFont::TypeWriter);
    font.setStretch(85);
    _copyTargetStatus->setFont(font);

    //ui->horizontalLayout_2
    ui->horizontalLayout_raw->addWidget(_copyTargetStatus);

    // connect update target
    connect(this, SIGNAL(newImageLabel()), SLOT(_updateCopyTargetDisplay()));

    connect(this, SIGNAL(fileOverlayChanged()), SLOT(_updateCopyTargetDisplay()));	
}

void ViewImageDialogExtended::show()
{
    setVisible(true);
    ViewImageDialog::show();
}

void ViewImageDialogExtended::toggleCopyTargetAction()
{
    if (_currentPosition < 0)
        return;

    QObject * sender = QObject::sender();
    for (int i=0; i<MAX_NUM_COPY_DESTINATIONS; i++) {
        if ((QObject*) _toggleCopyTarget[i] == sender) {
            // set current image copy target
            ImageData* ptr = getFileInfoPtr(_currentPosition).get();
            if (ptr == 0) {
                continue;
            }
            ImageDataExtended* tmpC = static_cast<ImageDataExtended*>(ptr);
            tmpC->toggleCopyDestinationStatus(i);
            if (ui->checkBox_hideRaw->isChecked()) {
                // make sure, corresponding raw files also obtain the toggle status
                for (int j=0; j<(int)getNumFilesOverall(); j++) {
                    ImageData* ptr = getFileInfoPtrOverallList(j).get();
                    if (ptr == 0) {
                        continue;
                    }
                    ImageDataExtended* tmpJ = static_cast<ImageDataExtended*>(ptr);
                    if (tmpJ->getFileInfo().canonicalFilePath() != tmpC->getFileInfo().canonicalFilePath()
                            && tmpJ->getFileInfo().canonicalPath() == tmpC->getFileInfo().canonicalPath()
                            && tmpJ->getFileInfo().completeBaseName().toLower() == tmpC->getFileInfo().completeBaseName().toLower()
                            && tmpJ->getCopyDestinationStatus(i) != tmpC->getCopyDestinationStatus(i)) {
                        tmpJ->toggleCopyDestinationStatus(i);
                    }
                }
            }
            break;
        }
    }
    _updateCopyTargetDisplay();
}


void ViewImageDialogExtended::setNumberOfDestinations(const QStringList& destinations)
{
    const int i = destinations.length();
    if (_useNumTargets == i) {
        return;
    }
    _targets = destinations;
    _useNumTargets = i;
    _updateCopyTargetDisplay();


    for (int i=0; i< _targetListsOverlay.size(); i++) {
        _targetListsOverlay[i]->deleteLater();
        _targetListsOverlay[i] = 0;
    }

    _targetListsOverlay.resize(_useNumTargets);
    for (int i=0; i<_useNumTargets; i++) {
        _targetListsOverlay[i] = new QLabel(this);//ui->scrollArea);
        QGraphicsDropShadowEffect* effect = new QGraphicsDropShadowEffect();
        effect->setBlurRadius(1);
        effect->setOffset(1.0);
        _targetListsOverlay[i]->setGraphicsEffect(effect);
        _targetListsOverlay[i]->setStyleSheet(_styleGreen);
    }
}



void ViewImageDialogExtended::_updateCopyTargetDisplay()
{
    //if not needed, don't use
    if (_useNumTargets == 0 || _currentPosition < 0 || _currNumImages - 1 < _currentPosition) {
        _copyTargetStatus->setText("");
        return;
    }

    QString out = tr("Copy to:");
    if (_useNumTargets != _targetListsOverlay.size()) {
        qDebug() << "number of targets and respective displays don't align";
    }

    const int add = _fullScreen && _fileOverlayActive ? 1 : 0;
    for (int i=0; i<qMin(_useNumTargets, MAX_NUM_COPY_DESTINATIONS); i++) {
        _targetListsOverlay[i]->setVisible(_fullScreen);
        QFontMetrics fm( _targetListsOverlay[i]->font() );
        ImageData* ptr = getFileInfoPtr(_currentPosition).get();
        if (ptr == 0) {
            continue;
        }
        if ( static_cast<ImageDataExtended*>(ptr)->getCopyDestinationStatus(i)) {
            out.append(tr(" %1", "copy location target number").arg(i+1));
            _targetListsOverlay[i]->setText(tr("Copy to target %1: yes (%2)").arg(i+1).arg(_targets[i]));
            _targetListsOverlay[i]->setStyleSheet(_styleGreen);
        } else {
            out.append(tr("  ", "copy location empty target"));
            _targetListsOverlay[i]->setText(tr("Copy to target %1: no (%2)").arg(i+1).arg(_targets[i]));
            _targetListsOverlay[i]->setStyleSheet(_styleRed);
        }
        _targetListsOverlay[i]->setMinimumWidth(ui->graphicsView->width()-10);
        _targetListsOverlay[i]->move(0, fm.height()*1.5*(i+add));
    }
    _copyTargetStatus->setText(out);
}

void ViewImageDialogExtended::_toggleFullscreen()
{
    ViewImageDialog::_toggleFullscreen();
    _updateCopyTargetDisplay();
}



void ViewImageDialogExtended::showInfo()
{
    if (!_infoHasBeenShown) {
        if (!MiscTools::MiscTools::lockDialog()) {
            QTimer::singleShot(10000, this, SLOT(showInfo()));
            return;
        }
        _infoHasBeenShown = true;
        Tools::provideInfo(this,
                           tr("Review Images"),
                           tr("<i>This is the image viewer dialog.</i><br><br>"
                              "Here you can review images, i.e. for copying.<br>"
                              "If you have setup multiple copy destinations, you can choose"
                              " to copy the image to each destination (default) or"
                              " remove images for certain or all destinations."
                              "Just skip trough the images: <p><b>For each image press"
                              " any of the keys from 1-9 to enable/disable copying"
                              " to the respective destination.</b> This will be indicated "
                              "by the destination number in the lower right corner.</p>"
                              "This gives you the opportunity for pre selection of"
                              " images/video, for instance to keep all files in backup"
                              " but to remove certain files in the actual galleries."),
                           tr("viewimages"));
        MiscTools::MiscTools::unlockDialog();
    }
}
