/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2015, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QBrush>
#include <QApplication>
#include <QPalette>
#include <QDebug>
#include <QDateTimeEdit>

#include "viewmodel.h"
#include "tools.h"

ViewModel::ViewModel(QObject *parent)
    : ImageViewModel(parent)
    , _showDetails(false)
    , _showCrop(false)
    , _showOnlyMissing(false)
    , _numMissing(0)
    , _lastTime(QDateTime::currentDateTime())
    , _useDifferentTags(false)
    , _useComments(false)
    , _numCopyDestinations(0)
{
    Q_INIT_RESOURCE(infoimages);
}

int ViewModel::rowCount(const QModelIndex& /* parent */) const
{
    if (_images.count() == 0) {
        return 1;
    } else if (!_showOnlyMissing) {
        return _images.count();
    } else {
        return (_numMissing == 0 ? 1 : _numMissing);
    }
}

int ViewModel::columnCount(const QModelIndex& /* parent */) const
{
    int numCols = _useComments ? IMAGE_VIEW_COLUMNS + 5: IMAGE_VIEW_COLUMNS + 4;
    numCols += _numCopyDestinations > 1 ? 1 : 0;
    return numCols;
}

bool ViewModel::_isCommentColumn(const int index) const
{
    if (_useComments && index == IMAGE_VIEW_COLUMNS+3) {
        return true;
    }
    return false;
}

bool ViewModel::_isStatusColumn(const int index) const
{
    int statusPos = 5;
    if (_useComments) {
        statusPos++;
    }

    if (_numCopyDestinations > 1) {
        statusPos++;
    }

    return index== statusPos;
}

bool ViewModel::_isTargetColumn(const int index) const
{
    int statusPos = 5;
    if (_useComments) {
        statusPos++;
    }

    return (_numCopyDestinations > 1 && index == statusPos) ;
}

QVariant ViewModel::data(const QModelIndex& index, int role) const
{
    if (!index.isValid()) {
        return QVariant();
    }

    if (_images.count() == 0
            && index.column() == IMAGE_VIEW_FILENAME_IDX
            && role == Qt::DisplayRole
            && !_showOnlyMissing)
    {
        return QString(tr("select a directory first"));
    } else if (_numMissing == 0
                && role == Qt::DisplayRole
               && _showOnlyMissing)
    {
        if ( index.column() == IMAGE_VIEW_FILENAME_IDX) {
            return QString(tr("no files without time stamp/album"));
        } else {
            return QVariant();
        }
    } else if (_images.count() == 0) {
        return QVariant();
    }

    if (_showOnlyMissing
            &&  (_backMap.size() == 0
                 || index.row() >= _backMap.size() )
            ) {
        return QVariant();
    }

    if (index.column() < IMAGE_VIEW_COLUMNS) {
        if (_showOnlyMissing
                && _backMap.size() > 0
                && index.row() < _backMap.size() ) {
            return ImageViewModel::data(_backMap[index.row()], index.column(), role);
        } else {
            return ImageViewModel::data(index, role);
        }
        return QVariant();
    }

    const ImageDataExtended i = _showOnlyMissing ? getImageDataExtended(_backMap[index.row()]) : getImageDataExtended(index.row());

    if (role == Qt::TextAlignmentRole) {
        return int(Qt::AlignCenter | Qt::AlignVCenter);
    } else if (role == Qt::ToolTipRole
               || role == Qt::StatusTipRole) {
        QString album = "None";
        for (int k=0; k<_albums.size(); k++) {
            if (_albums[k].ID == i.getAlbumID()) {
                album = _albums[k].name;
            }
        }
        QString extension = "";
        double size = i.getFileInfo().size();
        if (size * 1.5 > 1024) {
            extension = tr("Kb");
            size /= 1024;
        }
        if (size * 1.5 > 1024) {
            extension = tr("Mb");
            size /= 1024;
        }
        if (size * 1.5 > 1024) {
            extension = tr("Gb");
            size /= 1024;
        }

        return tr("File: '%1'\nFile time: '%2'\nAssigned album: '%3' "
                  "%4\nSize: %6")
                .arg(i.getFileInfo().canonicalFilePath())
                .arg(i.getTime().toString(Qt::DefaultLocaleLongDate))
                .arg(album)
                .arg(_useComments ? tr("\nComment: %1").arg(i.getComment()) : QString(""))
                .arg(QString("%1 %2").arg(size, 0, 'f', 2 ).arg(extension));
    } else if (role == Qt::EditRole) {
        if (index.column() == 4) {
            _currentEditing = index.row();
            if (i.getTime().isValid()) {
                return i.getTime();
            } else {
                // otherwise return last edited time
                return _lastTime;
            }
        } else if (_isCommentColumn(index.column())) {
            _currentEditing = index.row();
            return i.getComment();
        } else if (index.column() == 3) {
            _currentEditing = index.row();
            return i.getAlbumID();
        } else if (_isTargetColumn(index.column())) {
            _currentEditing = index.row();
            QString targets;
            for (int j=0; j<_numCopyDestinations; j++) {
                if (i.getCopyDestinationStatus(j)) {
                    if (targets.length() > 0) {
                        targets += ", ";
                    }
                    targets += QString("%1").arg(j+1);
                }
            }
            return targets;
        }
    } else if (Qt::CheckStateRole == role && index.column() == 4 && _useDifferentTags) {
        return i.getDayIsKnown() ? Qt::Checked : Qt::Unchecked;
    } else if (role == Qt::BackgroundRole) {
        QBrush colorGood( index.row()%2 == 0 ? QColor(200,255,200) : QColor(180,235,180));
        QBrush colorOk( index.row()%
                        2 == 0 ? QColor(255,255,200) : QColor(235,235,180));
        QBrush colorManual( index.row()%2 == 0 ? QColor(255,215,120) : QColor(235,195,100));
        QBrush colorBad( index.row()%2 == 0 ? QColor(255,200,200) : QColor(235,180,180));
        QBrush colorDone( index.row()%2 == 0 ? QColor(180,235,180) : QColor(160,215,160));
        QBrush colorAlternate = QApplication::palette("tableView_Images")
                .brush(QPalette::Active,
                       index.row()%2 == 0? QPalette::Base : QPalette::AlternateBase);

        if ( index.column() == IMAGE_VIEW_COLUMNS) {
            return colorAlternate;
        } else if (index.column() == IMAGE_VIEW_COLUMNS+1) {
            if (i.getAlbumID() == -1) {
                return colorBad;
            } else if (i.getIsManualAlbum()) {
                return colorManual;
            } else {
                return colorAlternate;
            }
        } else if (index.column() == IMAGE_VIEW_COLUMNS+2) {
            if ( i.getIsManualTime() ) {
                return colorManual;
            } else if (i.getIsSemiAutomatedTime()
                       || i.getIsShiftedTime()) {
                return colorOk;
            } else if (i.getIsAutomatedTime()) {
                return colorAlternate;
            } else {
                return colorBad;
            }
        } else if (_isCommentColumn(index.column())) {
            return colorAlternate;
        } else if (_isStatusColumn(index.column())) {
            if (i.getCopyStatus() == CopyStatusCopyError) {
                return colorBad;
            } else if (i.getCopyStatus() == CopyStatusAlreadyCopied) {
                return colorDone;
            } else if (i.getCopyStatus() == CopyStatusRemovedUnique) {
                return colorOk;
            } else if (i.getCopyStatus() == CopyStatusSuccess) {
                return colorGood;
            } else {
                return colorAlternate;
            }
        }
    } else if (role == Qt::DisplayRole) {
        if ( index.column() == IMAGE_VIEW_COLUMNS) {
            return i.getFileInfo().canonicalPath().replace(0,rootDirectory.canonicalPath().length(), ".");
        } else if  ( index.column() == IMAGE_VIEW_COLUMNS+1) {
            int albumId = i.getAlbumID();
            if (albumId >= 0) {
                bool found = false;
                for (int i=0; i<(int)_albums.size(); i++) {
                    if (_albums[i].ID == albumId) {
                        return _albums[i].ID;
                    }
                }
                if (!found) {
                    albumId = -1;
                }
            }
            return albumId;
        } else if  ( index.column() == IMAGE_VIEW_COLUMNS+2) {
            if (i.getIsManualTime() || i.getIsAutomatedTime() || i.getIsSemiAutomatedTime()) {
                return i.getTime().toString(Qt::DefaultLocaleLongDate);
            } else {
                return " --- ";
            }
        } else if (_isCommentColumn(index.column()) ) {
            return i.getComment();
        } else if (_isTargetColumn(index.column()) ) {
            QString targets;
            for (int j=0; j<_numCopyDestinations; j++) {
                if (i.getCopyDestinationStatus(j)) {
                    if (targets.length() > 0) {
                        targets += ", ";
                    }
                    targets += QString("%1").arg(j+1);
                }
            }
            return targets;
        } else if  ( _isStatusColumn(index.column()) ) {
            if (i.getCopyStatus() == CopyStatusCopyError) {
                return tr("Error - not copied");
            } else if (i.getCopyStatus() == CopyStatusAlreadyCopied) {
                return tr("Already copied");
            } else if (i.getCopyStatus() == CopyStatusRemovedUnique) {
                return tr("Duplicate");
            } else if (i.getCopyStatus() == CopyStatusSuccess) {
                return tr("Successful");
            } else {
                return QVariant();
            }
        }
    }

    return QVariant();
}

QVariant ViewModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (section < IMAGE_VIEW_COLUMNS) {
        return ImageViewModel::headerData(section, orientation, role);
    }

    if (role == Qt::DisplayRole && orientation == Qt::Horizontal) {
        if (section == IMAGE_VIEW_COLUMNS) {
            return tr("Source Sub Folder");
        } else if (section == IMAGE_VIEW_COLUMNS+1) {
            return tr("Assigned Album");
        } else if (section == IMAGE_VIEW_COLUMNS+2) {
            return _useDifferentTags ? tr("Creation Date (uncheck if day/time not known)"): tr("Creation Date");
        } else if (_isCommentColumn(section)) {
            return tr("Comment");
        } else if (_isStatusColumn(section)) {
            return tr("Status");
        } else if (_isTargetColumn(section)) {
            return tr("Destinations");
        } else {
            return QVariant();
        }
    }
    return QVariant();
}

bool ViewModel::setData ( const QModelIndex & index, const QVariant & value, int role)
{
    bool edited = false;
    _currentEditing = -1;
    if (index.column() < IMAGE_VIEW_COLUMNS) {
        return ImageViewModel::setData(index, value, role);
    } else if (index.column() == 4 ) {

        const int row = _showOnlyMissing ? _backMap[index.row()] : index.row();
        ImageDataExtended* data = dynamic_cast<ImageDataExtended*>(_images[row].get());
        if (role == Qt::CheckStateRole  && _useDifferentTags)
        {
            data->setDayIsKnown(value.toBool());
            edited = true;
        } else {
            // time selection
            if (value.toDateTime() != data->getAutomatedTime()) {
                data->setManualTime(value.toDateTime());
            }

            _lastTime = value.toDateTime();

            // now also update album, if not set manually
            if (!data->getIsManualAlbum()) {
                int ID = -1;
                for (int i=0; i<(int)_albums.size(); i++) {
                    if (_albums[i].from <= data->getTime()
                            && _albums[i].to > data->getTime()) {
                        ID = _albums[i].ID;
                    }
                }
                data->setAlbumID(ID, false);
            }
            edited = true;
        }
    } else if (_isCommentColumn(index.column())) {
        const QString comment = value.toString();
        const int row = _showOnlyMissing ? _backMap[index.row()] : index.row();
        ImageDataExtended* data = dynamic_cast<ImageDataExtended*>(_images[row].get());
        data->setComment(comment);
        edited = true;
    } else if (_isTargetColumn(index.column())) {
        //hier von string parsen
        const QString values = value.toString();
        const QStringList vals = values.split(",", QString::SkipEmptyParts);
        const int row = _showOnlyMissing ? _backMap[index.row()] : index.row();
        ImageDataExtended* data = dynamic_cast<ImageDataExtended*>(_images[row].get());

        bool anyBefore = false;
        for (int j=0; j<_numCopyDestinations; j++) {
            anyBefore |= data->getCopyDestinationStatus(j);
            if (data->getCopyDestinationStatus(j)) {
                data->toggleCopyDestinationStatus(j);
            }
        }

        for (int j=0; j< (int)vals.size(); j++) {
            const int pos = vals[j].toInt() - 1;
            if (pos < _numCopyDestinations
                    && pos >= 0
                    && !data->getCopyDestinationStatus(pos)) {
                data->toggleCopyDestinationStatus(pos);
            }
        }

        bool anyAfter = false;
        for (int j=0; j<_numCopyDestinations; j++) {
            anyAfter |= data->getCopyDestinationStatus(j);
        }

        if (!anyBefore && anyAfter) {
            data->setUsed(true);
        } else if (!anyAfter) {
            data->setUsed(false);
        }
        edited = true;
    } else if (index.column() == 3) {
        // album selection
        const int row = _showOnlyMissing ? _backMap[index.row()] : index.row();
        ImageDataExtended* data = dynamic_cast<ImageDataExtended*>(_images[row].get());
        int ID = -1;
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].from <= data->getTime()
                    && _albums[i].to > data->getTime()) {
                ID = _albums[i].ID;
            }
        }

        data->setAlbumID(value.toInt(), ID != value.toInt());
        edited = true;
    }

    _updateViews(index.row());

    if (edited) return true;
    return QAbstractTableModel::setData(index, value, role);
}

Qt::ItemFlags ViewModel::flags ( const QModelIndex & index ) const
{
    if ( index.column() < IMAGE_VIEW_COLUMNS) {
        return ImageViewModel::flags(index);
    } else if (index.column() == 3
               || _isCommentColumn(index.column())
               || _isTargetColumn(index.column()) ) {
        return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
    }else if (index.column() == 4) {
        if (_useDifferentTags) {
            return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable | Qt::ItemIsUserCheckable;
        } else {
            return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
        }
    } else {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
}

void ViewModel::setShowDetails(const bool showDetails)
{
    emit layoutAboutToBeChanged();
    _showDetails = showDetails;
    emit layoutChanged();
}


void ViewModel::setShowCrop(const bool showCrop)
{
    emit layoutAboutToBeChanged();
    _showCrop = showCrop;
    emit layoutChanged();
}

const ImageDataExtended& ViewModel::getImageDataExtended(const int pos) const
{
    if (pos <0 || pos > _images.size()) {
        throw "wrong index";
    }
    return * ( static_cast<ImageDataExtended*>(_images[pos].get()));
}


void ViewModel::setList(const QList<ImageDataPtr>& newList)
{
    emit layoutAboutToBeChanged();
    _images.clear();
    for (int i=0; i<newList.size(); i++) {
        _images.push_back( ImageDataPtr(new ImageDataExtended(*dynamic_cast<ImageDataExtended*>(newList[i].get()) ))  );
    }

    _computeNumMissing();

    emit layoutChanged();

    _updateViews();
}


void ViewModel::updateList(const int pos, const ImageDataExtended& im)
{
    if (pos >= _images.size() || pos < 0) {
        throw "wrong image index";
    }

    * (dynamic_cast<ImageDataExtended*>(_images[pos].get()) ) = im;
    _updateViews(pos);
}


void ViewModel::toggleNoTimeOrAlbum(const bool showMissing)
{
    emit layoutAboutToBeChanged();
    _computeNumMissing();

    _showOnlyMissing = showMissing;
    for (int i=0; i<_images.size(); i++) {
        ImageDataExtended tmp = getImageDataExtended(i);
        if (!tmp.getTime().isValid()
                || tmp.getAlbumID() == -1) {
            tmp.setShowNoTimeOrAlbumFlag(true);
            updateList(i, tmp);
        } else {
            tmp.setShowNoTimeOrAlbumFlag(false);
            updateList(i, tmp);
        }
    }
    emit layoutChanged();
}


void ViewModel::_computeNumMissing()
{
    _backMap.clear();
    _numMissing = 0;
    for (int i=0; i<_images.size(); i++) {
        if (!getImageDataExtended(i).getTime().isValid()
                || getImageDataExtended(i).getAlbumID() == -1) {
            _backMap.push_back(i);
            _numMissing++;
        }
    }
}

void ViewModel::timeDown(const QVector<int> rows)
{
    if (rows.size() == 0) return;

    const QString errTitle = tr("No start date");
    const QString errMsg = tr("The top most selected file has no valid date");

    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[0]]) :  getImageDataExtended(rows[0]);

    if (!itemSource.getTime().isValid()) {
        emit errorMessage(errTitle, errMsg);
        return;
    }

    for (int r=1; r<(int)rows.size(); r++) {
        const int row = _showOnlyMissing ? _backMap[rows[r]] : rows[r];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        itemTarget.setManualTime(itemSource.getTime());
        itemTarget.setAlbumID(-1, false);
        updateList(row, itemTarget);
    }

    albumReset(rows);
}

void ViewModel::timeUp(const QVector<int> rows)
{
    const QString errTitle = tr("No start date");
    const QString errMsg = tr("The bottom most selected file has no valid date");


    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[rows.size()-1]]) :  getImageDataExtended(rows[rows.size()-1]);

    if (!itemSource.getTime().isValid()) {
        emit errorMessage(errTitle, errMsg);
        return;
    }

    for (int i=rows.size()-2; i>=0; i--) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        itemTarget.setManualTime(itemSource.getTime());
        itemTarget.setAlbumID(-1, false);
        updateList(row, itemTarget);
    }

    albumReset(rows);
}

void ViewModel::timeReset(const QVector<int> rows)
{
    if (rows.size() == 0) return;
    for (int i=0; i<rows.size(); i++) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        itemTarget.resetToAutomated();

        if (!itemTarget.getIsManualAlbum()) {
            int ID = -1;
            for (int i=0; i<(int)_albums.size(); i++) {
                if (_albums[i].from <= itemTarget.getTime()
                        && _albums[i].to > itemTarget.getTime()) {
                    ID = _albums[i].ID;
                }
            }
            itemTarget.setAlbumID(ID, false);
        }

        updateList(row, itemTarget);
    }
}

void ViewModel::albumDown(const QVector<int> rows)
{
    if (rows.size() == 0) return;


    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[0]]) :  getImageDataExtended(rows[0]);

    for (int r=1; r<(int)rows.size(); r++) {
        const int row = _showOnlyMissing ? _backMap[rows[r]] : rows[r];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        int ID = -1;
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].from <= itemTarget.getTime()
                    && _albums[i].to > itemTarget.getTime()) {
                ID = _albums[i].ID;
            }
        }
        itemTarget.setAlbumID(itemSource.getAlbumID(), ID != itemSource.getAlbumID());
        updateList(row, itemTarget);
    }
}

void ViewModel::albumUp(const QVector<int> rows)
{

    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[rows.size()-1]]) :  getImageDataExtended(rows[rows.size()-1]);

    for (int i=rows.size()-2; i>=0; i--) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        int ID = -1;
        for (int i=0; i<(int)_albums.size(); i++) {
            if (_albums[i].from <= itemTarget.getTime()
                    && _albums[i].to > itemTarget.getTime()) {
                ID = _albums[i].ID;
            }
        }
        itemTarget.setAlbumID(itemSource.getAlbumID(), ID != itemSource.getAlbumID());
        updateList(row, itemTarget);
    }
}

void ViewModel::albumReset(const QVector<int> rows)
{
    if (rows.size() == 0) return;
    for (int i=0; i<rows.size(); i++) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        int ID = -1;
        for (int j=0; j<(int)_albums.size(); j++) {
            if (_albums[j].from <= itemTarget.getTime()
                    && _albums[j].to > itemTarget.getTime()) {
                ID = _albums[j].ID;
            }
        }
        itemTarget.setAlbumID(ID, false);
        updateList(row, itemTarget);
    }
}


void ViewModel::destinationDown(const QVector<int> rows)
{
    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[0]]) :  getImageDataExtended(rows[0]);

    for (int i=1; i< rows.size(); i++) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        for (int j=0; j<_numCopyDestinations; j++) {
            if (itemTarget.getCopyDestinationStatus(j) != itemSource.getCopyDestinationStatus(j)) {
                itemTarget.toggleCopyDestinationStatus(j);
            }
        }
        updateList(row, itemTarget);
    }

    albumReset(rows);
}

void ViewModel::destinationUp(const QVector<int> rows)
{
    ImageDataExtended itemSource = _showOnlyMissing ?  getImageDataExtended(_backMap[rows[rows.size()-1]]) :  getImageDataExtended(rows[rows.size()-1]);

    for (int i=rows.size()-2; i>=0; i--) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        for (int j=0; j<_numCopyDestinations; j++) {
            if (itemTarget.getCopyDestinationStatus(j) != itemSource.getCopyDestinationStatus(j)) {
                itemTarget.toggleCopyDestinationStatus(j);
            }
        }
        updateList(row, itemTarget);
    }

    albumReset(rows);
}

void ViewModel::destinationReset(const QVector<int> rows)
{
    if (rows.size() == 0) return;

    for (int i=0; i<rows.size(); i++) {
        const int row = _showOnlyMissing ? _backMap[rows[i]] : rows[i];
        ImageDataExtended itemTarget = getImageDataExtended(row);
        for (int j=0; j<_numCopyDestinations; j++) {
            if (!itemTarget.getCopyDestinationStatus(j)) {
                itemTarget.toggleCopyDestinationStatus(j);
            }
        }
        updateList(row, itemTarget);
    }
}

void ViewModel::updateFilesAlbums(const QVector<PhotoAlbum>& albums)
{
    _albums = albums;

    // assumes sorted albums
    for (int i=0; i<(int)_images.size(); i++) {
        ImageDataExtended* tmp = ( dynamic_cast<ImageDataExtended*>(_images[i].get()));
        if (tmp->getIsManualAlbum()) {
            // do not change the album if manually changed
            continue;
        }
        int albumId = -1;
        for (int j=0; j<(int)albums.size(); j++) {
            if (albums[j].from <= tmp->getTime()
                    && albums[j].to > tmp->getTime() ) {
                albumId = albums[j].ID;
            }
        }
        tmp->setAlbumID(albumId, false);
    }
    _updateViews();
}

static bool sortingByTime(const QPair<QPair<QDateTime, QString>,int>& e1, const QPair<QPair<QDateTime, QString>,int>& e2) {
    // sort by time, but if equal sort by filename
    return e1.first.first == e2.first.first ? e1.first.second < e2.first.second : e1.first.first < e2.first.first;
}

QVector<int> ViewModel::getImageTimeSortIndexes(const QVector<int> indexes) const
{
    // given desired (empty if all are desired) indexes, sort these by time and
    // return newly, sorted indexes
    QVector<QPair<QPair<QDateTime, QString>, int> > items;
    if (indexes.size() == 0) {
        items.resize(_images.size());
        for (int i=0; i < _images.size(); i++) {
            ImageDataExtended* tmp = ( dynamic_cast<ImageDataExtended*>(_images[i].get()));
            items[i] = QPair<QPair<QDateTime, QString>, int>(QPair<QDateTime, QString>(tmp->getTime(), tmp->getFileInfo().canonicalFilePath()), indexes[i]);
        }
    } else {
        items.resize(indexes.size());
        for (int i=0; i < indexes.size(); i++) {
            ImageDataExtended* tmp = ( dynamic_cast<ImageDataExtended*>(_images[indexes[i]].get()));
            items[i] = QPair<QPair<QDateTime, QString>, int>(QPair<QDateTime, QString>(tmp->getTime(), tmp->getFileInfo().canonicalFilePath()), i);
        }
    }
    qSort(items.begin(), items.end(), sortingByTime);

    QVector<int> indexesOut(items.size());
    for (int i=0; i< items.size(); i++) {
        indexesOut[i] = items[i].second;
    }
    return indexesOut;
}


void ViewModel::setUseDifferentTags(const bool use)
{
    if (use == _useDifferentTags) {
        return;
    }

    emit layoutAboutToBeChanged();
    _useDifferentTags = use;
    emit layoutChanged();
}


 void ViewModel::setUseComments(const bool useComments)
 {
     emit layoutAboutToBeChanged();
     _useComments= useComments;
     emit layoutChanged();
 }


 void ViewModel::setNumDestinations(const int num)
 {
     emit layoutAboutToBeChanged();
    _numCopyDestinations = num;
    emit layoutChanged();
 }
